/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;

import java.util.LinkedList;

/**
 * This is the base class for all First In First Out (FIFO) queuing disciplines.
 * There is an unbounded version ({@link UnboundedFIFOQDisc UnboundedFIFOQDisc})
 * and a bounded version ({@link BoundedFIFOQDisc BoundedFIFOQDisc}).
 * 
 * @see LIFOQDisc
 */
public abstract class FIFOQDisc extends StackableQueuingDiscipline {
   protected final java.util.List<QueueEntry> queue;

   public FIFOQDisc(SimNode ownNode) {
      super(ownNode);
      queue = new LinkedList<QueueEntry>();
   }

   public FIFOQDisc(SimNode ownNode, @SuppressWarnings("unused") Parameters pars) {
      this(ownNode);
   }

   @Override
   public int getMaxSizeInUnits() {
      return Integer.MAX_VALUE;
   }

   @Override
   public boolean isEmpty() {
      return queue.isEmpty();
   }

   @Override
   protected void push(QueueEntry qe) {
      queue.add(qe);
   }

   @Override
   protected QueueEntry peekInternal() {
      if (queue.isEmpty()) {
         throw new SimulationException("Get Message from empty queue");
      }
      return queue.get(0);
   }

   @Override
   protected QueueEntry pop() {
      if (queue.isEmpty()) {
         throw new SimulationException(simNode.getFullName() + "Get Message from empty queue");
      }
      QueueEntry qe = queue.remove(0);
      if (getCurrentNumberOfUnits() < 0) {
         throw new SimulationException(simNode.getFullName() + "Actual number of units in the queue is less than 0");
      }
      return qe;
   }
}
