/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

import java.util.LinkedList;
import java.util.TreeMap;
/**
 * This class implements a fair queuing on the basis of Virtual Finish Times, 
 * and therefore achieves fairness regarding the number of units (e.g. Bytes)
 * transmitted.  
 * @author dwagner
 *
 */
public class UnitFairQueuing extends FairQueuing {
   class FinishTimeStackableQueuingDiscipline extends SortableQDiscContainer {
      double lastFinishTime, finishTimeNext;
      QueueEntry headOfQDisc = null;

      public FinishTimeStackableQueuingDiscipline(StackableQueuingDiscipline q) {
         qdisc = q;
      }

      @Override
      public double getsortIndex() {
         return finishTimeNext;
      }

      public void updateSortIndex(double cvt) {
         if (!qdisc.isEmpty() && qdisc.peek() != headOfQDisc) {
            // head changed, need to update sortIndex = virtual finish time of
            // next packet
            lastFinishTime = finishTimeNext;
            finishTimeNext = Math.max(lastFinishTime, cvt) + qdisc.peek().getMessage().getLength();
            headOfQDisc = qdisc.peek();
         }
      }

      @Override
      public long getLengthInMetric() {
         return qdisc.getCurrentNumberOfUnits();
      }
   }

   /** as required by {@link ReflectionConstructable} */
   public static UnitFairQueuing createInstance(SimNode ownNode, Parameters pars) {
      return new UnitFairQueuing(ownNode, pars);
   }

   protected TreeMap<Double, LinkedList<FinishTimeStackableQueuingDiscipline>> activeQs;
   protected double currentVirtualTime;

   public UnitFairQueuing(SimNode ownNode, Parameters pars) {
      super(ownNode, pars);
      currentVirtualTime = 0;
      activeQs = new TreeMap<Double, LinkedList<FinishTimeStackableQueuingDiscipline>>();
   }

   /**
    * This is just a simplified option to implement the dropping estimation. It
    * just removes one message from all queues longer than the queue the message
    * would be inserted in but in fact, it might be necessary to remove several
    * messages from one of these queues.
    * 
    * When overriding it, keep in mind that with a stochastic classifier packets
    * at the head of the queue might belong to a completely different flow
    */
   @Override
   public boolean wouldStore(Message msg) {
      if (getCurrentNumberOfUnits() + msg.getLength() > getMaxSizeInUnits()) {
         long tooMuch = getCurrentNumberOfUnits() + msg.getLength() - getMaxSizeInUnits();
         if (queues.getConsumer(msg) != null) {
            for (long index : longestQList.tailMap(queues.getConsumer(msg).getQDisc().getCurrentNumberOfUnits(), false).descendingKeySet()) {
               for (SortableQDiscContainer sq : longestQList.get(index)) {
                  tooMuch -= sq.getQDisc().peek().getMessage().getLength();
                  if (tooMuch <= 0)
                     return true;
               }
            }
         } else {
            return true;
         }
         return false;
      } else {
         return true;
      }
   }

   @Override
   public SortableQDiscContainer createSortableQDisc(StackableQueuingDiscipline q) {
      return new FinishTimeStackableQueuingDiscipline(q);
   }

   @Override
   protected boolean activeQListIsEmpty() {
      return activeQs.isEmpty();
   }

   @Override
   protected void addToActiveQList(SortableQDiscContainer sq) {
      FinishTimeStackableQueuingDiscipline fsq = (FinishTimeStackableQueuingDiscipline) sq;
      double pos = fsq.getsortIndex();
      if (activeQs.containsKey(pos)) {
         activeQs.get(pos).add(fsq);
      } else {
         LinkedList<FinishTimeStackableQueuingDiscipline> newList = new LinkedList<FinishTimeStackableQueuingDiscipline>();
         newList.add(fsq);
         activeQs.put(pos, newList);
      }
      if (activeQ == null) {
         assert activeQs.size() == 1;
         activeQ = activeQs.firstEntry().getValue().getFirst();
         currentVirtualTime = activeQ.getsortIndex();
      }
   }

   @Override
   protected void updateActiveQList(SortableQDiscContainer sq, double oldPos) {
      // update
      FinishTimeStackableQueuingDiscipline fsq = (FinishTimeStackableQueuingDiscipline) sq;
      fsq.updateSortIndex(currentVirtualTime);
      // remove
      double newPos = fsq.getsortIndex();
      assert activeQs.containsKey(oldPos);
      // remove from old place in sortedQList
      activeQs.get(oldPos).remove(sq);
      if (activeQs.get(oldPos).size() == 0)
         activeQs.remove(oldPos);
      // if activeQs.size() == 0 advanceActiveQ will set activeQ = null
      // insert
      if (sq.getQDisc().getCurrentNumberOfMessages() != 0)
         addToActiveQList(sq);
   }

   @Override
   protected void advanceActiveQ() {
      if (activeQs.size() > 0) {
         activeQ = activeQs.firstEntry().getValue().getFirst();
         assert currentVirtualTime <= activeQ.getsortIndex();
         currentVirtualTime = activeQ.getsortIndex();
      } else {
         activeQ = null;
      }
   }
}