/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control;

import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.Model;
import ikr.simlib.model.SimNodeManager;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.exceptions.ImportException;

/**
 * <code>ImportSimulation</code> does not actually run the simulation itself.
 * Instead, it goes through the batches and collects batch results from previous
 * simulation runs and produces the final results.
 */
public class ImportSimulation extends StdSimulation {

   public ImportSimulation(Model model, SimulationEnvironment simEnv) {
      super(model, simEnv);
   }

   @Override
   public void runSimulation() {
      if (model == null) {
         throw new SimulationException("no model");
      }

      // skip the transient phase
      SimulationControlManager.getInstance().startTransientPhase();
      SimulationControlManager.getInstance().stopTransientPhase();

      // import the batches
      for (int batchNumber = 1; batchNumber <= simEnv.getNumberOfBatches(); batchNumber++) {

         // check if an input file is available
         if (simEnv.getImportReaders().size() <= batchNumber - 1) {
            throw new SimulationException("Not enough import log files (batch number " + batchNumber + ").");
         }

         // get input file
         ResultReader reader = simEnv.getImportReaders().get(batchNumber - 1);

         // read input file
         try {
            reader.readStartResults();
            SimNodeManager.getInstance().doImportBatch(reader);
            reader.readEndResults();
            reader.close();
         } catch (ImportException e) {
            throw new RuntimeException("error while importing result file no. " + batchNumber, e);
         }

         // print results
         printBatchResults(batchNumber);
      }

      // end simulation
      SimulationControlManager.getInstance().stopSimulation();
   }
}
