/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control.notification;

import java.util.LinkedList;
import java.util.List;

/**
 * MultiNotificationHandler manages any number of Notifier objects. Notifier
 * objects must be registered with the handler by calling the AddNotifier
 * method. <br>
 * After all Notifier objects have called their Notify method, the
 * HandleNotification method of the handler is called. After that
 * ResetNotification is called and the process starts all over. <br>
 * Notification handlers are used to control simulation phases. They may be used
 * for more general applications as well.
 */
public class MultiNotificationHandler {

   /**
    * Boolean operation for SimNofificationHandlers
    */
   public enum BooleanOperation {
      AND, OR
   }

   /**
    * all members
    */
   protected List<Notifier> notifiers;

   /**
    * members that haven't called Notify yet
    */
   protected List<Notifier> needToNotifyMembers;

   /**
    * determines condition for calling HandleNotification
    */
   protected BooleanOperation notificationCondition;

   /**
    * flag that allows changes of notification condition only once
    */
   protected boolean isOperationSet;

   protected NotificationHandler handler;

   public MultiNotificationHandler(NotificationHandler handler) {
      notifiers = new LinkedList<Notifier>();
      needToNotifyMembers = new LinkedList<Notifier>();

      this.handler = handler;

      notificationCondition = BooleanOperation.AND;
      isOperationSet = false;

      resetNotification();
   }

   public void notify(Notifier from) {
      switch (notificationCondition) {
      case AND:
         needToNotifyMembers.remove(from);
         if (needToNotifyMembers.isEmpty() && !notifiers.isEmpty()) {
            handler.handleNotification();
            resetNotification();
         }
         break;
      case OR:
         if (!notifiers.isEmpty()) {
            handler.handleNotification();
            resetNotification();
         }
         break;
      }
   }

   public void resetNotification() {
      needToNotifyMembers.clear();
      needToNotifyMembers.addAll(notifiers);
   }

   public void addNotifier(Notifier n) {

      if (n.getOwner() != null) {
         throw new NotificationException("Notifier already owned by other NotificationHandler");
      }

      n.setOwner(this);

      notifiers.add(n);
      needToNotifyMembers.add(n);
   }

   public void removeNotifier(Notifier n) {
      if (!notifiers.remove(n))
         throw new NotificationException("Notifier not handled by this handler");

      if (needToNotifyMembers.size() == 1 && notificationCondition == BooleanOperation.AND && needToNotifyMembers.contains(n))
         notify(n); // notify after removing it

      n.setOwner(null);
   }

   /**
    * Allows to change the necessary condition for calling HandleNotification(),
    * do not use this method after start of simulation.
    * 
    * @param op
    */
   public void setBooleanOperation(BooleanOperation op) {
      if (isOperationSet) {
         throw new NotificationException("Boolean operation already set, no further changes allowed");
      }

      isOperationSet = true;
      notificationCondition = op;
   }
}
