/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions;

import java.util.List;

/**
 * This class contains an array-like structure which holds two vectors of same
 * length. <code>fValVector</code> holds parameter values (class border) and
 * <code>fDistVector</code> holds the corresponding probabilities of the
 * cumulative distribution function. The user should enter items as pairs of
 * values and cumulative distribution densities (into a single vector which is
 * divided later). <em>The class width may differ!</em>
 * 
 * <p>
 * <code>next</code> is supposed to return a continuous value to a given
 * empirical distribution (using linear interpolation as in piecewise linear
 * distribution).
 * </p>
 */
public class DistributionVector {

   public final double[] valVector;
   public final double[] distVector;

   public DistributionVector(List<Double> vector) {
      final int N = vector.size();

      if (N % 2 != 0) {
         throw new DistributionException("vector must hold equal number of values and CDF-probs");
      }

      for (int i = 1; i < N - 2; i += 2) {
         if ((vector.get(i) < 0.0) || (vector.get(i) > 1.0)) {
            throw new DistributionException("expecting probabilities >= 0.0 and <=1");
         }

         if (vector.get(i) > vector.get(i + 2)) {
            throw new DistributionException("expecting a CDF (cumulative distribution function) with sorted vals/probs");
         }
      }

      if ((vector.get(N - 1) < 0) || (vector.get(N - 1) > 1.0)) {
         throw new DistributionException("expecting probabilities >= 0.0 and <=1");
      }

      if (vector.get(1) != 0.0) {
         throw new DistributionException("expecting probabilities of 0.0 for first value.");
      }

      valVector = new double[N / 2];
      distVector = new double[N / 2];

      for (int i = 0; i < N / 2; i += 1) {
         valVector[i] = vector.get(i * 2);
         distVector[i] = vector.get(i * 2 + 1);
      }
   }

   public DistributionVector(double[] vector) {
      if (vector.length % 2 != 0) {
         throw new DistributionException("vector must hold equal number of values and CDF-probs");
      }

      for (int i = 1; i < vector.length - 2; i += 2) {
         if ((vector[i] < 0.0) || (vector[i] > 1.0)) {
            throw new DistributionException("expecting probabilities >= 0.0 and <=1");
         }

         if (vector[i] > vector[i + 2]) {
            throw new DistributionException("expecting a CDF (cumulative distribution function) with sorted vals/probs");
         }
      }

      if ((vector[vector.length - 1] < 0) || (vector[vector.length - 1] > 1.0)) {
         throw new DistributionException("expecting probabilities >= 0.0 and <=1");
      }

      if (vector[1] != 0.0) {
         throw new DistributionException("expecting probabilities of 0.0 for first value.");
      }

      valVector = new double[vector.length / 2];
      distVector = new double[vector.length / 2];

      for (int i = 0; i < vector.length / 2; i += 1) {
         valVector[i] = vector[i * 2];
         distVector[i] = vector[i * 2 + 1];
      }
   }

   public DistributionVector(DistributionVector rhs) {
      final int N = rhs.valVector.length;

      valVector = new double[N];
      distVector = new double[N];

      System.arraycopy(rhs.valVector, 0, valVector, 0, N);
      System.arraycopy(rhs.distVector, 0, distVector, 0, N);
   }

   public double next(double searchValue) {
      if ((searchValue >= 1) || (searchValue < 0)) {
         throw new DistributionException("search value must be <1 and >=0");
      }

      int a = 0;
      int b = distVector.length - 1;
      int k;
      double rv;
      double lowprob, highprob, classwidth;

      while (a < b) {
         k = (a + b) / 2;
         if (searchValue <= distVector[k]) {
            b = k;
         } else {
            a = k + 1;
         }
      }

      /**
       * The given probability points to a value betweeN fvalVector[b-1] and
       * fvalVector[b]. make a linear interpolation:
       */
      highprob = distVector[b];
      if (b == 0) {
         lowprob = 0.0;
         classwidth = valVector[b];
      } else {
         lowprob = distVector[b - 1];
         classwidth = valVector[b] - valVector[b - 1];
      }

      if (highprob == lowprob) {
         rv = valVector[b];
      } else {
         rv = valVector[b] - classwidth * ((highprob - searchValue) / (highprob - lowprob));
      }

      return rv;
   }
}
