/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions;

import java.util.List;

/**
 * A discrete probability vector. The user should enter items as probability
 * density and may transform them into a distribution.
 */
public class ProbabilityVector {

   private final double[] cumulatedVector;

   public ProbabilityVector(List<Double> vector) {
      cumulatedVector = new double[vector.size()];

      double sum = 0;
      final double cEpsilon = 1e-10;

      for (int i = 0; i < vector.size(); i++) {
         if ((vector.get(i) < 0) || (vector.get(i) > 1)) {
            throw new DistributionException("Probability must be >= 0 and <= 1");
         }

         sum += vector.get(i);
         cumulatedVector[i] = sum;
      }

      if (Math.abs(sum - 1) > cEpsilon) {
         throw new DistributionException("sum must be = 1");
      }
   }

   public ProbabilityVector(double[] vector) {
      cumulatedVector = new double[vector.length];

      double sum = 0;
      final double cEpsilon = 1e-10;

      for (int i = 0; i < vector.length; i++) {
         if ((vector[i] < 0) || (vector[i] > 1)) {
            throw new DistributionException("Probability must be >= 0 and <= 1");
         }

         sum += vector[i];
         cumulatedVector[i] = sum;
      }

      if (Math.abs(sum - 1) > cEpsilon) {
         throw new DistributionException("sum must be = 1");
      }
   }

   public ProbabilityVector(ProbabilityVector rhs) {
      cumulatedVector = new double[rhs.cumulatedVector.length];

      System.arraycopy(rhs.cumulatedVector, 0, cumulatedVector, 0, rhs.cumulatedVector.length);
   }

   /**
    * Returns a discrete Value to a given probability.
    */
   public int next(double searchValue) {
      if ((searchValue >= 1) || (searchValue < 0)) {
         throw new DistributionException("search value must be <1 and >=0");
      }

      int a = 0;
      int b = cumulatedVector.length - 1;
      int k;

      while (a < b) {
         k = (a + b) / 2;
         if (searchValue <= cumulatedVector[k]) {
            b = k;
         } else {
            a = k + 1;
         }
      }
      return b;
   }
}
