/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Beta Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>
 * <ul>
 * <li>
 * Distribution of a random variable \(T= \frac{Z_1}{Z_1+Z_2}\), if \(Z_1\) and
 * \(Z_2\) are gamma distributed with the parameters \(\alpha_1\) and \(\beta\)
 * as well as \(\alpha_2\) and \(\beta\).
 * <li>
 * suitable for the distribution of random shares (values between 0 and 1)
 * <li>
 * Special cases contained are the uniform distribution between 0 and 1 as well
 * as several triangular distributions
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * shape parameter \(\alpha_1 > 0\)
 * <li>
 * shape parameter \(\alpha_2 > 0\)
 * </ul>
 * The distribution can also be described with parameters for the mean value (in
 * the range between 0 and 1) and coefficient of variation:
 * <ul>
 * <li>
 * \(\alpha_1 = \frac{1-E[T] \cdot (1+c_T^2)}{c_T^2}\)
 * <li>
 * \(\alpha_2 = \alpha_1 \cdot (\frac{1}{E[T]} -1)\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \frac{t^{\alpha_1 -1} \cdot (1-t)^{\alpha_2
 * -1}}{B(\alpha_1 , \alpha_2)} \mbox{ for } 0 < x < 1 \), <br>
 * <br>
 * whereby \(B(x,y)\) is the Beta function</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>No closed form of the distribution function (except for special cases)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \frac{\alpha_1}{\alpha_1 + \alpha_2}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \frac{\alpha_1 \cdot \alpha_2}{(\alpha_1 + \alpha_2)^2 \cdot
 * (\alpha_1 + \alpha_2 +1)}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \sqrt{\frac{\alpha_1}{\alpha_2 \cdot (\alpha_1 + \alpha_2 +1)}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.BetaDistribution
         <br>
         [...].Distribution.Alpha1 = 1.5
         <br>
         [...].Distribution.Alpha2 = 3.0
      </code> <br>
 * <br>
 * or with the mean value and variation coefficient: <br>
 * <br>
 * <code>
         [...].Distribution.Mean = 0.33
         <br>
         [...].Distribution.CoefficientOfVariation = 0.60
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class BetaDistribution extends ContinuousDistribution {
   private static final String ALPHA1 = "Alpha1";
   private static final String ALPHA2 = "Alpha2";
   private static final String MEAN = "Mean";
   private static final String COEFFICIENT_OF_VARIATION = "CoefficientOfVariation";

   /** as required by {@link ReflectionConstructable} */
   public static BetaDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double alpha1, alpha2;

      if (pars.hasParameter(ownNode, ALPHA1) && pars.hasParameter(ownNode, ALPHA2)) {
         alpha1 = pars.get(ownNode, ALPHA1).asDouble();
         alpha2 = pars.get(ownNode, ALPHA2).asDouble();
      } else if (pars.hasParameter(ownNode, MEAN) && pars.hasParameter(ownNode, COEFFICIENT_OF_VARIATION)) {
         double mean = pars.get(ownNode, MEAN).asDouble();
         double coefficientOfVariation = pars.get(ownNode, COEFFICIENT_OF_VARIATION).asDouble();
         double coVSquare = coefficientOfVariation * coefficientOfVariation;
         alpha1 = (1 - mean * (1 + coVSquare)) / coVSquare;
         alpha2 = alpha1 * (1 / mean - 1);
      } else {
         throw new SimulationException("not enough parameters");
      }
      return new BetaDistribution(alpha1, alpha2, rng);
   }

   public final GammaDistribution gamma1;
   public final GammaDistribution gamma2;

   public BetaDistribution(double alpha1, double alpha2, RandomNumberGenerator rng) {
      super(rng);
      gamma1 = new GammaDistribution(alpha1, 1, rng);
      gamma2 = new GammaDistribution(alpha2, 1, rng);
   }

   public BetaDistribution(double alpha1, double alpha2) {
      this(alpha1, alpha2, getDefaultRNG());
   }

   @Override
   public double next() {
      final double y1 = gamma1.next();
      final double y2 = gamma2.next();

      return y1 / (y1 + y2);
   }
}
