/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Coxian Phase Model <br>
 * 
 * <center>
 * <img src="{@docRoot}/ikr/simlib/doc-files/coxianPhaseModelP28.svg" width="600" alt="Coxian Phase Model"> <br>
 * Coxian Phase Model </center>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Distribution according to the Coxian phase model: Serial switching of a
 * selection of one of \(k\) phases each with a negative-exponentially
 * distributed phase duration period (parameter \(\lambda_i\)), whereby after
 * each phase the system is exited with the probability \(q_i\). Both the
 * hyperexponential as well as the hypoexponential distributions are contained
 * within this model.
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * order \(k>0\)
 * <li>
 * rates \(\lambda_i\) or mean values \(m_i\) of the individual phases \((i = 1,
 * \dots,k)\)
 * <li>
 * exit probabilities \(q_i (i = 0, ... , k-1)\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = q_0 + \sum\limits_{i=1}^k ( \prod\limits_{v=0}^{i-1} (1-q_v)
 * ) \cdot q_i \cdot \prod\limits_{j=1}^i \frac{\lambda_j}{\lambda_j + s} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.CoxianDistribution
         <br>
         [...].Distribution.Order = 2
         <br>
         [...].Distribution.Means = [ 2.5 8 ]
         <br>
         [...].Distribution.QuitProbabilities = [ 0.2 0.5 ]
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class CoxianDistribution extends ContinuousDistribution {

   private static final String MEAN = "Mean";
   private static final String QUIT_PROBS = "QuitProbabilitites";

   /** as required by {@link ReflectionConstructable} */
   public static CoxianDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      double[] means = pars.get(ownNode, MEAN).asDoubleVector();
      double[] quitProbs = pars.get(ownNode, QUIT_PROBS).asDoubleVector();
      assert means.length == quitProbs.length;
      return new CoxianDistribution(means, quitProbs, rng);
   }

   public final NegExpDistribution[] phases;
   public final double[] quitProbs;

   public CoxianDistribution(double[] means, double[] quitProbs, RandomNumberGenerator rng) {
      super(rng);

      if (means.length != quitProbs.length) {
         throw new DistributionException("number of means is different from number of quit probs");
      }

      final int N = means.length;

      phases = new NegExpDistribution[N];
      this.quitProbs = new double[N];

      for (int i = 0; i < N; i++) {
         if (quitProbs[i] < 0 || quitProbs[i] > 1) {
            throw new DistributionException("quit prob. must be >= 0 and <= 1");
         }

         this.quitProbs[i] = quitProbs[i];
         phases[i] = new NegExpDistribution(means[i]);
      }
   }

   public CoxianDistribution(double[] means, double[] quitProbs) {
      this(means, quitProbs, getDefaultRNG());
   }

   @Override
   public double next() {
      double sum = 0;

      for (int i = 0; i < phases.length; i++) {
         if (rng.next() <= quitProbs[i]) {
            break;
         } else {
            sum += phases[i].next();
         }
      }

      return sum;
   }
}
