/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

import java.util.logging.Logger;

/**
 * General Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>A distribution with a certain mean value \(\mu\)and coefficient of
 * variation \(c\) is generated by linking two phases (phase model)
 * <ul>
 * <li>
 * case I \((c=0)\): <br>
 * Only one phase with a constant distribution (mean value \(\mu\)).
 * <li>
 * case II \((0 < c < 1)\): <br>
 * Serial switching of a constant distribution with the mean value \(m_1 = \mu
 * \cdot c\) and a negative-exponential distribution with the mean value \(m_2 =
 * \mu \cdot (1-c)\)
 * <li>
 * case III \((c=1)\): <br>
 * Only one phase with a negative-exponential distribution (mean value \(\mu\)).
 * <li>
 * case IV \((c > 1)\): <br>
 * Parallel switching of two phases with negative-exponential distributions and
 * the parameters <br>
 * \(m_{1/2} = \frac{\mu}{1 \pm \sqrt{ \frac{c^2 -1}{c^2 +1} }}\) (mean value)
 * and <br>
 * \(p_{1/2} = \frac{1}{2} \cdot (1 \pm \sqrt{ \frac{c^2 -1}{c^2 +1} }\)
 * (branching probabilities). <br>
 * <br>
 * This corresponds to a hyperexponential distribution of 2nd order, its
 * parameters fulfill the symmetry condition \(p_1 \cdot m_1 = p_2 \cdot m_2\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * mean value \(\mu > 0\)
 * <li>
 * coefficient of variation \(c \ge 0\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>
 * <ul>
 * <li>
 * case I: \(P(T=t) = f(t) = \delta(t - \mu) \)
 * <li>
 * case II: \(P(T=t) = f(t) = \frac{1}{m_2} \cdot exp( -\frac{t-m_1}{m_2} )\)
 * <li>
 * case III: \(P(T=t) = f(t) = \frac{1}{\mu} \cdot exp( -\frac{t}{\mu} )\)
 * <li>
 * case IV: \(P(T=t) = f(t) = \frac{p_1}{m_1} \cdot exp( -\frac{t}{m_1} ) +
 * \frac{p_2}{m_2} \cdot exp( -\frac{t}{m2} )\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \mu \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= (c \cdot \mu)^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>
 * <ul>
 * <li>
 * case I: \(\phi(s) = exp( -\mu s )\)
 * <li>
 * case II: \(\phi(s) = \frac{exp(-m_1 s)}{1+m_2 \cdot s}\)
 * <li>
 * case III: \(\phi(s) = \frac{1}{1+\mu \cdot s}\)
 * <li>
 * case IV: \(\phi(s) = \frac{p_1}{1+m_1 \cdot s} + \frac{p_2}{1+m_2 \cdot s}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.GeneralDistribution
         <br>
         [...].Distribution.Mean = 2.3
         <br>
         [...].Distribution.CoefficientOfVariation = 1.5
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class GeneralDistribution extends ContinuousDistribution {

   private static final String MEAN = "Mean";
   private static final String COEFF_OF_VARIATION = "CoefficientOfVariation";

   /** as required by {@link ReflectionConstructable} */
   public static GeneralDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      final double coefficientOfVariation = pars.get(ownNode, COEFF_OF_VARIATION).asDouble();
      return new GeneralDistribution(mean, coefficientOfVariation, rng);
   }

   public static enum SubType {
      SINGLE, SUM, ALTERNATIVE
   }

   public final ContinuousDistribution distribution1;
   public final ContinuousDistribution distribution2;
   public final SubType subType;
   public final double branchingProbability;

   public GeneralDistribution(double mean, double coefficientOfVariation, RandomNumberGenerator rng) {
      super(rng);

      final double max_cv = 100;

      if (coefficientOfVariation < 0) {
         throw new DistributionException("Coefficient of variation < 0!");
      }

      if (mean < 0) {
         throw new DistributionException("Mean < 0!");
      }

      if (coefficientOfVariation > max_cv) {
         Logger log = Logger.getLogger("ikr.simlib.distributions");

         log.warning("coefficient of variation critically large!");
      }

      if (coefficientOfVariation == 0) {
         subType = SubType.SINGLE;
         distribution1 = new ConstantDistribution(mean);
         distribution2 = null;
         branchingProbability = 0;
      } else if ((coefficientOfVariation > 0) && (coefficientOfVariation < 1)) {
         subType = SubType.SUM;
         distribution1 = new NegExpDistribution(mean * coefficientOfVariation);
         distribution2 = new ConstantDistribution(mean * (1 - coefficientOfVariation));
         branchingProbability = 0;
      } else if (coefficientOfVariation == 1) {
         subType = SubType.SINGLE;
         distribution1 = new NegExpDistribution(mean);
         distribution2 = null;
         branchingProbability = 0;
      } else {
         double cv_squared, auxterm;

         subType = SubType.ALTERNATIVE;
         cv_squared = coefficientOfVariation * coefficientOfVariation;
         auxterm = Math.sqrt((cv_squared - 1) / (cv_squared + 1));
         distribution1 = new NegExpDistribution(mean / (1 + auxterm));
         distribution2 = new NegExpDistribution(mean / (1 - auxterm));
         branchingProbability = (1 + auxterm) / 2;
      }
   }

   public GeneralDistribution(double mean, double coefficientOfVariation) {
      this(mean, coefficientOfVariation, getDefaultRNG());
   }

   @Override
   public double next() {
      double result = 0;

      if (subType == SubType.SINGLE) {
         result = distribution1.next();
      } else if (subType == SubType.SUM) {
         result = distribution1.next() + distribution2.next();

      } else if (subType == SubType.ALTERNATIVE) {
         if (rng.next() < branchingProbability) {
            result = distribution1.next();
         } else {
            result = distribution2.next();
         }
      }

      return result;
   }
}
