/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.distributions.ProbabilityVector;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Hyperexponential Distribution to the Order of k <br>
 * 
 * <center>
 * <img src="{@docRoot}/ikr/simlib/doc-files/hyperExpP27.svg" width="600" alt="Hyperexponential distribution to the order of k"> <br>
 * Hyperexponential distribution to the order of k </center>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Selecting one of \(k\) random variables that are negative-exponentially
 * distributed with the individual parameters \(\lambda_i\) and the
 * probabilities \(p_i\) (parallel switching in the phase model).
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * order \(k > 0\)
 * <li>
 * rates \(\lambda_i\) or mean values \(m_i\) of the individual phases \((i = 1,
 * \dots, k)\)
 * <li>
 * branching probabilities \(p_i(i=1,...,k)\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \sum\limits_{i=1}^k \lambda_i \cdot p_i \cdot
 * exp(-\lambda_i t) \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = 1 - \sum\limits_{i=1}^k p_i \cdot exp(-\lambda_i t)
 * \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \sum\limits_{i=1}^k \frac{p_i}{\lambda_i} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= 2 \cdot \sum\limits_{i=1}^k \frac{p_i}{\lambda_i^2} - (
 * \sum\limits_{i=1}^k \frac{p_i}{\lambda_i} )^2 \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \sqrt{\frac{2 \cdot \sum\limits_{i=1}^k
 * \frac{p_i}{\lambda_i^2}}{(\sum\limits_{i=1}^k \frac{p_i}{\lambda_i})^2}} -1
 * \le 1\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = \sum\limits_{i=1}^{k} p_i \cdot \frac{\lambda_i}{\lambda_i
 * +s}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.HyperExpDistribution
         <br>
         [...].Distribution.Order = 2
         <br>
         [...].Distribution.Means = [ 2.5 , 8]
         <br>
         [...].Distribution.BranchProbabilities = [ 0.4 0.6 ]
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class HyperExpDistribution extends ContinuousDistribution {

   private static final String MEANS = "Means";
   private static final String BRANCH_PROBS = "BranchProbabilities";

   /** as required by {@link ReflectionConstructable} */
   public static HyperExpDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double[] means = pars.get(ownNode, MEANS).asDoubleVector();
      final double[] branchProbs = pars.get(ownNode, BRANCH_PROBS).asDoubleVector();
      return new HyperExpDistribution(means, branchProbs, rng);
   }

   public final NegExpDistribution[] phases;
   public final ProbabilityVector branchProbs;

   public HyperExpDistribution(double[] means, double[] branchProbs, RandomNumberGenerator rng) {
      super(rng);

      if (means.length != branchProbs.length) {
         throw new DistributionException("number of means is different from number of branch probs");
      }

      phases = new NegExpDistribution[means.length];
      for (int i = 0; i < means.length; i++) {
         phases[i] = new NegExpDistribution(means[i]);
      }

      this.branchProbs = new ProbabilityVector(branchProbs);
   }

   public HyperExpDistribution(double[] means, double[] branchProbs) {
      this(means, branchProbs, getDefaultRNG());
   }

   public HyperExpDistribution(double mean1, double mean2, double branchProb1, RandomNumberGenerator rng) {
      super(rng);

      phases = new NegExpDistribution[2];
      phases[0] = new NegExpDistribution(mean1);
      phases[1] = new NegExpDistribution(mean2);

      this.branchProbs = new ProbabilityVector(new double[] { branchProb1, 1 - branchProb1 });
   }

   public HyperExpDistribution(double mean1, double mean2, double branchProb1) {
      this(mean1, mean2, branchProb1, getDefaultRNG());
   }

   @Override
   public double next() {
      int phaseIndex = branchProbs.next(rng.next());

      return phases[phaseIndex].next();
   }
}
