/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Hypoexponential Distribution to the Order of k <br>
 * 
 * <center><img src="{@docRoot}/ikr/simlib/doc-files/hypoExpP26.svg" width="600" alt="Hypoexponential distribution to the order of k"> <br>
 * Hypoexponential distribution to the order of k </center>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Generalization of the Erlang k distribution (serial switching of \(k\)
 * phases with negative-exponentially distributed durations with individual
 * parameters \(\lambda_i\) in the phase model).
 * </td>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * order \(k > 0\)
 * <li>
 * rates \(\lambda_i\) or mean values \(m_i\) of the individual phases \((i = 1,
 * \dots, k)\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = g_1(t) \otimes ... \otimes g_k(t)\) with \(g_i(t) =
 * \lambda_i \cdot exp(-\lambda_i t)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \sum\limits_{i=1}^k \frac{1}{\lambda_i} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \sum\limits_{i=1}^k \frac{1}{\lambda_i^2} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \frac{\sqrt{\sum\limits_{i=1}^k
 * \frac{1}{\lambda_i^2}}}{\sum\limits_{i=1}^k \frac{1}{\lambda_i}} \le 1\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = \prod\limits_{i=1}^{k} \frac{\lambda_i}{\lambda_i +s}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.HypoExpDistribution
         <br>
         [...].Distribution.Order = 2
         <br>
         [...].Distribution.Means = [ 2.5 , 8]
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class HypoExpDistribution extends ContinuousDistribution {

   public final NegExpDistribution[] phases;

   private static final String MEANS = "Means";

   /** as required by {@link ReflectionConstructable} */
   public static HypoExpDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double[] means = pars.get(ownNode, MEANS).asDoubleVector();
      return new HypoExpDistribution(means, rng);
   }

   public HypoExpDistribution(int nrOfPhases, RandomNumberGenerator rng) {
      super(rng);
      phases = new NegExpDistribution[nrOfPhases];
   }

   public HypoExpDistribution(int nrOfPhases) {
      this(nrOfPhases, getDefaultRNG());
   }

   public HypoExpDistribution(double[] means, RandomNumberGenerator rng) {
      this(means.length, rng);

      for (int i = 0; i < means.length; i++) {
         phases[i] = new NegExpDistribution(means[i]);
      }
   }

   public HypoExpDistribution(double[] means) {
      this(means, getDefaultRNG());
   }

   /**
    * 2nd order
    */
   public HypoExpDistribution(double mean1, double mean2, RandomNumberGenerator rng) {
      this(2, rng);

      phases[0] = new NegExpDistribution(mean1);
      phases[1] = new NegExpDistribution(mean2);
   }

   public HypoExpDistribution(double mean1, double mean2) {
      this(mean1, mean2, getDefaultRNG());
   }

   @Override
   public double next() {
      double sum = 0;

      for (NegExpDistribution p : phases) {
         sum += p.next();
      }

      return sum;
   }
}
