/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Lognormal Distribution <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>
 * <ul>
 * <li>
 * Distribution of \(T = exp(Z)\), if \(Z\) is normally distributed with the
 * parameters \(\mu\) and \(\sigma^2\).
 * <li>
 * Limit distribution of the product of many independent random variables with
 * arbitrary statistical characteristics if the contribution of a single random
 * variable remains very small.
 * <li>
 * In its form similar to a gamma or Weibull distribution-
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * \(\mu\) (mean value of \(Z\))
 * <li>
 * \(\sigma > 0\) (standart deviation of \(Z\))
 * </ul>
 * Alternative parameters mean value and coefficient of variation:
 * <ul>
 * <li>
 * \(\mu = ln(\frac{E[T]}{\sqrt{1+c_T^2}})\)
 * <li>
 * \(\sigma = \sqrt{ln(1+c_T^2)}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \frac{1}{\sqrt{2\pi} \sigma \cdot t} \cdot exp(
 * -\frac{(ln t -\mu)^2}{2 \sigma^2} ) \mbox{ for } t>0 \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>No closed form</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= exp(\mu + \frac{\sigma^2}{2})\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= exp(2 \mu + \sigma^2) \cdot (exp(\sigma^2) -1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \sqrt{exp(\sigma^2) -1}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.LognormalDistribution
         <br>
         [...].Distribution.My = 0.11
         <br>
         [...].Distribution.Sigma = 1.27
      </code> <br>
 * <br>
 * or with the mean value and variation coefficient: <br>
 * <br>
 * <code>
         [...].Distribution.Mean = 2.5
         <br>
         [...].Distribution.CoefficientOfVariation = 2.0
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class LognormalDistribution extends ContinuousDistribution {

   public final NormalDistribution baseDist;
   public final double my;
   public final double sigmaSquare;

   private static final String MY = "My";
   private static final String SIGMA = "Sigma";
   private static final String MEAN = "Mean";
   private static final String COEFFICIENT_OF_VARIATION = "CoefficientOfVariation";

   /** as required by {@link ReflectionConstructable} */
   public static LognormalDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      double my = Double.NaN;
      double sigma = Double.NaN;

      // get sigma
      if (pars.hasParameter(ownNode, SIGMA)) {
         sigma = pars.get(ownNode, SIGMA).asDouble();

      } else {
         double coefficientOfVariation = pars.get(ownNode, COEFFICIENT_OF_VARIATION).asDouble();
         sigma = Math.sqrt(Math.log(1 + coefficientOfVariation * coefficientOfVariation));
      }

      // check sigma
      assert sigma > 0;

      // get my
      if (pars.hasParameter(ownNode, MY)) {
         my = pars.get(ownNode, MY).asDouble();

      } else {
         double mean = pars.get(ownNode, MEAN).asDouble();
         my = Math.log(mean) - sigma * sigma / 2.0;
      }

      return new LognormalDistribution(my, sigma, rng);
   }

   public LognormalDistribution(double my, double sigma, RandomNumberGenerator rng) {
      super(rng);

      if (sigma <= 0) {
         throw new DistributionException("sigma is <= 0!");
      }

      baseDist = new NormalDistribution(my, sigma, rng);
      this.my = my;
      sigmaSquare = sigma * sigma;
   }

   public LognormalDistribution(double my, double sigma) {
      this(my, sigma, getDefaultRNG());
   }

   @Override
   public double next() {
      return Math.exp(baseDist.next());
   }
}
