/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Pareto Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>- Like the Weibull distribution, the Pareto distribution is often used to
 * characterize Internet traffic because of its heavy tail. The superposition of
 * on-off sources with heavy-tailed phases is known to produce self-similar
 * traffic.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * minimum value \(k > 0\)
 * <li>
 * shape parameter \(\alpha > 0\)
 * </ul>
 * In case \(\alpha > 2\) this distribution can also be described with
 * parameters mean value \(\mu\) and coefficient of variation \(c\):
 * <ul>
 * <li>
 * \(\alpha = 1+ \frac{\sqrt{1+c^2}}{c}\)
 * <li>
 * \(k = \frac{1}{ \frac{c}{\sqrt{1+c^2}} +1 } \cdot \mu\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \frac{\alpha \cdot k^{\alpha}}{t^{\alpha +1}} \mbox{
 * for } t \ge k \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = 1 - (\frac{k}{t})^{\alpha} \mbox{ for } t \ge k \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \begin{cases} \frac{\alpha}{\alpha -1} \cdot k &\mbox{ for }
 * \alpha > 1 \\ \infty &\mbox{ for } \alpha \le 1 \end{cases} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \begin{cases} (\frac{\alpha}{\alpha -2} - (\frac{\alpha}{\alpha
 * -1})^2) \cdot k^2 &\mbox{ for } \alpha > 2 \\ \infty &\mbox{ for } \alpha \le
 * 2 \end{cases}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \frac{1}{\sqrt{(\alpha -2) \cdot \alpha}} \mbox{ for } \alpha >
 * 2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.ParetoDistribution
         <br>
         [...].Distribution.Alpha1 = 2.2
         <br>
         [...].Distribution.MinValue = 1.0
      </code> <br>
 * <br>
 * or with any of the following combinations: <br>
 * (Mean OR MinValue) AND (CoefficientOfVariation OR Alpha), e.g., <br>
 * <br>
 * <code>
         [...].Distribution.Mean = 1.83
         <br>
         [...].Distribution.CoefficientOfVariation = 1.51
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class ParetoDistribution extends ContinuousDistribution {

   public final double alphaReciprocal;
   public final double minValue;

   private static final String ALPHA = "Alpha";
   private static final String MIN_VALUE = "MinValue";
   private static final String MEAN = "Mean";
   private static final String COEFF_OF_VARIATION = "CoefficientOfVariation";

   /** as required by {@link ReflectionConstructable} */
   public static ParetoDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      double alpha = Double.NaN;

      if (pars.hasParameter(ownNode, ALPHA)) {
         alpha = pars.get(ownNode, ALPHA).asDouble();

      } else if (pars.hasParameter(ownNode, COEFF_OF_VARIATION)) {
         final double coefficientOfVariation = pars.get(ownNode, COEFF_OF_VARIATION).asDouble();
         alpha = 1 + Math.sqrt(1 + 1.0 / (coefficientOfVariation * coefficientOfVariation));

      } else {
         throw new SimulationException("alpha value or coefficient of variation must be specified");
      }

      double minValue = Double.NaN;

      if (pars.hasParameter(ownNode, MIN_VALUE)) {
         minValue = pars.get(ownNode, MIN_VALUE).asDouble();

      } else if (pars.hasParameter(ownNode, MEAN)) {
         final double mean = pars.get(ownNode, MEAN).asDouble();
         minValue = mean * (1 - 1 / alpha);

      } else {
         throw new SimulationException("mean value of min value must be specified.");
      }

      return new ParetoDistribution(alpha, minValue, rng);
   }

   public ParetoDistribution(double alpha, double minValue, RandomNumberGenerator rng) {
      super(rng);

      if (!(alpha > 0)) {
         throw new DistributionException("alpha must be > 0");
      }

      if (minValue <= 0) {
         throw new DistributionException("minValue must be > 0");
      }

      alphaReciprocal = 1 / alpha;
      this.minValue = minValue;
   }

   public ParetoDistribution(double alpha, double minValue) {
      this(alpha, minValue, getDefaultRNG());
   }

   @Override
   public double next() {
      return minValue / Math.pow(rng.next(), alphaReciprocal);
   }
}
