/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Uniform Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>All (continuous) values \(t\) in the interval \(b_l < t < b_u\) appear
 * with the same probability</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * lower limit \(b_l\)
 * <li>
 * upper limit \(b_u > b_l \)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \begin{cases} \frac{1}{b_u -b_l} &\mbox{for } b_l \le t
 * \le b_u \\ 0 &\mbox{else } \end{cases}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = \begin{cases} 0 &\mbox{for } t < b_l \\
 * \frac{t-b_l}{b_u-b_l} &\mbox{for } b_l \le t < b_u \\ 1 &\mbox{for } t \ge
 * b_l \end{cases} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \frac{b_l + b_u}{2}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \frac{(b_u - b_l)^2}{12}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \frac{1}{\sqrt{3}} \cdot \frac{b_u-b_l}{b_u+b_l}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = \frac{1}{b_u-b_l} \cdot \frac{exp(-b_l s) - exp(-b_u s)}{s}\)
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.UniformDistribution
         <br>
         [...].Distribution.LowerBound = 1.5
         <br>
         [...].Distribution.UpperBound = 13.5
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class UniformDistribution extends ContinuousDistribution {

   public final double lowerBound;
   public final double interval;

   private static final String LOWER_BOUND = "LowerBound";
   private static final String UPPER_BOUND = "UpperBound";

   /** as required by {@link ReflectionConstructable} */
   public static UniformDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double lowerBound = pars.get(ownNode, LOWER_BOUND).asDouble();
      final double upperBound = pars.get(ownNode, UPPER_BOUND).asDouble();
      return new UniformDistribution(lowerBound, upperBound, rng);
   }

   public UniformDistribution(double lowerBound, double upperBound, RandomNumberGenerator rng) {
      super(rng);

      if (lowerBound > upperBound) {
         throw new DistributionException("Lower bound > Upper bound!");
      }

      this.lowerBound = lowerBound;
      interval = upperBound - lowerBound;
   }

   public UniformDistribution(double lowerBound, double upperBound) {
      this(lowerBound, upperBound, getDefaultRNG());
   }

   @Override
   public double next() {
      return lowerBound + interval * rng.next();
   }
}
