/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Weibull Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>The Weibull distribution is often used to model internet traffic because
 * of its heavy tail.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * shape parameter \(\alpha > 0\)
 * <li>
 * scale parameter \(\beta > 0\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \alpha \cdot \beta^{-\alpha} \cdot t^{\alpha -1} \cdot
 * exp(-(\frac{t}{\beta})^{\alpha}) \mbox{ for } t>0 \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = 1- exp(-(\frac{t}{\beta})^{\alpha}) \mbox{ for }
 * t>0\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \frac{\beta}{\alpha} \cdot \Gamma(\frac{1}{\alpha}\), whereas
 * \(\Gamma(x)\) is the gamma function</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \frac{\beta^2}{\alpha} \cdot \left\{ 2\Gamma(\frac{2}{\alpha})
 * - \frac{1}{\alpha} \cdot \Gamma(\frac{1}{\alpha})^2 \right\} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T=
 * \sqrt{\frac{2\alpha\Gamma(\frac{2}{\alpha})}{(\Gamma(\frac{1}{\alpha}))^2}
 * -1}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.WeibullDistribution
         <br>
         [...].Distribution.Alpha = 0.3
         <br>
         [...].Distribution.Beta = 0.03
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class WeibullDistribution extends ContinuousDistribution {

   public final double alphaReciprocal;
   public final double beta;

   private static final String ALPHA = "Alpha";
   private static final String BETA = "Beta";

   /** as required by {@link ReflectionConstructable} */
   public static WeibullDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double alpha = pars.get(ownNode, ALPHA).asDouble();
      final double beta = pars.get(ownNode, BETA).asDouble();
      return new WeibullDistribution(alpha, beta, rng);
   }

   public WeibullDistribution(double alpha, double beta, RandomNumberGenerator rng) {
      super(rng);

      if (alpha <= 0) {
         throw new DistributionException("alpha must be > 0");
      }

      if (beta <= 0) {
         throw new DistributionException("beta must be > 0");
      }

      alphaReciprocal = 1 / alpha;
      this.beta = beta;
   }

   public WeibullDistribution(double alpha, double beta) {
      this(alpha, beta, getDefaultRNG());
   }

   @Override
   public double next() {
      return beta * Math.pow(-Math.log(rng.next()), alphaReciprocal);
   }
}
