/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.ProbabilityVector;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.continuous.NegExpDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * GMPP (Generally Modulated Poisson Process)<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Modeling of sources with multiple states (usually 2 - 5) in ATM networks</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>State machine with \(m\) states. <br>
 * <ul>
 * <li>
 * Markovian process with the rate \(\lambda_i\) in the state \(i\)
 * <li>
 * number of arrivals \(X_i\) in state \(i\) is arbitrarily (discretely)
 * distributed (often shifted geometrically -> MMPP); in the silence state
 * \(X_i\) means the number of time slots of the silence phase.
 * <li>
 * according to definition: \(p_{ii} = 0\) and \(\sum\limits_{i=1}^m p_{ij} =
 * 1\).
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * number of states \(m\)
 * <li>
 * transition probabilities \(p_{ij}\)
 * <li>
 * distribution of the number of arrivals \(X_i\) in the individual states
 * <li>
 * arrival rates \(\lambda_i\) in the states
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>Moments and distribution of the arrival intervals</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].distribution = ikr.simlib.distributions.continuous.statedep.GMPPDistribution
         <br>
         [...].distribution.States = 2
         <br>
         [...].distribution.MeanIATs = [1 5]
         <br>
         [...].distribution.StateDuration_0 = Constant
         <br>
         [...].distribution.StateDuration_0.Mean = 10000
         <br>
         [...].distribution.StateDuration_1 = Constant
         <br>
         [...].distribution.StateDuration_1.Mean = 10000
         <br>
         [...].distribution.TransitionProbs = [[0 1] [1 0]]
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * P. J. KHN: Reminder on queueing theory for ATM networks. Telecommunication
 * Systems, No. 5, 1996, pp. 1-24.
 * <li>
 * G. D. STAMOULIS, M. E. ANAGNOSTOU, A. D. GEORGANTAS: Traffic source models
 * for ATM networks: a survey. Computer Communications, Vol. 17, No. 6, Juni,
 * 1994.
 * <li>
 * T. THEIMER: How to compute the moments of a GMDP, RACE 1022, Document
 * UST_123_0023_CD_CC, 1989.
 * </ul>
 * </td>
 * </tr>
 * </table>
 * 
 */
public class GMPPDistribution extends ContinuousDistribution {

   private static final String STATES = "States";
   private static final String MEAN_IATS = "MeanIATs";
   private static final String TRANSITION_PROBS = "TransitionProbs";
   private static final String STATE_DURATIONS = "StateDurations_";

   /** as required by {@link ReflectionConstructable} */
   public static GMPPDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int noOfStates = pars.get(ownNode, STATES).asInteger();
      final double[] interArrivalTimes = pars.get(ownNode, MEAN_IATS).asDoubleVector();
      final double[][] transitionProbs = pars.get(ownNode, TRANSITION_PROBS).asDoubleMatrix();

      final ContinuousDistribution[] stateDurations = new ContinuousDistribution[noOfStates];
      for (int i = 0; i < noOfStates; i++) {
         stateDurations[i] = pars.get(ownNode, STATE_DURATIONS + i).asInstance(ContinuousDistribution.class, ownNode.createChildNode(STATE_DURATIONS + i), pars, rng);

      }

      // FIXME: make sure that vector and matrix dimensions match

      return new GMPPDistribution(noOfStates, interArrivalTimes, stateDurations, transitionProbs, rng);
   }

   public static class StateParameters {

      public final NegExpDistribution interArrivalTime;
      public final ContinuousDistribution stateDuration;
      public final ProbabilityVector cumulatedTransitionProbs;

      public StateParameters(double meanIAT, ContinuousDistribution stateDuration, double[] transitionProbs) {
         interArrivalTime = new NegExpDistribution(meanIAT);
         this.stateDuration = stateDuration;
         cumulatedTransitionProbs = new ProbabilityVector(transitionProbs);
      }
   }

   public final int noOfStates;
   public final StateParameters[] stateParameters;

   private double residualPhaseTime;
   private int currentState;

   public GMPPDistribution(int noOfStates, double[] meanIATs, ContinuousDistribution[] stateDurations, double[][] transitionProbs, RandomNumberGenerator rng) {
      super(rng);

      this.noOfStates = noOfStates;
      stateParameters = new StateParameters[noOfStates];

      for (int i = 0; i < noOfStates; i++) {
         stateParameters[i] = new StateParameters(meanIATs[i], stateDurations[i], transitionProbs[i]);
      }
   }

   public GMPPDistribution(int noOfStates, double[] meanIATs, ContinuousDistribution[] stateDurations, double[][] transitionProbs) {
      this(noOfStates, meanIATs, stateDurations, transitionProbs, getDefaultRNG());
   }

   @Override
   public double next() {
      boolean stop = false;
      double result = 0;

      do {
         double interArrivalTime = stateParameters[currentState].interArrivalTime.next();
         if (interArrivalTime < residualPhaseTime) {
            residualPhaseTime -= interArrivalTime;
            result += interArrivalTime;
            stop = true;
         } else {
            result += residualPhaseTime;
            currentState = stateParameters[currentState].cumulatedTransitionProbs.next(rng.next());
            residualPhaseTime = stateParameters[currentState].stateDuration.next();
         }
      } while (!stop);

      return result;
   }

   public double getResidualPhaseTime() {
      return residualPhaseTime;
   }

   public int getCurrentState() {
      return currentState;
   }
}
