/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.continuous.UniformDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;
import ikr.simlib.support.Vectors;

import java.util.Arrays;

/**
 * WSS (Wide Sense Stationary Process) <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Noise generator (conversion of white noise to bandwidth-limited, colored
 * noise with the help of a filter)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>
 * <ul>
 * <li>
 * White noise at the input of a filter is uniformly distributed in the interval
 * \((-\sqrt{3}, \sqrt{3})\).
 * <li>
 * Behavior is defined by \(n\) filter coefficients \(h_0, ..., h_{n-1}\) as
 * well as the mean value \(\mu\).
 * <li>
 * Condition for positive event intervals: \(\sum\limits_{i=0}^{n-1} |h_i| \cdot
 * \sqrt{3} < \mu\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * mean value \(\mu\)
 * <li>
 * filter coefficients \(h_i (i = 0, ..., n-1, n > 0)\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>auto covariance sequence: \(R_i = \sum\limits_{k=0}^{n-1-k} h_k \cdot
 * h_{k+i} (i = 0, ..., n)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].distribution = ikr.simlib.distributions.continuous.statedep.WSSDistribution
         <br>
         [...].distribution.Mean = 2
         <br>
         [...].distribution.Coefficients = [0.2 0.3 -0.05]
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class WSSDistribution extends ContinuousDistribution {

   public static final double WSS_RANDOM_AMPLITUDE = Math.sqrt(3);

   public final double mean;
   public final double[] filterCoefficients;
   public final double[] filterMemory;
   public final UniformDistribution uniformDist;
   private int currentPos;

   private static final String MEAN = "Mean";
   private static final String COEFFICIENTS = "Coefficients";

   /** as required by {@link ReflectionConstructable} */
   public static WSSDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      final double[] list = pars.get(ownNode, COEFFICIENTS).asDoubleVector();
      return new WSSDistribution(mean, list, rng);
   }

   public WSSDistribution(double mean, double[] filterCoefficients, RandomNumberGenerator rng) {
      super(rng);

      if (Vectors.sum(filterCoefficients) * WSS_RANDOM_AMPLITUDE > mean) {
         throw new DistributionException("Possibility of negative samples since sum of filter coefficients * amplitude > mean");
      }

      this.mean = mean;
      this.filterCoefficients = Arrays.copyOf(filterCoefficients, filterCoefficients.length);
      filterMemory = new double[filterCoefficients.length];
      uniformDist = new UniformDistribution(-WSS_RANDOM_AMPLITUDE, WSS_RANDOM_AMPLITUDE);
      currentPos = 0;
   }

   public WSSDistribution(double mean, double[] filterCoefficients) {
      this(mean, filterCoefficients, getDefaultRNG());
   }

   @Override
   public double next() {
      if (currentPos == 0) {
         currentPos = filterMemory.length - 1;
      } else {
         currentPos--;
      }

      filterMemory[currentPos] = uniformDist.next();

      double sum = 0;
      for (int i = 0; i < filterCoefficients.length; i++) {
         int memPos = (currentPos + i) % filterMemory.length;
         sum += filterMemory[memPos] * filterCoefficients[i];
      }

      return mean + sum;
   }

   @Override
   public void reset() {
      Arrays.fill(filterMemory, 0);
   }

   public int getCurrentPos() {
      return currentPos;
   }
}
