/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.transform;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Nested Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>The nested distribution gives the distribution of a sum of random
 * variables \(T_1, T_2, ..., T_N\) that each are described by a continuous
 * ("inner") distribution. The number of addends \(N\) itself is a random
 * variable with discrete ("outer") distribution. In literature, this
 * distribution is also denoted as compound distribution.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * inner distribution with PDF \(g(t)\) and DF \(G(t)\)
 * <li>
 * outer distribution \(p_i\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \sum\limits_{n=0}^{\infty} p_n \cdot (g_1(t) \otimes
 * ... \otimes g_n(t)) \mbox{ with } g_i(t) \equiv g(t) \forall i\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T] = E[N] \cdot E[T_i]\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= VAR[T_i] \cdot E[N] + VAR[N] \cdot (E[T_i])^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T = \sqrt{\frac{c_{T_i}^2}{E[N]} + c_N^2}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = H(\Psi(s))\) , if \(H(z)\) represents the generating function
 * of the external and \(\Psi(s)\) the LST of the inner distribution.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.transform.NestedDistribution
         <br>
         [...].Distribution.InnerDist = ikr.simlib.distributions.continuous.ConstantDistribution
         <br>
         [...].Distribution.InnerDist.Mean = 2.3
         <br>
         [...].Distribution.OuterDist = ikr.simlib.distributions.discrete.GeometricDistribution
         <br>
         [...].Distribution.OuterDist.Mean = 9
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class NestedDistribution extends ContinuousDistribution {
   
   private static final String INNER_DISTRIBUTION = "InnerDist";
   private static final String OUTER_DISTRIBUTION = "OuterDist";


   /** as required by {@link ReflectionConstructable} */
   public static NestedDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      return new NestedDistribution(pars.get(ownNode, INNER_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(INNER_DISTRIBUTION), pars, rng), pars.get(ownNode, OUTER_DISTRIBUTION).asInstance(DiscreteDistribution.class, ownNode.createChildNode(OUTER_DISTRIBUTION), pars, rng));
   }

   public final ContinuousDistribution innerDist;
   public final DiscreteDistribution outerDist;

   public NestedDistribution(ContinuousDistribution innerDist, DiscreteDistribution outerDist) {
      super();

      this.innerDist = innerDist;
      this.outerDist = outerDist;
   }

   @Override
   public double next() {
      final int outerValue = outerDist.next();

      double result = 0;

      for (int i = 0; i < outerValue; i++) {
         result += innerDist.next();
      }

      return result;
   }
}
