/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.transform;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Piece-wise Linear Distribution <br>
 * 
 * <center><img src="{@docRoot}/ikr/simlib/doc-files/piecewiseLinP39.svg"
 * width="600" alt="Piece-wise Linear Distribution"><br>
 * Piece-wise Linear Distribution</center>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Distribution of a real random variable \(T = d \cdot (Z + Y)\), whereby
 * \(Z\) is an arbitrarily distributed integer number random variable, \(Y\) is
 * an uniformly distributed continuous random variable between 0 and 1 and \(d\)
 * is the scale factor. <br>
 * Application example: empirical distribution according to [1], whereby the
 * distribution of \(Z\) is a known histogram</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * scale factor/time slot duration \(d\)
 * <li>
 * discrete distribution \(p_i\) of \(Z\) / the number of time slots
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \frac{1}{d} \cdot \sum\limits_{n=0}^{\infty} p_n \cdot
 * (\sigma(t-nd) - \sigma(t- (n+1)d))\) <br>
 * <br>
 * whereby \(\sigma(x)\) represents the step function.
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T] = (E[N] + \frac{1}{2}) \cdot d\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.transform.PiecewiseLinearDistribution
         <br>
         [...].Distribution.SlotDuration = 1.5
         <br>
         [...].Distribution.NoOfSlotsDist = ikr.simlib.distributions.discrete.GeometricDistribution
         <br>
         [...].Distribution.NoOfSlotsDist.Mean = 9
         <br>
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>A. M. LAW, W. D. KELTON: Simulation Modeling & Analysis, 2nd edition,
 * McGraw- Hill, 1991.</td>
 * </table>
 * 
 */
public class PiecewiseLinearDistribution extends ContinuousDistribution {

   private static final String NO_OF_SLOTS_DISTRIBUTION = "NoOfSlotsDist";
   private static final String SLOT_DURATION = "SlotDuration";

   /** as required by {@link ReflectionConstructable} */
   public static PiecewiseLinearDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final DiscreteDistribution noOfSlotsDist = pars.get(ownNode, NO_OF_SLOTS_DISTRIBUTION).asInstance(DiscreteDistribution.class, ownNode.createChildNode(NO_OF_SLOTS_DISTRIBUTION), pars, rng);
      final double slotDuration = pars.get(ownNode, SLOT_DURATION).asDouble();
      return new PiecewiseLinearDistribution(noOfSlotsDist, slotDuration, rng);
   }

   public final DiscreteDistribution noOfSlotsDist;
   public final double slotDuration;

   public PiecewiseLinearDistribution(DiscreteDistribution noOfSlotsDist, double slotDuration, RandomNumberGenerator rng) {
      super(rng);
      this.noOfSlotsDist = noOfSlotsDist;
      this.slotDuration = slotDuration;
   }

   public PiecewiseLinearDistribution(DiscreteDistribution noOfSlotsDist, double slotDuration) {
      this(noOfSlotsDist, slotDuration, getDefaultRNG());
   }

   @Override
   public double next() {
      return slotDuration * (noOfSlotsDist.next() + rng.next());
   }
}
