/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.transform;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Linear Transformed Continuous Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Distribution of a random variable \(T\), that results from a linear
 * transformation \(T = aZ + b\) of the random variable \(Z\) with a given
 * continuous distribution ("base distribution").</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * base distribution with PDF \(g(t)\) and DF \(G(t)\)
 * <li>
 * factor \(a \neq 0\)
 * <li>
 * offset \(b\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = g(\frac{t-b}{a})\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = G(\frac{t-b}{a})\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T] = a \cdot E[Z] +b\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= a^2 \cdot VAR[Z]\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T = \frac{1}{\frac{1}{c_z} + \frac{b}{a \sqrt{VAR[Z]}}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.transform.TransformedDistribution
         <br>
         [...].Distribution.BaseDist = ikr.simlib.distributions.continuous.NegExpDistribution
         <br>
         [...].Distribution.BaseDist.Mean = 2.5
         <br>
         [...].Distribution.Factor = 2.0 # optional, default = 1
         <br>
         [...].Distribution.Offset = 10.0 # optional, default = 0
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class TransformedDistribution extends ContinuousDistribution {

   private static final String OFFSET = "Offset";
   private static final String FACTOR = "Factor";
   private static final String BASE_DISTRIBUTION = "BaseDist";

   /** as required by {@link ReflectionConstructable} */
   public static TransformedDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double offset = pars.get(ownNode, OFFSET).asDouble();
      final double factor = pars.get(ownNode, FACTOR).asDouble();
      final ContinuousDistribution baseDist = pars.get(ownNode, BASE_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(BASE_DISTRIBUTION), pars, rng);
      return new TransformedDistribution(baseDist, offset, factor);
   }

   public final ContinuousDistribution baseDist;
   public final double offset;
   public final double factor;

   public TransformedDistribution(ContinuousDistribution baseDist, double offset, double factor) {
      super();
      this.baseDist = baseDist;
      this.offset = offset;
      this.factor = factor;
   }

   public TransformedDistribution(ContinuousDistribution baseDist, double offset) {
      this(baseDist, offset, 1);
   }

   public TransformedDistribution(ContinuousDistribution baseDist) {
      this(baseDist, 0, 1);
   }

   @Override
   public double next() {
      return offset + factor * baseDist.next();
   }
}
