/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Discrete Deterministic Value<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>A constant integer value \(d\) is returned</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>constant integer (mean) value \(d\)</td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = \begin{cases} 1 &\mbox{for } i=d \\ 0 & \mbox{else }
 * \end{cases} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]=d\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]=0\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_X=0\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)=z^d \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.discrete.DiscreteConstantDistribution
         <br>
         [...].Distribution.Mean = 5</code></td>
 * </tr>
 * </table>
 */

public class DiscreteConstantDistribution extends DiscreteDistribution {

   public final int mean;

   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static DiscreteConstantDistribution createInstance(SimNode ownNode, Parameters pars, @SuppressWarnings("unused") RandomNumberGenerator rng) {
      final int mean = pars.get(ownNode, MEAN).asInteger();
      return new DiscreteConstantDistribution(mean);
   }

   public DiscreteConstantDistribution(int mean) {
      this.mean = mean;
   }

   @Override
   public int next() {
      return mean;
   }
}
