/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Geometric Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Probability for \(i\) failures prior to the first success in independent
 * Bernoulli experiments with the parameter \(q(0 \le q \le 1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>success probability \(q(0 < q \le 1)\) <br>
 * <br>
 * With mean parameter \(m : q = \frac{1}{1+m}\)</td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = (1-q)^i \cdot q = (\frac{m}{m+1})^i \cdot \frac{1}{m+1}
 * \mbox{for } i = 0,1,2,... \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]=\frac{1-q}{q}=m \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]=\frac{1-q}{q^2} = m \cdot (m+1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \sqrt{ \frac{1}{1-q}} = \sqrt{ \frac{m+1}{m} } \ge 1\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)= \frac{q}{1-z(1-q)} = \frac{1}{m+1-zm} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.discrete.GeometricDistribution
         <br>
         [...].Distribution.Mean = 2.5</code></td>
 * </tr>
 * </table>
 * 
 */
public class GeometricDistribution extends DiscreteDistribution {

   public final double geomet_q;

   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static GeometricDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      return new GeometricDistribution(mean, rng);
   }

   public GeometricDistribution(double mean, RandomNumberGenerator rng) {
      super(rng);

      if (mean <= 0) {
         throw new DistributionException("Mean <= 0");
      }

      geomet_q = Math.log(mean / (mean + 1));
   }

   public GeometricDistribution(double mean) {
      this(mean, getDefaultRNG());
   }

   @Override
   public int next() {
      return (int) Math.floor(Math.log(rng.next()) / geomet_q);
   }
}
