/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Negative Binomial Distribution <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Probability, that \(i\) failed Bernoulli trials with the parameter \(q\)
 * will precede \(k\) successes</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * success probability \(q(0 \le q \le 1)\)
 * <li>
 * number of successful trial \(k \ge 0\)
 * </ul>
 * With mean value and variance parameters:
 * <ul>
 * <li>
 * \(q= \frac{VAR[X]}{E[X]} \)
 * <li>
 * \(k= \frac{E[X]}{VAR[X] - E[X]}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = \binom{i+k-1}{k-1} \cdot q^k \cdot (1-q)^i = \binom{i+
 * \frac{m^2}{v-m} -1}{ \frac{m^2}{v-m} -1} \cdot
 * (\frac{m}{v})^{\frac{m^2}{v-m}} \cdot (1- \frac{m}{v})^i \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]= \frac{1-q}{q} \cdot k = m\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]= \frac{1-q}{q^2} \cdot k = \frac{m}{q} = v\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \frac{1}{ \sqrt{(1-q) \cdot k}}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)= \frac{q^k}{(1-(1-q) \cdot z)^k} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.discrete.NegBinDistribution
         <br>
         [...].Distribution.Mean = 5
         <br>
         [...].Distribution.Variance = 10
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class NegBinDistribution extends DiscreteDistribution {

   public final double negBin_p;
   public final double negBin_r;

   private static final String MEAN = "Mean";
   private static final String VARIANCE = "Variance";

   /** as required by {@link ReflectionConstructable} */
   public static NegBinDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      final double variance = pars.get(ownNode, VARIANCE).asDouble();
      return new NegBinDistribution(mean, variance, rng);
   }

   public NegBinDistribution(double mean, double variance, RandomNumberGenerator rng) {
      super(rng);

      if (mean <= 0) {
         throw new DistributionException("mean <= 0");
      }

      if (mean >= variance) {
         throw new DistributionException("mean >= variance");
      }

      negBin_p = 1 - mean / variance;
      negBin_r = mean * (1 - negBin_p) / negBin_p;
   }

   public NegBinDistribution(double mean, double variance) {
      this(mean, variance, getDefaultRNG());
   }

   public double nextBetaRandomNumber(double _a, double _b) {
      // X = Y/(Y+Z) is Beta distributed if Y+Z<=1 is true.
      // Y=U1^(1/a); Z=U2^(1/b); U1, U2 need to be independent (0;1)
      // distributed random numbers.

      double y, z;
      double a = 1 / _a;
      double b = 1 / _b;

      do {
         y = Math.pow(rng.next(), a);
         z = Math.pow(rng.next(), b);
      } while (y + z > 1);

      return y / (y + z);
   }

   public double nextGammaRandomNumber(double alpha, double beta) {
      /**
       * The Gamma distributed sample is computed using a Beta and an
       * exponential distributed RandomNumber (see D.Walters work for details).
       */

      double k;
      double gamma;
      double u = 0, x = 0, y = 0, z = 0;
      int i;

      k = Math.floor(alpha);
      gamma = alpha - k;

      if (k != 0) {
         u = 1;
         for (i = 1; i <= k; i++)
            u *= rng.next();
         x = -Math.log(u);
      }
      if (gamma > 0) {
         y = nextBetaRandomNumber(gamma, 1 - gamma);
         z = -Math.log(rng.next());
      }
      return beta * (x + y * z);
   }

   @Override
   public int next() {
      /**
       * A Gamma distributed random number Y=Ga(r, p/1-p) is used as mean for a
       * Poisson distributed random Number. Then X=P(Y) is negative binomial
       * distributed.
       */

      PoissonDistribution pd = new PoissonDistribution(nextGammaRandomNumber(negBin_r, negBin_p / (1 - negBin_p)));

      return pd.next();
   }
}
