/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Shifted Geometric Distribution <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Probability for \(i-1\) failures prior to the first success in
 * independent Bernoulli experiments with the parameter \(q(0 \le q \le 1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>Success probability \(q\) <br>
 * <br>
 * With mean value parameter \(m : q = \frac{1}{m}\)</td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = (1-q)^{i-1} \cdot q = (\frac{m-1}{m})^i \cdot \frac{1}{m-1}
 * \mbox{for } i = 1,2,... \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]=\frac{1}{q} = m \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]=\frac{1-q}{q^2} = m \cdot (m-1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \sqrt{ 1-q } = \sqrt{ \frac{m-1}{m}} \le 1\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)= \frac{qz}{1-z(1-q)} = \frac{z}{m-z(m-1)}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.discrete.ShiftedGeometricDistribution
         <br>
         [...].Distribution.Mean = 2.5</code></td>
 * </tr>
 * </table>
 * 
 */
public class ShiftedGeometricDistribution extends DiscreteDistribution {

   public final double shiGeo_q;

   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static ShiftedGeometricDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      return new ShiftedGeometricDistribution(mean, rng);
   }

   public ShiftedGeometricDistribution(double mean, RandomNumberGenerator rng) {
      super(rng);

      if (mean <= 1) {
         throw new DistributionException("Mean <= 1");
      }

      shiGeo_q = Math.log((mean - 1) / mean);
   }

   public ShiftedGeometricDistribution(double mean) {
      this(mean, getDefaultRNG());
   }

   @Override
   public int next() {
      return (int) Math.floor(Math.log(rng.next()) / shiGeo_q) + 1;
   }
}
