/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete.statedep;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.distributions.ProbabilityVector;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.distributions.discrete.ShiftedGeometricDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;
import ikr.simlib.support.Vectors;

/**
 * DMAP (Discrete time Markovian Arrival Process)<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Modeling from sources with multiple states</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>
 * <ul>
 * <li>
 * State machine with \(m\) states.
 * <li>
 * Discrete Markovian process with the transition probability \(p_{ij}\) from
 * state \(i\) to state \(j\) after each time slot \(\Delta t\), which is a
 * combination of two components: \(p_{ij} = c_{ij} + d_{ij}\)
 * <li>
 * Transition probability from state \(i\) to state \(j\) without an arrival
 * event upon transition: \(c_{ij}\)
 * <li>
 * Transition probability from state \(i\) to state \(j\) with an arrival event
 * upon transition: \(d_{ij}\)
 * <li>
 * According to definition: \(\sum\limits_i p_{ij} = 1\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * number of states \(m\)
 * <li>
 * transition probabilities without arrival \(c_{ij}\)
 * <li>
 * transition probabilities with arrival \(d_{ij}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>see C. BLONDIA, T. THEIMER</td>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].distribution = ikr.simlib.distributions.discrete.statedep.DMAPDistribution
         <br>
         [...].distribution.States = 2
         <br>
         [...].distribution.CMAP = [[0.2 0.5] [0.1 0.3]]
         <br>
         [...].distribution.DMAP = [[0.2 0.1] [0.2 0.4]]
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * C. BLONDIA, T. THEIMER: A Discrete-Time Model for ATM Traffic, RACE 1022,
 * Document PRLB_123_0018_CD_CC/UST_123_0022_CD_CC, 1989.
 * <li>
 * P. J. KHN: "Reminder on queueing theory for ATM networks." Telecommunication
 * Systems, No. 5, 1996, pp. 1-24.
 * </ul>
 * </td>
 * </tr>
 * </table>
 * 
 */
public class DMAPDistribution extends DiscreteDistribution {

   private static final String STATES = "States";
   private static final String CMAP = "CMAP";
   private static final String DMAP = "DMAP";

   /** as required by {@link ReflectionConstructable} */
   public static DMAPDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int noOfStates = pars.get(ownNode, STATES).asInteger();
      final double[][] cMatrix = pars.get(ownNode, CMAP).asDoubleMatrix();
      final double[][] dMatrix = pars.get(ownNode, DMAP).asDoubleMatrix();
      return new DMAPDistribution(noOfStates, cMatrix, dMatrix, rng);
   }

   public static class DMAPStateParameters {

      /**
       * Distribution for number of slots until a state transition or event
       * occurs
       */
      public final ShiftedGeometricDistribution idleTimeDist;

      /**
       * Probability that a event without a transition occurs under condition
       * that transition or event occurs
       */
      public final double eventNoTransProb;

      /**
       * Transition probabilities under condition that transition occurs
       */
      public final ProbabilityVector cumulatedTransitionProbs;

      /**
       * Probabilities that event occurs on transition to a certain state
       */
      public final double[] eventProbs;

      public DMAPStateParameters(double noTransNoEventProb, double eventNoTransProb, double[] transitionProbs, double[] eventProbs) {
         idleTimeDist = new ShiftedGeometricDistribution(1 / (1 - noTransNoEventProb));
         this.eventNoTransProb = eventNoTransProb;
         cumulatedTransitionProbs = new ProbabilityVector(transitionProbs);
         this.eventProbs = eventProbs;
      }
   }

   public final int noOfStates;
   public final DMAPStateParameters[] stateParameters;

   private int currentState;

   public DMAPDistribution(int noOfStates, double[][] c, double[][] d, RandomNumberGenerator rng) {
      super(rng);

      this.noOfStates = noOfStates;
      stateParameters = new DMAPStateParameters[noOfStates];
      currentState = 0;

      final double cEpsilon = 1e-10;

      for (int i = 0; i < noOfStates; i++) {
         double noTransNoEventProb = c[i][i];
         double transitionProbs[] = Vectors.add(c[i], d[i]);
         double[] eventProbs = new double[noOfStates];
         double eventNoTransProb;

         /** Check whether parameters are correct */
         if (Math.abs(Vectors.sum(transitionProbs) - 1) > cEpsilon) {
            throw new DistributionException("sum of (c_ij + d_ij) mus be 1");
         }
         if (Math.abs(noTransNoEventProb - 1) <= cEpsilon) {
            throw new DistributionException("c_ii must be < 1");
         }
         for (int k = 0; k < noOfStates; k++) {
            if ((d[i][k] < 0) || (d[i][k] > 1) || (c[i][k] < 0) || (c[i][k] > 1)) {
               throw new DistributionException("transition prob. must be >= 0 and <= 1");
            }
         }

         if (c[i][i] + d[i][i] < 1) {
            transitionProbs[i] = 0;
            Vectors.normalize(transitionProbs);
            eventNoTransProb = d[i][i] / (1 - c[i][i]);
         } else {
            transitionProbs[i] = 1;
            eventNoTransProb = 1;
         }

         for (int k = 0; k < noOfStates; k++) {
            double sum = c[i][k] + d[i][k];
            if (sum > 0) {
               eventProbs[k] = d[i][k] / sum;
            } else {
               eventProbs[k] = 0;
            }
         }

         stateParameters[i] = new DMAPStateParameters(noTransNoEventProb, eventNoTransProb, transitionProbs, eventProbs);
      }

      reset();
   }

   public DMAPDistribution(int noOfStates, double[][] c, double[][] d) {
      this(noOfStates, c, d, getDefaultRNG());
   }

   @Override
   public int next() {
      int silenceCounter = 0;
      boolean eventFlag;

      do {
         silenceCounter += stateParameters[currentState].idleTimeDist.next();
         if (rng.next() <= stateParameters[currentState].eventNoTransProb) {
            eventFlag = true;
         } else {
            int nextState = stateParameters[currentState].cumulatedTransitionProbs.next(rng.next());
            eventFlag = (rng.next() <= stateParameters[currentState].eventProbs[nextState]);
            currentState = nextState;
         }
      } while (!eventFlag);

      return silenceCounter;
   }

   @Override
   public void reset() {
      currentState = 0;
   }

   public int getCurrentState() {
      return currentState;
   }
}
