/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete.transform;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Ceiled Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>The "ceiled distribution" is a discrete distribution that converts the
 * value obtained from a continuous ("base") distribution to an integer value
 * using the <code>
         ceil()</code> function from the Java <code>
         Math</code> package. <code>
         ceil(x)</code>
 * implements the ceiling operator \(\left \lceil x \right \rceil\), i.e. it
 * yields the smallest integer value not less than \(x\).</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>continuous base distribution \(f(x)\) (DF \(F(x)\))</td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = F(i)-F(i-1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.discrete.transform.CeiledDistribution
         <br>
         [...].Distribution.BaseDist = ikr.simlib.distributions.continouos.NegExpDistribution
         <br>
         [...].Distribution.BaseDist.Mean = 2.5
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class CeiledDistribution extends DiscreteDistribution {

   private static final String BASE_DISTRIBUTION = "BaseDist";

   /** as required by {@link ReflectionConstructable} */
   public static CeiledDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final ContinuousDistribution baseDist = pars.get(ownNode, BASE_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(BASE_DISTRIBUTION), pars, rng);
      return new CeiledDistribution(baseDist);
   }

   public final ContinuousDistribution baseDist;

   public CeiledDistribution(ContinuousDistribution baseDist) {
      super();
      this.baseDist = baseDist;
   }

   /**
    * Returns the ceiled value of the base distribution or
    * <code>Integer.MIN/MAX_VALUE</code> if the value exceeds the integer range.
    */
   @Override
   public int next() {
      final double contValue = baseDist.next();

      if (contValue > Integer.MIN_VALUE && contValue < Integer.MAX_VALUE) {
         return (int) Math.ceil(contValue);
      } else if (contValue < 0) {
         return Integer.MIN_VALUE;
      } else {
         return Integer.MAX_VALUE;
      }
   }
}
