/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete.transform;

import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Discrete Bounded Distribution<br>
 * 
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Discrete analogon to the continuous bounded distribution, whereby the
 * upper and lower limits each belong to the value range of the random variable
 * \(X\).</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * base distribution \(q_i\) with the corresponding DF \(Q_i\)
 * <li>
 * lower limit \(b_l\) (integer number)
 * <li>
 * upper limit \(b_u\) (integer number)
 * <li>
 * boolean value <code>
               resampling</code>
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = p_i = \begin{cases} \frac{q_i}{Q_{b_u}-Q_{b_i}} &\mbox{for }
 * b_l \le i \le b_u \\ 0 &\mbox{else} \end{cases}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.discrete.transform.DiscreteBoundedDistribution
         <br>
         [...].Distribution.BaseDist = ikr.simlib.distributions.discrete.Binomial
         <br>
         [...].Distribution.BaseDist.Mean = 10.0
         <br>
         [...].Distribution.BaseDist.Variance = 5.0
         <br>
         [...].Distribution.LowerBound = 0 # optional, default 0
         <br>
         [...].Distribution.UpperBound = 20 # optional, def. inf.
         <br>
         [...].Distribution.resampling = false # optional, def. tr.
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class DiscreteBoundedDistribution extends DiscreteDistribution {

   public final DiscreteDistribution baseDist;
   public final int lowerBound;
   public final int upperBound;
   public final boolean resampling;

   private static final String BASE_DISTRIBUTION = "BaseDist";
   private static final String LOWER_BOUND = "LowerBound";
   private static final String LOWER_BOUND_DEFAULT = "0";
   private static final String UPPER_BOUND = "UpperBound";
   private static final String UPPER_BOUND_DEFAULT = new Integer(Integer.MAX_VALUE).toString();
   private static final String RESAMPLING = "Resampling";
   private static final String RESAMPLING_DEFAULT = "true"; 

   /** as required by {@link ReflectionConstructable} */
   public static DiscreteBoundedDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final DiscreteDistribution baseDist = pars.get(ownNode, BASE_DISTRIBUTION).asInstance(DiscreteDistribution.class, ownNode.createChildNode(BASE_DISTRIBUTION), pars, rng);
      final int lowerBound = pars.getOrUseDefault(ownNode, LOWER_BOUND, LOWER_BOUND_DEFAULT).asInteger();
      final int upperBound = pars.getOrUseDefault(ownNode, UPPER_BOUND, UPPER_BOUND_DEFAULT).asInteger();
      final boolean resampling = pars.getOrUseDefault(ownNode, RESAMPLING, RESAMPLING_DEFAULT).asBoolean();

      return new DiscreteBoundedDistribution(baseDist, lowerBound, upperBound, resampling);
   }

   public DiscreteBoundedDistribution(DiscreteDistribution baseDist, int lowerBound, int upperBound, boolean resampling) {
      super();

      this.baseDist = baseDist;
      this.lowerBound = lowerBound;
      this.upperBound = upperBound;
      this.resampling = resampling;
   }

   public DiscreteBoundedDistribution(DiscreteDistribution baseDist, int lowerBound, int upperBound) {
      this(baseDist, lowerBound, upperBound, Boolean.parseBoolean(RESAMPLING_DEFAULT));
   }

   public DiscreteBoundedDistribution(DiscreteDistribution baseDist, int lowerBound) {
      this(baseDist, lowerBound, Integer.parseInt(UPPER_BOUND_DEFAULT));
   }

   public DiscreteBoundedDistribution(DiscreteDistribution baseDist) {
      this(baseDist, Integer.parseInt(LOWER_BOUND_DEFAULT));
   }

   @Override
   public int next() {
      int result;

      if (resampling) {
         do {
            result = baseDist.next();
         } while ((result < lowerBound) || (result > upperBound));
      } else {
         result = baseDist.next();
         if (result < lowerBound) {
            result = lowerBound;
         } else if (result > upperBound) {
            result = upperBound;
         }
      }

      return result;
   }
}
