/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities;

import ikr.simlib.model.SimNode;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.PortException;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.OutputPort;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * This is the basic entity of the simulation package. All components should be
 * derived from this class. It has all functions to support hierarchical models.
 */
public abstract class Entity {

   /**
    * 
    */
   protected final SimNode simNode;

   protected static class PortAlias {
      public final Port port;
      public final String aliasPortName;

      public PortAlias(Port port, String aliasPortName) {
         this.port = port;
         this.aliasPortName = aliasPortName;
      }
   }

   protected List<Port> ports;
   protected List<PortAlias> portAliases;

   public Entity(SimNode ownNode) {
      // save the entity's SimNode
      simNode = ownNode;

      // create the port lists
      ports = new ArrayList<Port>();
      portAliases = new ArrayList<PortAlias>();
   }

   /**
    * Returns a read-only list of this entity's ports.
    * 
    * @return the list of ports
    */
   public List<Port> getPorts() {
      return ports;
   }

   /**
    * Adds a port to this entity.
    */
   public void addPort(Port port) {
      ports.add(port);
   }

   /**
    * searches port list and returns port with corresponding name
    */
   public Port getPortByName(String localPortName) {
      // search port list
      for (Port port : ports) {
         if (port.getLocalName().equals(localPortName))
            return port;
      }

      // search port aliases list
      for (PortAlias portAlias : portAliases) {
         if (portAlias.aliasPortName.equals(localPortName)) {
            return portAlias.port;
         }
      }

      throw new PortException("Port " + localPortName + " of entity " + simNode.getFullName() + " doesn't exist");
   }

   /**
    * searches port list and returns true, if a port with the right name is in
    * it
    */
   public boolean isPortKnown(String localPortName) {
      // search port list
      for (Port port : ports) {
         if (port.getLocalName().equals(localPortName))
            return true;
      }

      // search port aliases list
      for (PortAlias portAlias : portAliases) {
         if (portAlias.aliasPortName.equals(localPortName))
            return true;
      }

      return false;
   }

   public void aliasPort(Port originalPort, String aliasLocalPortName) {
      if (isPortKnown(aliasLocalPortName))
         throw new PortException("the port name " + aliasLocalPortName + " of entity " + simNode.getFullName() + " is already in use");

      portAliases.add(new PortAlias(originalPort, aliasLocalPortName));
   }

   public void aliasPort(Entity originalPortEntity, String originalLocalPortName, String aliasLocalPortName) {
      Port originalPort = originalPortEntity.getPortByName(originalLocalPortName);

      aliasPort(originalPort, aliasLocalPortName);
   }

   public void unaliasPort(String aliasLocalPortName) {
      Iterator<PortAlias> it = portAliases.iterator();
      while (it.hasNext()) {
         PortAlias portAlias = it.next();
         if (portAlias.aliasPortName.equals(aliasLocalPortName))
            it.remove();
      }
   }

   public void connect(String fromLocalPortName, Entity toEntity, String toLocalPortName) {
      // check port types
      Port fromPort = getPortByName(fromLocalPortName);
      OutputPort outputPort = null;
      if (fromPort instanceof OutputPort)
         outputPort = (OutputPort) fromPort;
      else
         throw new PortException("Wrong port type: port " + fromPort.getName() + " is not an OutputPort");

      Port toPort = toEntity.getPortByName(toLocalPortName);
      InputPort inputPort = null;
      if (toPort instanceof InputPort)
         inputPort = (InputPort) toPort;
      else
         throw new PortException("Wrong port type: port " + toPort.getName() + " is not an InputPort");

      // connect
      outputPort.connect(inputPort);
   }

   public void disconnect(Entity fromEntity, String fromLocalPortName, Entity toEntity, String toLocalPortName) {
      // check port types
      Port fromPort = getPortByName(fromLocalPortName);
      OutputPort outputPort = null;
      if (fromPort instanceof OutputPort)
         outputPort = (OutputPort) fromPort;
      else
         throw new PortException("Wrong port type: port " + fromPort.getName() + " is not an OutputPort");

      Port toPort = toEntity.getPortByName(toLocalPortName);
      InputPort inputPort = null;
      if (toPort instanceof InputPort)
         inputPort = (InputPort) toPort;
      else
         throw new PortException("Wrong port type: port " + toPort.getName() + " is not an InputPort");

      // disconnect
      outputPort.disconnect(inputPort);
   }

   public String getName() {
      return simNode.getName();
   }

   public SimNode getNode() {
      return simNode;
   }
}
