/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.generators;

import ikr.simlib.distributions.Distribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.factories.Factory;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * <code>DStdGenerator</code> generates messages according to a renewal point
 * process for which IAT distribution is provided/parsed. If a distribution for
 * the length of a message is provided/parsed it sets the message length in an
 * iid fashion. That is, each time a message is created, the message length is
 * drawn according to the given distribution. A MessageFactory may be provided
 * to create messages other than Message.
 * 
 * <p>
 * If a length distribution is provided, the length of the default Message
 * generated by the <code>MessageFactory</code> is overwritten.
 * <p>
 * StdGenerator is a very flexible Generator with the following characteristics:
 * <ul>
 * <li>Messages are generated according to the constant interarrival time.
 * <li>The constant interarrival time can be modified with the
 * <code>setInterArrivalTime()</code> method.
 * <li>Message length follows an optional discrete distribution. Each time a new
 * message is generated by a <code>MessageFactory</code>, the length is adapted
 * by calling <code>setLength()</code> of the message and passing the new length
 * according to the length distribution.
 * <li>If no length distribution is provided, the length of the generated
 * message is not changed. Therefore control of the length remains at the
 * <code>MessageFactory</code>.
 * <li>A message factory is required and could either be provided by passing it
 * upon generation of <code>StdGenerator</code> or during the parsing procedure.
 * <li>All possible combinations of message factories and distributions are
 * described in the source code.
 * </ul>
 */
public class DStdGenerator extends Generator {

   private static final String IAT = "IAT";
   private static final String LENGTH_DIST = "LengthDist";

   /** as required by {@link ReflectionConstructable} */
   public static DStdGenerator createInstance(SimNode ownNode, Parameters pars) {
      final Duration interArrivalTime = pars.get(ownNode, IAT).asInstance(Duration.class, ownNode.createChildNode(IAT), pars);
      DiscreteDistribution messageLengthDist = null;
      if (pars.hasParameter(ownNode, LENGTH_DIST)) {
         messageLengthDist = pars.get(ownNode, LENGTH_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(LENGTH_DIST), pars, Distribution.getDefaultRNG());
      }

      return new DStdGenerator(interArrivalTime, messageLengthDist, ownNode);
   }

   protected Duration interArrivalTime;
   protected DiscreteDistribution messageLengthDist;

   /** to be used for subclasses parsing */
   protected DStdGenerator(SimNode ownNode, Factory<Message> factory, boolean activateOnStart) {
      super(ownNode, factory, activateOnStart);
   }

   protected DStdGenerator(SimNode ownNode, Factory<Message> factory) {
      this(ownNode, factory, true);
   }

   protected DStdGenerator(SimNode ownNode, boolean activateOnStart) {
      super(ownNode, null, activateOnStart);
   }

   protected DStdGenerator(SimNode ownNode) {
      this(ownNode, true);
   }

   /**
    * Constructor with factory but without parsing. Distributions are required,
    * factory is optional.
    * 
    */
   public DStdGenerator(Duration iat, DiscreteDistribution lengthdist, SimNode ownNode, Factory<Message> factory, boolean activateOnStart) {
      super(ownNode, factory, activateOnStart);

      interArrivalTime = iat;
      messageLengthDist = lengthdist;
   }

   public DStdGenerator(Duration iat, DiscreteDistribution lengthdist, SimNode ownNode, Factory<Message> factory) {
      this(iat, lengthdist, ownNode, factory, true);
   }

   public DStdGenerator(Duration iat, DiscreteDistribution lengthdist, SimNode ownNode, boolean activateOnStart) {
      this(iat, lengthdist, ownNode, null, activateOnStart);
   }

   public DStdGenerator(Duration iat, DiscreteDistribution lengthdist, SimNode ownNode) {
      this(iat, lengthdist, ownNode, true);
   }

   /**
    * Constructor without parsing of distributions (w and w/o factory)
    */
   public DStdGenerator(Duration iat, SimNode ownNode, Factory<Message> factory, boolean activateOnStart) {
      this(iat, null, ownNode, factory, activateOnStart);
   }

   public DStdGenerator(Duration iat, SimNode ownNode, Factory<Message> factory) {
      this(iat, ownNode, factory, true);
   }

   public DStdGenerator(Duration iat, SimNode ownNode, boolean activateOnStart) {
      this(iat, null, ownNode, null, activateOnStart);
   }

   public DStdGenerator(Duration iat, SimNode ownNode) {
      this(iat, ownNode, true);
   }

   @Override
   protected Message createMessage() {
      Message tempMes = super.createMessage();
      if (messageLengthDist != null)
         tempMes.setLength(messageLengthDist.next());
      return tempMes;
   }

   @Override
   protected PointInTime getNextEventTime(PointInTime currentTime) {
      return  currentTime.plus(interArrivalTime);
   }

   public void setInterArrivalTime(Duration interArrivalTime) {
      this.interArrivalTime = interArrivalTime;
   }
}
