/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.phases;

import ikr.simlib.entities.Entity;
import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.OutputPort;
import ikr.simlib.ports.output.SynchronousOutputPort;
import ikr.simlib.statistics.integral.IntegralStatistic;
import ikr.simlib.statistics.integral.StdIntegralStatistic;

public abstract class InfiniteServer extends Entity implements ReflectionConstructable2<SimNode, Parameters> {

   protected final InputPort inputPort;
   protected final SynchronousOutputPort outputPort;
   protected final IntegralStatistic occupancy;
   protected int activePhases;

   public InfiniteServer(SimNode ownNode) {
      super(ownNode);

      inputPort = new InputPort(this) {
         @Override
         protected void handleMessageIndication(Message msg) {
            InfiniteServer.this.handleMessageIndication();
         }
      };
      outputPort = new SynchronousOutputPort(this);
      occupancy = new StdIntegralStatistic(ownNode.createChildNode("Occupancy"));
      activePhases = 0;
   }

   private void handleMessageIndication() {
      // get message
      final Message m = inputPort.getMessage();
      // wait service time
      Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            onExecute(m);
         }
      }, new CalendarCallback() {
         @Override
         public void execute() {
            onCancel();
         }
      }, Calendar.getInstance().getSystemTime().plus(getNextServiceTime()));
      activePhases++;
      occupancy.update(activePhases);
   }

   private void onExecute(Message msg) {
      // send message to next port
      InfiniteServer.this.outputPort.sendMessage(msg);
      InfiniteServer.this.activePhases--;
      InfiniteServer.this.occupancy.update(InfiniteServer.this.activePhases);
   }

   private void onCancel() {
      InfiniteServer.this.activePhases--;
      InfiniteServer.this.occupancy.update(InfiniteServer.this.activePhases);
   }

   protected abstract Duration getNextServiceTime();

   public InputPort getInput() {
      return inputPort;
   }

   public OutputPort getOutput() {
      return outputPort;
   }

}
