/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.time;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.exceptions.ImportException;

import java.util.Collection;

/**
 * Represents the relative time between two points in time.
 * <p>
 * Can be added to {@link PointInTime} objects or other {@link Duration}
 * objects. For the generation of new {@link Duration} objects, double values
 * are parsed to the internal representation using long values. The method
 * {@link Math#round(double)} is used to get the best fitting long value.
 */
public class Duration extends TimeBase implements Comparable<Duration> {
   private static final long serialVersionUID = 8300617116656794591L;

   public static Duration createInstance(SimNode ownNode, Parameters pars) {
      TimeUnits unitToParse = null;
      for (TimeUnits unit : TimeUnits.values()) {
         if (pars.hasParameter(ownNode, unit.getName())) {
            if (unitToParse != null) {
               throw new RuntimeException("multiple unit values defined for " + ownNode.getFullName());
            }
            unitToParse = unit;
         }
      }
      if (unitToParse == null) {
         throw new RuntimeException("no valid unit value defined for " + ownNode.getFullName());
      }

      final double value = pars.get(ownNode, unitToParse.getName()).asDouble();
      return from(value, unitToParse);
   }

   public static Duration ZERO = new Duration(0);

   public static Duration fromDays(double days) {
      return from(days, TimeUnits.d);
   }

   public static Duration fromHours(double hours) {
      return from(hours, TimeUnits.h);
   }

   public static Duration fromMinutes(double minutes) {
      return from(minutes, TimeUnits.m);
   }

   public static Duration fromSeconds(double seconds) {
      return from(seconds, TimeUnits.s);
   }

   public static Duration fromMilliSeconds(double milliSeconds) {
      return from(milliSeconds, TimeUnits.ms);
   }

   public static Duration fromMicroSeconds(double microSeconds) {
      return from(microSeconds, TimeUnits.us);
   }

   public static Duration fromNanoSeconds(double nanoSeconds) {
      return from(nanoSeconds, TimeUnits.ns);
   }

   public static Duration fromPicoSeconds(double picoSeconds) {
      return from(picoSeconds, TimeUnits.ps);
   }

   public static Duration from(double value, TimeUnits unit) {
      return new Duration(TimeBase.secondsToInternal(value / unit.getTimeBase()));
   }

   public static Duration importBatch(ResultReader reader, String name) throws ImportException {
      reader.readStartNode(Duration.class.getSimpleName(), name);
      final String timeUnit = reader.readString(EXPORT_UNIT);
      final long value = reader.readLong(EXPORT_VALUE);
      reader.readEndNode();

      if (!getTimeUnit().equals(timeUnit))
         throw new ImportException("unknown", "unknown", "invalid time unit: expected " + getTimeUnit() + " but read " + timeUnit);

      return new Duration(value);
   }

   /** Does not need to be positive */
   protected Duration(long value) {
      super(value);
   }

   public static Duration calculateSumOf(Collection<Duration> durations) {
      long sum = 0;
      for (Duration d : durations)
         sum += d.getValue();
      return new Duration(sum);
   }

   /**
    * This method returns the duration between two {@link PointInTime} objects.<br>
    * Be aware: this method assumes ordered arguments and may return negative
    * durations! The result is positive if and only if
    * second.isLaterThan(first).
    * 
    * @param first
    *           typically the earlier {@link PointInTime}
    * @param second
    *           typically the later {@link PointInTime}
    */
   public Duration(PointInTime first, PointInTime second) {
      super(second.getValue() - first.getValue());
   }

   public Duration plus(Duration other) {
      return new Duration(getValue() + other.getValue());
   }

   /**
    * This method is to answer the relation of two durations, e.g. how many
    * seconds go in an hour?
    * 
    * @param other
    *           divisor of the fraction
    * @return the fraction this/other
    */
   public double ratio(Duration other) {
      return getValue() * 1.0 / other.getValue();
   }

   public Duration multiplyBy(long scale) {
      return new Duration(getValue() * scale);
   }

   public Duration multiplyBy(double scale) {
      return new Duration(Math.round(getValue() * scale));
   }

   public Duration divideBy(long scale) {
      return new Duration(getValue() / scale);
   }

   public Duration negated() {
      return new Duration(-1 * getValue());
   }

   @Override
   public int compareTo(Duration other) {
      return Long.compare(getValue(), other.getValue());
   }

   public boolean isGreaterThan(Duration other) {
      return this.getValue() > other.getValue();
   }

   public boolean isGreaterThanOrEqualTo(Duration other) {
      return this.getValue() >= other.getValue();
   }

   public boolean isLessThan(Duration other) {
      return this.getValue() < other.getValue();
   }

   public boolean isLessThanOrEqualTo(Duration other) {
      return this.getValue() <= other.getValue();
   }

   /**
    * Return the longer of two durations.
    * 
    * @param other
    *           the other value
    * @return the longer of the two durations
    */
   public Duration selectLonger(Duration other) {
      final Duration ret;
      if (this.isGreaterThan(other)) {
         ret = this;
      } else {
         ret = other;
      }
      return ret;
   }

   /**
    * Return the shorter of two durations
    * 
    * @param other
    *           the other value
    * @return the shorter of the two durations
    */
   public Duration selectShorter(Duration other) {
      final Duration ret;
      if (this.isGreaterThan(other)) {
         ret = other;
      } else {
         ret = this;
      }
      return ret;
   }
}
