/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.meters.count;

import ikr.simlib.control.simulationcontrol.ControlCounter;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;

/**
 * The previous implementation has been replaced with two helper methods, which
 * instanciate a {@link MessageCountMeter} and a {@link ControlCounter}.
 * <p>
 * The combination of both instances is used to indicate the end of a
 * phase/partial phase of the simulation. Whenever a message is encountered, a
 * counter value is increased. When the counter reaches the configured
 * threshold, the {@link ControlCounter} calls the method
 * <code>endOfTransientPhase()</code> or <code>endOfBatch()</code>, depending on
 * the current state of the simulation.
 * <p>
 * You can use {@link #parse(SimNode, Parameters)} to parse the parameters and
 * create objects of type {@link ControlCounter} and {@link MessageCountMeter}.
 */
public class ControlCountMeter {

   /**
    * Helper method to create something which resembles the
    * <code>ControlCountMeter</code>, which was included in previous versions of
    * this library. Combines a {@link MessageCountMeter} with a
    * {@link ControlCounter}.
    * 
    * @param transientMsgCount
    *           number of messages after which the transient phase ends
    * @param batchMsgCount
    *           number of messages after which a batch ends
    * @return a new meter instance
    */
   public static MessageCountMeter create(int transientMsgCount, int batchMsgCount) {
      return new MessageCountMeter(new ControlCounter(transientMsgCount, batchMsgCount));
   }

   public static final String TRANSIENT_CALLS = "TransientCalls";
   public static final String BATCH_CALLS = "BatchCalls";

   /**
    * use this method to parse the parameters and create a meter and a control
    * counter
    */
   public static MessageCountMeter parse(SimNode ownNode, Parameters pars) {
      final int transientCalls = pars.get(ownNode, TRANSIENT_CALLS).asInteger();
      final int batchCalls = pars.get(ownNode, BATCH_CALLS).asInteger();
      return create(transientCalls, batchCalls);
   }
}
