/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.meters.length;

import ikr.simlib.messages.Message;
import ikr.simlib.meters.OnePointMeter;
import ikr.simlib.ports.Port;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * An message length meter can be attached to a port to measure message lengths.
 * The {@link #useMessage(Message)} method can be overridden to use only those
 * messages for measurement that satisfy certain criteria, e.g., based on
 * message content, or message type.
 */
public class MessageLengthMeter extends OnePointMeter {

   /**
    * creates a new {@link MessageLengthMeter} with a statistic that accepts
    * Long values
    * 
    * @param statistic
    *           statistic to fill with the values
    * @return the new {@link MessageLengthMeter}
    */
   public static MessageLengthMeter createWithLongStatistic(UnweightedSampleCollector<Long> statistic) {
      return new MessageLengthMeter(statistic);
   }

   /**
    * creates a new {@link MessageLengthMeter} with a statistic that accepts
    * Double values. The measured durations are converted to seconds and then
    * filled into the statistic.
    * 
    * @param statistic
    *           statistic to fill with the values
    * @return the new {@link MessageLengthMeter}
    */
   public static MessageLengthMeter createWithDoubleStatistic(final UnweightedSampleCollector<? super Double> statistic) {
      return new MessageLengthMeter(new UnweightedSampleCollector<Long>() {
         @Override
         public void update(Long value) {
            statistic.update(value.doubleValue());
         }
      });
   }

   /** statistic to update with the legnths */
   private final UnweightedSampleCollector<Long> statistic;

   protected MessageLengthMeter(UnweightedSampleCollector<Long> statistic) {
      this.statistic = statistic;
   }

   /**
    * called when a message was seen
    * 
    * @param msg
    *           the message
    * @param length
    *           the length of the message
    */
   protected void evaluateSample(Message msg, long length) {
      statistic.update(length);
   }

   /**
    * filters messages that should be used for measurement
    * 
    * @param msg
    *           the observed message
    */
   protected boolean useMessage(Message msg) {
      return true;
   }

   @Override
   protected void evaluateMessage(Message msg, Port port) {
      if (useMessage(msg)) {
         evaluateSample(msg, msg.getLength());
      }
   }

}
