/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.model;

import ikr.simlib.parameters.NamingConstants;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.Printable;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.regex.Pattern;

/**
 * <h1>StdSimNode</h1> The default implementation of a
 * {@link ikr.simlib.model.SimNode} is called <code>StdSimNode</code>, which
 * already implements the methods listed above.
 * 
 * The interface {@link ikr.simlib.model.SimNode} has the following set of
 * methods:
 * <ul>
 * <li>building the model hierarchy:
 * <ul>
 * <li>{@link #addChild(SimNode)} adds a child node to an existing node
 * <li>{@link #getChildren()} returns all child nodes of this node
 * <li>{@link #getParent()} returns the parent node
 * </ul>
 * <li>naming nodes:
 * <ul>
 * <li>{@link #getName()} returns the name of this node, only
 * <li>{@link #getFullName()} returns the fully qualified name of this node
 * </ul>
 * <li>exporting, importing, and printing results:
 * <ul>
 * <li>{@link #getPrintHandler()} return a handler for printing results
 * <li>{@link #getExportHandler()} returns a handler for importing and exporting
 * results
 * </ul>
 * </ul>
 */
public class StdSimNode implements SimNode {

   /**
    * This pattern defines the rules for node names.
    */
   public static final Pattern NAME_PATTERN = Pattern.compile("[A-Za-z_][A-Za-z_0-9]*");

   /**
    * Static method called only one time by {@link Model} to create the root
    * node.
    */
   public static SimNode createRootNode(String name) {
      return new StdSimNode(name, null);
   }

   /**
    * The name of the node
    */
   private final String name;

   /**
    * The reference to the owning node
    */
   private final SimNode parent;

   /**
    * The list of child SimNodes
    */
   private final LinkedHashMap<String, SimNode> children;

   /**
    * 
    */
   private Printable printHandler;

   /**
    * 
    */
   private BatchExportable exportHandler;

   /**
    * Creates a new model node.
    * 
    * Set to <code>private</code> since SimLib "node creation change". Use
    * {@link #createChildNode(String)} to create nodes instead.
    * 
    * @param name
    *           The name of the new node. It must be unique for nodes having the
    *           same owner.
    * @param parentNode
    *           The owner of this node. A <code>null</code> value indicates the
    *           creation of the root node.
    */
   private StdSimNode(String name, SimNode parentNode) {
      this.parent = parentNode;

      // no handlers by default
      setPrintHandler(null);
      setExportHandler(null);

      // check name
      if (!NAME_PATTERN.matcher(name).matches())
         throw new ModelException("Illegal model node name. Names must begin with a letter or underscore and can also contain digits in later characters.");
      this.name = name;

      children = new LinkedHashMap<String, SimNode>();

      if (parentNode != null)
         parentNode.addChild(this);
   }

   /**
    * @return the current owner or null if this is a top-level PrintServer.
    */
   @Override
   public SimNode getParent() {
      return parent;
   }

   /**
    * @return the local name
    */
   @Override
   public final String getName() {
      return name;
   }

   /**
    * @return the full name
    */
   @Override
   public final String getFullName() {
      if (getParent() != null) {
         return getParent().getFullName() + NamingConstants.DELIMITER + getName();
      } else {
         return getName();
      }
   }

   /**
    * Adds a new PrintServer as a child ModelNode.
    */
   @Override
   public void addChild(SimNode newNode) {
      // check if node name exists
      if (children.containsKey(newNode.getName()))
         throw new ModelException("Node names must be unique. A node with name \"" + newNode.getName() + "\" already exists.");

      children.put(newNode.getName(), newNode);
   }

   /**
    * Returns a read-only list of this print servers children.
    */
   @Override
   public List<SimNode> getChildren() {
      return new ArrayList<SimNode>(this.children.values());
   }

   @Override
   public void setPrintHandler(Printable printHandler) {
      this.printHandler = printHandler;
   }

   @Override
   public Printable getPrintHandler() {
      return printHandler;
   }

   @Override
   public void setExportHandler(BatchExportable exportHandler) {
      this.exportHandler = exportHandler;
   }

   @Override
   public BatchExportable getExportHandler() {
      return exportHandler;
   }

   @Override
   public String toString() {
      return getFullName();
   }

   /**
    * Creates and returns a child node with name <i>n</i>
    * 
    * @param n
    *           The name of the new node. It must be unique for nodes having the
    *           same owner.
    */
   @Override
   public SimNode createChildNode(String n) {
      return new StdSimNode(n, this);
   }

   @Override
   public boolean subtreeHasPrintHandler() {
      if (printHandler != null)
         return true;
      for (SimNode child : children.values())
         if (child.subtreeHasPrintHandler())
            return true;
      return false;
   }

   @Override
   public boolean subtreeHasExportHandler() {
      if (exportHandler != null)
         return true;
      for (SimNode child : children.values())
         if (child.subtreeHasExportHandler())
            return true;
      return false;
   }
}
