/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.parameters.reflection;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.Name;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.ElementScanner7;
import javax.lang.model.util.SimpleTypeVisitor6;
import javax.lang.model.util.Types;
import javax.tools.Diagnostic.Kind;

/**
 * Checks for correct implementation of static createInstance() methods as
 * defined in {@link ReflectionConstructable}.
 */
@SupportedAnnotationTypes("*")
@SupportedSourceVersion(SourceVersion.RELEASE_7)
public class ReflectionConstructableProcessor extends AbstractProcessor {
   private MyScanner scanner;

   @Override
   public synchronized void init(ProcessingEnvironment pe) {
      super.init(pe);
      scanner = new MyScanner(pe);
   }

   @Override
   public boolean process(Set<? extends TypeElement> annotations, RoundEnvironment roundEnv) {
      // process all non-abstract classes
      for (Element e : roundEnv.getRootElements())
         if (e.getKind() == ElementKind.CLASS && !e.getModifiers().contains(Modifier.ABSTRACT))
            processClass(e);
      return false; // continue processing
   }

   /**
    * analyze the class and print errors if requirements are not met
    */
   private void processClass(Element e) {
      final ClassInformation ci = new ClassInformation(processingEnv);
      scanner.scan(e, ci);
      for (EnforcedStaticCreateMethod failed : ci.getFailedRequirements())
         processingEnv.getMessager().printMessage(Kind.ERROR, "missing static " + ReflectionConstructable.CREATE_INSTANCE_METHOD_NAME + "() method enforced by " + failed.toString(), failed.theClass);
   }

   private static class MyScanner extends ElementScanner7<Void, ClassInformation> {

      private final Set<String> reflectionConstructableNames;
      private final ProcessingEnvironment processingEnv;

      public MyScanner(ProcessingEnvironment processingEnv) {
         this.processingEnv = processingEnv;
         this.reflectionConstructableNames = new HashSet<String>();
         this.reflectionConstructableNames.add(ReflectionConstructable.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable1.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable2.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable3.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable4.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable5.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable6.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable7.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable8.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable9.class.getName());
         this.reflectionConstructableNames.add(ReflectionConstructable10.class.getName());
      }

      @Override
      public Void visitExecutable(ExecutableElement e, ClassInformation ci) {
         // check for a static method which creates objects of this class;
         // extracts all parameters
         if (e.getKind() == ElementKind.METHOD && e.getModifiers().contains(Modifier.STATIC) && e.getSimpleName().toString().equals(ReflectionConstructable.CREATE_INSTANCE_METHOD_NAME)) {
            final List<TypeMirror> types = extractParameterTypes(e);
            ci.implementedStaticMethods.add(new ImplementedStaticCreateMethod(types, e.getReturnType()));
         }
         return super.visitExecutable(e, ci);
      }

      private static List<TypeMirror> extractParameterTypes(ExecutableElement e) {
         final List<TypeMirror> types = new ArrayList<TypeMirror>();
         for (VariableElement elem : e.getParameters())
            types.add(elem.asType());
         return types;
      }

      @Override
      public Void visitType(TypeElement e, ClassInformation ci) {
         // called for class definitions
         if (e.getKind() == ElementKind.CLASS) {
            // get all implemented interfaces and super classes
            final Set<TypeElement> superClasses = new HashSet<TypeElement>();
            collectSuperClassesAndInterfaces(e, superClasses);
            // check which of them directly implements the
            // ReflectionConstructable interface
            for (TypeElement clazz : superClasses) {
               if (!this.reflectionConstructableNames.contains(clazz.getQualifiedName().toString())) {
                  for (TypeMirror iFace : clazz.getInterfaces()) {
                     final Name ifaceName = ((TypeElement) processingEnv.getTypeUtils().asElement(iFace)).getQualifiedName();
                     if (this.reflectionConstructableNames.contains(ifaceName.toString())) {
                        // get the generic type arguments of the interface
                        // implements statement
                        final List<TypeMirror> types = new ArrayList<TypeMirror>();
                        iFace.accept(new InterfaceArgumentTypeVisitor(), types);
                        // add this signature to the list of enforced
                        // constructors
                        ci.enforcedStaticMethods.add(new EnforcedStaticCreateMethod(types, clazz, e));
                     }
                  }
               }
            }
         }
         return super.visitType(e, ci);
      }
   }

   private static class EnforcedStaticCreateMethod {
      final List<TypeMirror> parameters;
      final TypeElement annotatedClass;
      final TypeElement theClass;

      private EnforcedStaticCreateMethod(List<TypeMirror> parameters, TypeElement annotatedClass, TypeElement theClass) {
         this.parameters = parameters;
         this.annotatedClass = annotatedClass;
         this.theClass = theClass;
      }

      @Override
      public String toString() {
         return annotatedClass + " (parameters " + parameters + ")";
      }
   }

   private static class ImplementedStaticCreateMethod {
      final List<TypeMirror> parameters;
      final TypeMirror returnType;

      private ImplementedStaticCreateMethod(List<TypeMirror> parameters, TypeMirror returnType) {
         this.parameters = parameters;
         this.returnType = returnType;
      }
   }

   private static class ClassInformation {
      final List<ImplementedStaticCreateMethod> implementedStaticMethods = new ArrayList<ImplementedStaticCreateMethod>();
      final List<EnforcedStaticCreateMethod> enforcedStaticMethods = new ArrayList<EnforcedStaticCreateMethod>();
      final Types typeUtils;

      public ClassInformation(ProcessingEnvironment processingEnv) {
         this.typeUtils = processingEnv.getTypeUtils();
      }

      List<EnforcedStaticCreateMethod> getFailedRequirements() {
         final List<EnforcedStaticCreateMethod> retval = new ArrayList<EnforcedStaticCreateMethod>();
         for (EnforcedStaticCreateMethod enf : enforcedStaticMethods) {
            boolean found = false;
            for (ImplementedStaticCreateMethod impl : implementedStaticMethods) {
               if (typeUtils.isAssignable(impl.returnType, enf.annotatedClass.asType()) && enf.parameters.size() == impl.parameters.size()) {
                  boolean paramsMatch = true;
                  for (int i = 0; i < enf.parameters.size(); i++)
                     if (!typeUtils.isAssignable(enf.parameters.get(i), impl.parameters.get(i)))
                        paramsMatch = false;
                  if (paramsMatch)
                     found = true;
               }
            }
            if (!found)
               retval.add(enf);
         }
         return retval;
      }
   }

   /**
    * collect all interfaces and super classes implemented by teh given type
    * 
    * @param type
    * @param superClasses
    */
   private static void collectSuperClassesAndInterfaces(TypeElement type, Set<TypeElement> superClasses) {
      if (superClasses.add(type)) {
         type.getSuperclass().accept(new CollectSuperClassesTypeVisitor(), superClasses);
         for (TypeMirror iFace : type.getInterfaces())
            iFace.accept(new CollectSuperClassesTypeVisitor(), superClasses);
      }
   }

   private static class CollectSuperClassesTypeVisitor extends SimpleTypeVisitor6<Void, Set<TypeElement>> {
      @Override
      public Void visitDeclared(DeclaredType t, Set<TypeElement> superClasses) {
         collectSuperClassesAndInterfaces((TypeElement) t.asElement(), superClasses);
         return super.visitDeclared(t, superClasses);
      }
   }

   private static class InterfaceArgumentTypeVisitor extends SimpleTypeVisitor6<Void, List<TypeMirror>> {
      @Override
      public Void visitDeclared(DeclaredType t, List<TypeMirror> p) {
         p.addAll(t.getTypeArguments());
         return super.visitDeclared(t, p);
      }
   }

}
