/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.ports.input;

import ikr.simlib.entities.Entity;
import ikr.simlib.messages.Message;

/**
 * Convenience class to use instead of manually deriving from {@link InputPort}.
 * <p>
 * Allows to specify a callback function which is executed whenever message
 * indication is signalled from the connected output port. Three interfaces
 * support different method signatures to make usage with java 1.8 closure
 * syntax more convenient.
 */
public class CallbackInputPort extends InputPort {

   /** callback interface with port and message arguments */
   public interface MessageIndicationHandler {
      void handleMessageIndication(InputPort port, Message msg);
   }

   /** callback interface with message argument */
   public interface MessageIndicationHandlerMessage {
      void handleMessageIndication(Message msg);
   }

   /** callback interface with no arguments */
   public interface MessageIndicationHandlerNoArgs {
      void handleMessageIndication();
   }

   private final MessageIndicationHandler msgIndicationhandler;

   public CallbackInputPort(Entity owner, String portName, MessageIndicationHandler msgIndicationhandler) {
      super(owner, portName);
      this.msgIndicationhandler = msgIndicationhandler;
   }

   public CallbackInputPort(Entity owner, MessageIndicationHandler msgIndicationhandler) {
      this(owner, InputPort.INPUT_PORT_NAME, msgIndicationhandler);
   }

   public CallbackInputPort(Entity owner, String portName, final MessageIndicationHandlerMessage msgIndicationhandler) {
      this(owner, portName, new MessageIndicationHandler() {
         @Override
         public void handleMessageIndication(InputPort port, Message msg) {
            msgIndicationhandler.handleMessageIndication(msg);
         }
      });
   }

   public CallbackInputPort(Entity owner, MessageIndicationHandlerMessage msgIndicationhandler) {
      this(owner, InputPort.INPUT_PORT_NAME, msgIndicationhandler);
   }

   public CallbackInputPort(Entity owner, String portName, final MessageIndicationHandlerNoArgs msgIndicationhandler) {
      this(owner, portName, new MessageIndicationHandler() {
         @Override
         public void handleMessageIndication(InputPort port, Message msg) {
            msgIndicationhandler.handleMessageIndication();
         }
      });
   }

   public CallbackInputPort(Entity owner, MessageIndicationHandlerNoArgs msgIndicationhandler) {
      this(owner, InputPort.INPUT_PORT_NAME, msgIndicationhandler);
   }

   @Override
   protected void handleMessageIndication(Message msg) {
      msgIndicationhandler.handleMessageIndication(this, msg);
   }

}
