/*
 * Copyright (C) 2007-2014 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

/**
 * Ports are the main abstraction used for communication between entities. Ports
 * use a handshake protocol to exchange messages. Ports of different entities or
 * ports belonging to the same entity may be connected using the <code>
   connect()
</code> -method
 * of the <code>
   Entity
</code> class. Ports always belong to one entity. Input and output
 * ports are only conceptionally different to make type checking easier. They
 * have the same functionality.
 * 
 * <p>
 * Every port may have several message filters attached to it. They may be used
 * to filter messages, or to get information about messages, e.g. for counting
 * or statistics. <br>
 * A port may also have one input message handler and one output message
 * handler. Message handlers are the interface between ports and entities.
 * Whenever a message has to be sent, the corresponding message handler is
 * called by the port.
 * </p>
 * 
 * <p>
 * The following protocol is used to exchange messages: If an entity has
 * prepared a new message to send, it calls the <code>
      messageIndication()
   </code>-method of its output
 * port. The message then travels along the connected ports to the destination
 * input port. Finally, the <code>
      handleMsgIndication()
   </code>-method of the input message handler is
 * responsible of handling the message. If it can handle the message, it
 * immediately calls the <code>
      getMessage()
   </code>-method of the input port, which passes the call
 * to the output handler of the source port. The <code>
      handleGetMessage()
   </code> -method of this port
 * must then supply the message (active sender). <br>
 * If <code>
      getMessage()
   </code> is not called, the sender is blocked. A receiver port may call
 * <code>
      isMessageAvailable()
   </code> which is passed from the input port to the corresponding output port.
 * There, the <code>
      handleIsMessageAvailable()
   </code> -method of the output message handler is called (active
 * receiver). <code>
      getMessage()
   </code> may only be called within <code>
      handleMsgIndication()
   </code> or after a successful
 * <code>
      isMessageAvaialbale()
   </code> -call! <br>
 * The port registers itself automatically at the port manager.
 * </p>
 * 
 * <p>
 * The difference between input and output ports is implemented by the
 * introduction of two special port classes
 * {@link ikr.simlib.ports.input.InputPort} and
 * {@link ikr.simlib.ports.output.OutputPort}.
 * </p>
 * <br>
 * <br>
 * <h2>Basic Characteristics of Ports</h2> All ports are derived from the same
 * base class <code>
   Port
</code>. There are sending ports (base class
 * {@link ikr.simlib.ports.output.OutputPort}) and receiving ports (base class
 * {@link ikr.simlib.ports.input.InputPort}). This is identified by a field
 * within <code>
   Port
</code> , which can be accessed with <code>
   getPortType()
</code> . <br>
 * Each port belongs to a certain model component, which must be declared upon
 * initialization as a parameter (owner). In addition, each port has a local
 * name (declared upon constructor call), which must be unique within the model
 * component and together with the global name of the superjacent model
 * component composes the global port name. <br>
 * <br>
 * <h2>Connecting Ports</h2> Two model components are connected by creating a
 * directed connection between two corresponding ports which are to be used for
 * their communication. There are several ways of accomplishing this: <br>
 * <br>
 * <ul>
 * <li>
 * Connection of an output port from model component A with an input port of
 * model component B by calling the <code>
         connect()
      </code> method from A or B and while
 * declaring the <code>
         Entity
      </code> entities and their corresponding port names. The
 * following is valid:
 * 
 * <ul>
 * <li>
 * A may not be owner of B.
 * 
 * <li>
 * B may not be owner of A.
 * </ul>
 * 
 * <li>
 * A and B may not be identical (feedback).
 * 
 * <li>
 * Connection of an output port model component A with an output port of the
 * subjacent model component B by calling <code>
         connect()
      </code>
 * 
 * <li>
 * Connection of an input port from model component A with an input port of the
 * superjacent model component B by calling <code>
         connect()
      </code> .
 * <li>
 * As an alternative in both previous cases it is also possible to make the port
 * of the subjacent model component addressable for the superjacent model
 * component by using the Entity method <code>
         aliasPort()
      </code> , so that the port of the
 * superjacent model component becomes superfluous.
 * <li>
 * Each port can have a maximum of one incoming and one outgoing connection, so
 * it is not possible to connect a model component via one output port with
 * multiple input ports of other model components. <br>
 * <br>
 * <li>
 * If an incoming connection exists, the address of the preceding port (in the
 * private part) is stored as a data element. If an outgoing connection exists
 * the address of the succeeding port is stored. Both occur when connecting with
 * <code>
         connect()
      </code> .
 * </ul>
 * 
 * <h2>Communication between Ports</h2> A handshake protocol has been defined
 * for the exchange of messages between two model components via ports, which
 * has been realized with three functions: <br>
 * <br>
 * <ul>
 * <li>
 * <code>
         messageIndication()
      </code> notifies the receiver, that a message is waiting at the sender.
 * <li>
 * <code>
         isMessageAvailable()
      </code> can be called from the receiver side to inquire if the sender is
 * holding a message.
 * <li>
 * <code>
         getMessage()
      </code> is called from the receiver side to pick up a waiting message from
 * the sender.
 * </ul>
 * <b>Sender offers message ("active sender")</b>
 * <ol>
 * <li>
 * Sender entity calls <code>
         messageIndication()
      </code> of the output port.
 * <li>
 * Output port calls <code>
         messageIndication()
      </code> of the successor port (= input port of the receiver
 * entity).
 * <li>
 * Input port calls <code>
         handleMessageIndication()
      </code> of its message handler because there is no successor
 * port.
 * <li>
 * <code>
         InputMessageHandler
      </code> calls a handler function of the receiver entity (e.g., <code>
         handleMessageIndication()
      </code> ) or
 * reacts directly.
 * <li>
 * Receiver entity or <code>
         InputMessageHandler
      </code> decides if the message can be retrieved at the
 * current time and if necessary calls <code>
         getMessage()
      </code> of the input port.
 * <li>
 * Input port calls <code>
         getMessage()
      </code> of the predecessor port (= output port of the sender
 * entity).
 * <li>
 * Output port calls <code>
         handleGetMessage()
      </code> of its message handler, because there is no
 * predecessor port.
 * <li>
 * Output port calls either a handler function of the sender entity (e.g.,
 * <code>
         handleGetMessage())
      </code> or reacts directly.
 * </ol>
 * <center>
 * <img src="{@docRoot}/ikr/simlib/doc-files/messageExchangeP79.svg" width="600" alt="Simplified
 * illustration of events during message exchange"> <br>
 * Simplified illustration of events during message exchange </center></td> <br>
 * <br>
 * A reference to the sent message finally lands as a return value of the
 * original <code>
   getMessage()
</code> call of the receiver. To be exact, calls made by the <code>
   messageIndication()
</code>
 * and <code>
   getMessage()
</code> methods of a port are not directly delivered to the succeeding or
 * preceding port (steps 2 and 6) or the corresponding message handler (steps 3
 * and 7). In fact the handler methods of any connected message filters are
 * called first and then the <code>
   TPort
</code> method <code>
   handleMessageIndication()
</code> or <code>
   handleGetMessage()
</code> are called which
 * take care of delivery. <br>
 * <br>
 * <b>Receiver inquires about the next message ("active receiver")</b>
 * <ol>
 * <li>
 * Receiver entity calls <code>
         isMessageAvailable()
      </code> of the input port.
 * <li>
 * Input port calls <code>
         isMessageAvailable()
      </code> of the output port of the sender entity.
 * <li>
 * Output port calls <code>
         handleIsMessageAvailable()
      </code> .
 * <li>
 * Output port calls either a handler function of the sender entity (e.g.,
 * <code>
         handleIsMessageAvailable()
      </code> ) or reacts directly. The return value announces if a message is
 * available or not and finally lands at the receiver entity.
 * <li>
 * The receiver entity generally calls <code>
         getMessage()
      </code> of the input port if true is
 * returned.
 * <li>
 * The input port calls <code>
         getMessage()
      </code> of the output port of the sender entity.
 * <li>
 * The output port calls <code>
         handleGetMessage()
      </code> of its message handler because it has no
 * preceding port.
 * <li>
 * Output portcalls either a handler function of the sender entity (e.g., <code>
         handleGetMessage()
      </code>
 * ) or reacts directly.
 * </ol>
 * A reference to the sent message finally lands as a return value of the
 * original <code>
   getMessage()
</code> call by the receiver. The extension and refining of
 * sequences in reference to message filters and the port's own <code>
   handle<*>()
</code> methods
 * as in the "active sender" scenario. <br>
 * <br>
 * <h2>Special Ports</h2> There are several classes that have been derived from
 * {@link ikr.simlib.ports.input.InputPort} or
 * {@link ikr.simlib.ports.output.OutputPort} in the simulation library, that
 * have ports which represent specially defined functions:
 * <ul>
 * <li>
 * The class {@link ikr.simlib.ports.output.NonBlockingOutputPort} serves to
 * model output ports which may not be blocked.
 * <ul>
 * <li>
 * The class is publicly derived from {@link ikr.simlib.ports.output.OutputPort}.
 * <li>
 * The model component which the port belongs to calls the method <code>
               sendMessage()
            </code> in
 * order to deliver a message which causes a <code>
               messageIndication()
            </code> call within the port.
 * <li>
 * The following model component must have retrieved the message at its input
 * port via <code>
               getMessage()
            </code> before the sending model component can send the next message
 * with <code>
               sendMessage()
            </code> .
 * <li>
 * Requests by the receiving model component with <code>
               isMessageAvailable()
            </code> are answered with true
 * until the last message delivered to the output port with <code>
               sendMessage()
            </code> has been
 * picked up.
 * </ul>
 * <li>
 * The same statements made about the
 * {@link ikr.simlib.ports.output.NonBlockingOutputPort} are valid for the class
 * {@link ikr.simlib.ports.output.SynchronousOutputPort} with the single
 * exception that messages sent by <code>
         sendMessage()
      </code> and announced to the following model
 * component with <code>
         messageIndication()
      </code> must be directly picked up i.e., the <code>
         handleMessageIndication()
      </code> method
 * (or the corresponding method with this functionality) within the message
 * handler or the model component on the receiver side must call <code>
         getMessage()
      </code> .
 * <li>
 * The class {@link ikr.simlib.ports.input.PolledInputPort} is publicly derived
 * from {@link ikr.simlib.ports.input.InputPort}. Its main characteristic is
 * that it ignores message offers by <code>
         messageIndication()
      </code> , so that a message exchange must
 * always be initiated by the receiver ("active receiver").
 */
package ikr.simlib.ports;