/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result;

import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;

/**
 * Classes supporting batch parallelization must implement this interface to
 * support import and export of batch states.
 * 
 * <p>
 * A fair amount of freedom is given to implementors in how the results are
 * written to the export files. The only constraints are, that
 * </p>
 * <ol>
 * <li>Results have to be ex-/imported in the same order</li>
 * <li>and neither more or less data may imported than has been exported (and
 * vice versa).
 * <li>The rules of the {@link ResultWriter} and {@link ResultReader} interfaces
 * have to be followed (especially matching start and end tags).</li>
 * </ol>
 * 
 */
public interface BatchExportable {

   /**
    * Exports the batch state to a result writer. Everything written to the
    * result writer has to be read in the same order in
    * {@link #importBatch(ResultReader) importBatch} despite the keywords.
    * 
    * @param writer
    *           the result writer which preserves the state
    * @throws PrintException
    *            if any exception occurs while writing a result to the writer
    */
   void exportBatch(ResultWriter writer) throws PrintException;

   /**
    * Imports the state of a batch from the result reader. The exported results
    * have to read in the same order as written in
    * {@link #exportBatch(ResultWriter) exportBatch} despite the keywords.
    * 
    * @param reader
    *           the result reader which contains the batch state
    * @throws ImportException
    *            if any exception occurs while reading a result from the reader
    */
   void importBatch(ResultReader reader) throws ImportException;
}
