/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result;

/**
 * <p>
 * Common result names used by the SimLib statistics are grouped here. The
 * <code>as...</code> methods should be used for custom keywords to receive the
 * SimLib naming convention (unless you want your own ;).
 * </p>
 * 
 * <p>
 * Export keywords should be decorated by {@link ResultKeyword#asExport()} to
 * make them stand out in result files. If you encounter such an keyword in
 * batch or final results this is a bug. <strong>Do not use undecorated names
 * from this class as export keywords.</strong> This is extremely error prone.
 * </p>
 */
public class ResultKeywords {

   public static final ResultKeyword LIST = new ResultKeyword("").asList();

   public static final ResultKeyword MIN = new ResultKeyword("min");
   public static final ResultKeyword BATCH_MIN = MIN.asBatch();
   public static final ResultKeyword MAX = new ResultKeyword("max");
   public static final ResultKeyword BATCH_MAX = MAX.asBatch();

   public static final ResultKeyword MEAN = new ResultKeyword("").asMean();
   public static final ResultKeyword MEAN_CINT = MEAN.asConfidenceInterval();
   public static final ResultKeyword BATCH_MEAN = MEAN.asBatch();

   public static final ResultKeyword MEDIAN = new ResultKeyword("median");
   public static final ResultKeyword BATCH_MEDIAN = MEDIAN.asBatch();

   public static final ResultKeyword DEVIATION = new ResultKeyword("deviation");
   public static final ResultKeyword DEVIATION_CINT = DEVIATION.asConfidenceInterval();
   public static final ResultKeyword BATCH_DEVIATION = DEVIATION.asBatch();

   /**
    * Confidence interval.
    */
   public static final ResultKeyword CINT = new ResultKeyword("").asConfidenceInterval();

   /**
    * Upper bound.
    */
   public static final ResultKeyword UBOUND = new ResultKeyword("ubound");
   public static final ResultKeyword BATCH_UBOUND = UBOUND.asBatch();

   /**
    * Lower bound.
    */
   public static final ResultKeyword LBOUND = new ResultKeyword("lbound");

   /**
    * Coefficient of variation.
    */
   public static final ResultKeyword COV = new ResultKeyword("cov");
   public static final ResultKeyword COV_MEAN = COV.asMean();
   public static final ResultKeyword COV_CINT = COV.asConfidenceInterval();
   public static final ResultKeyword COV_LIST = COV.asList();
   public static final ResultKeyword BATCH_COV = COV.asBatch();
   public static final ResultKeyword BATCH_COV_LIST = BATCH_COV.asList();

   /**
    * Coefficient.
    */
   public static final ResultKeyword COEFF = new ResultKeyword("coeff");
   public static final ResultKeyword COEFF_CINT = COEFF.asConfidenceInterval();
   public static final ResultKeyword BATCH_COEFF = COEFF.asBatch();

   public static final ResultKeyword PROBABILITY = new ResultKeyword("prob");
   public static final ResultKeyword PROBABILITY_CINT = PROBABILITY.asConfidenceInterval();
   public static final ResultKeyword BATCH_PROBABILITY = PROBABILITY.asBatch();

   public static final ResultKeyword DISTRIBUTION = new ResultKeyword("dist");
   public static final ResultKeyword DISTRIBUTION_CINT = DISTRIBUTION.asConfidenceInterval();
   public static final ResultKeyword BATCH_DISTRIBUTION = DISTRIBUTION.asBatch();

   /**
    * Complementary distribution.
    */
   public static final ResultKeyword CDIST = new ResultKeyword("cdist");

   public static final ResultKeyword VARIANCE = new ResultKeyword("variance");
   public static final ResultKeyword VARIANCE_LIST = VARIANCE.asList();
   public static final ResultKeyword BATCH_VARIANCE = VARIANCE.asBatch();
   public static final ResultKeyword BATCH_VARIANCE_LIST = BATCH_VARIANCE.asList();

   public static final ResultKeyword QUANTILE = new ResultKeyword("quantile");

   /**
    * Probability density function.
    */
   public static final ResultKeyword PDF = new ResultKeyword("pdf");
   public static final ResultKeyword BATCH_PDF = PDF.asBatch();

   public static final ResultKeyword AREA = new ResultKeyword("area");
   public static final ResultKeyword BATCH_AREA = AREA.asBatch();

   public static final ResultKeyword NUMBER_OF_EVENTS = new ResultKeyword("numberofevents");
   public static final ResultKeyword BATCH_NUMBER_OF_EVENTS = NUMBER_OF_EVENTS.asBatch();

   public static final ResultKeyword NUMBER_OF_BATCHES = new ResultKeyword("numberofbatches");

   // TODO: This is often synonymous to NUMBER_OF_EVENTS, especially when
   // summations provide this result (usually done through getNumberOfSamples).
   // In cases where indeed samples are counted, but NOE used, we should change
   // it.
   public static final ResultKeyword NUMBER_OF_SAMPLES = new ResultKeyword("numberofsamples");
   public static final ResultKeyword BATCH_NUMBER_OF_SAMPLES = NUMBER_OF_SAMPLES.asBatch();

   public static final ResultKeyword SUM = new ResultKeyword("sum");
   public static final ResultKeyword BATCH_SUM = SUM.asBatch();

   public static final ResultKeyword RATE = new ResultKeyword("rate");
   public static final ResultKeyword RATE_CINT = RATE.asConfidenceInterval();
   public static final ResultKeyword BATCH_RATE = RATE.asBatch();

   public static final ResultKeyword SAMPLE_RATE = new ResultKeyword("samplerate");
   public static final ResultKeyword SAMPLE_RATE_CINT = SAMPLE_RATE.asConfidenceInterval();
   public static final ResultKeyword BATCH_SAMPLE_RATE = SAMPLE_RATE.asBatch();

   public static final ResultKeyword UNDERFLOW_PROB = new ResultKeyword("uprob");
   public static final ResultKeyword UNDERFLOW_PROB_CINT = UNDERFLOW_PROB.asConfidenceInterval();
   public static final ResultKeyword BATCH_UNDERFLOW_PROB = UNDERFLOW_PROB.asBatch();

   public static final ResultKeyword OVERFLOW_PROB = new ResultKeyword("oprob");
   public static final ResultKeyword OVERFLOW_PROB_CINT = OVERFLOW_PROB.asConfidenceInterval();
   public static final ResultKeyword BATCH_OVERFLOW_PROB = OVERFLOW_PROB.asBatch();

   public static final ResultKeyword BLOCKSIZE = new ResultKeyword("blocksize");
   public static final ResultKeyword BATCH_BLOCKSIZE = BLOCKSIZE.asBatch();

   public static final ResultKeyword FI = new ResultKeyword("fi");
   // FIXME: Guess we could use FI instead of this
   public static final ResultKeyword FI_MEAN = FI.asMean();
   public static final ResultKeyword FI_CINT = FI.asConfidenceInterval();
   public static final ResultKeyword FI_LIST = FI.asList();

   public static final ResultKeyword BUCKET = new ResultKeyword("bucket");
   public static final ResultKeyword BUCKET_MEAN = BUCKET.asMean();
   public static final ResultKeyword BATCH_BUCKET_MEAN = BUCKET_MEAN.asBatch();

   public static final ResultKeyword NORM_VAR = new ResultKeyword("normvar");
   public static final ResultKeyword NORM_VAR_LIST = NORM_VAR.asList();

   public static final ResultKeyword UNIT = new ResultKeyword("unit");
}
