/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result;

import ikr.simlib.result.exceptions.ImportException;

/**
 * The result input interface is used to input results in a serial way. This is
 * the most basic way to read results.
 * 
 * <p>
 * The result tree is traversed through a set of events by calling
 * {@link #next()}. Each event can trigger a state change which changes the set
 * of methods that can be called.
 * </p>
 * 
 * <p>
 * <b>Available methods in a certain state</b>
 * <table border="1">
 * <tr>
 * <th>Event</th>
 * <th>Available methods</th>
 * </tr>
 * </thead>
 * <tr>
 * <td><code>START_RESULTS</code></td>
 * <td>next, hasNext</td>
 * </tr>
 * <tr>
 * <td><code>END_RESULTS</code></td>
 * <td>hasNext</td>
 * </tr>
 * <tr>
 * <td><code>START_NODE</code></td>
 * <td>next, hasNext, getName, getTypeName</td>
 * </tr>
 * <tr>
 * <td><code>END_NODE</code></td>
 * <td>next, hasNext</td>
 * </tr>
 * <tr>
 * <td><code>RESULT</code></td>
 * <td>next, hasNext, getName, getTypeName, getValue, getInt | getLong |
 * getDouble | getString (depends on the result type)</td>
 * </tr>
 * <tr>
 * <td><code>START_ARRAY</code></td>
 * <td>next, hasNext, getName, getTypeName</td>
 * </tr>
 * <tr>
 * <td><code>END_ARRAY</code></td>
 * <td>next, hasNext</td>
 * </tr>
 * <td><code>START_ELEMENT</code></td>
 * <td>next, hasNext, getName, getTypeName, getIndex</td>
 * </tr>
 * <tr>
 * <td><code>END_ELEMENT</code></td>
 * <td>next, hasNext</td>
 * </tr>
 * </table>
 * </p>
 * 
 */
public interface ResultReader {

   /**
    * Read events which occur in a result file.
    */
   enum Event {
      /**
       * The result tree root has been read.
       */
      START_RESULTS,

      /**
       * The end of the result tree has been read. No events remain after this
       * one has been read.
       */
      END_RESULTS,

      /**
       * A result node has been read.
       */
      START_NODE,

      /**
       * The end of a result node has been read.
       */
      END_NODE,

      /**
       * A result has been read.
       */
      RESULT,

      /**
       * A result array has been read.
       */
      START_ARRAY,

      /**
       * The end of an array has been read.
       */
      END_ARRAY,

      /**
       * A result array element has been read.
       */
      START_ELEMENT,

      /**
       * The end of an array element has been read.
       */
      END_ELEMENT
   }

   /**
    * Closes the reader and underlying streams.
    */
   void close() throws ImportException;

   /**
    * Returns the next result read event.
    * 
    * @return the next read event
    */
   Event next() throws ImportException;

   /**
    * Reads the next result (item, node or array).
    * 
    * @return either {@link Event#RESULT}, {@link Event#START_NODE} or
    *         {@link Event#START_ARRAY}
    */
   Event nextResult() throws ImportException;

   /**
    * Returns whether any read events remain or the result tree has been fully
    * traversed.
    * 
    * <p>
    * Calling {@link #next()} after this mehtod returned {@code false} will
    * result in an exception.
    * </p>
    * 
    * @return whether any read events remain
    */
   boolean hasNext() throws ImportException;

   /**
    * Returns the name of the current result. If the current result has no name
    * {@code null} is returned.
    * 
    * @return the name of the current result or {@code null}
    */
   String getName() throws ImportException;

   /**
    * Returns the type of the current result.
    * 
    * @return the type of the current result
    */
   String getTypeName() throws ImportException;

   /**
    * Returns the index of the current result element.
    * 
    * @return the index of the current element
    */
   long getIndex() throws ImportException;

   /**
    * Reads the unparsed value of the current result from the input.
    * 
    * @return the unparsed value of the current result
    */
   String getValue() throws ImportException;

   /**
    * Reads the start of the result tree.
    */
   void readStartResults() throws ImportException;

   /**
    * Reads the end of the result tree.
    */
   void readEndResults() throws ImportException;

   /**
    * Reads the next result array.
    * 
    * @param expectedName
    *           the expected name
    */
   void readStartArray(String expectedName) throws ImportException;

   /**
    * Reads the end of the current result array.
    */
   void readEndArray() throws ImportException;

   /**
    * Reads the next array element. If the end of the array has been reached -1
    * is returned. If called after the end of the array, an exception will be
    * thrown.
    * 
    * @param expectedIndex
    *           the expected index (to be compared with the index read from the
    *           result file)
    * @return the index of the next element or -1
    */
   long readStartArrayItem(long expectedIndex) throws ImportException;

   /**
    * Reads the end of the current array element.
    */
   void readEndArrayItem() throws ImportException;

   /**
    * Reads the next result node.
    * 
    * @param expectedType
    *           the expected type name
    * @param expectedName
    *           the expected name
    */
   void readStartNode(String expectedType, String expectedName) throws ImportException;

   /**
    * Reads the end of the current result node.
    */
   void readEndNode() throws ImportException;

   /**
    * Returns the current long integer result.
    * 
    * @return the current long integer result
    */
   long getLong() throws ImportException;

   /**
    * Reads the next long integer result from the input. This method is
    * equivalent of calling
    * 
    * <pre>
    * {@literal
    * nextResult();
    * getLong();
    * }
    * </pre>
    * 
    * @return the next long integer result
    */
   long readLong(String name) throws ImportException;

   /**
    * Returns the current double result.
    * 
    * @return the current double result
    */
   double getDouble() throws ImportException;

   /**
    * Reads the next double result from the input. This method is equivalent of
    * calling
    * 
    * <pre>
    * {@literal
    * nextResult();
    * getDouble();
    * }
    * </pre>
    * 
    * @return the next double result
    */
   double readDouble(String name) throws ImportException;

   /**
    * Returns the current string result.
    * 
    * @return the current string result
    */
   String getString() throws ImportException;

   /**
    * Reads the next string result from the input. This method is equivalent of
    * calling
    * 
    * <pre>
    * {@literal
    * nextResult();
    * getString();
    * }
    * </pre>
    * 
    * @return the next string result
    */
   String readString(String name) throws ImportException;
}
