/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result.xml;

import ikr.simlib.result.PathHelper;
import ikr.simlib.result.ResultFileType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.PrintException;

import java.io.IOException;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.Deque;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

/**
 * A result writer which writes XML to an output stream.
 * 
 */
public class XMLResultWriter implements ResultWriter {

   private enum State {
      /**
       * Waiting to start writing results.
       */
      DOCUMENT,

      /**
       * All results were written.
       */
      FINISHED,

      /**
       * We're at the root of the document.
       */
      AT_ROOT,

      /**
       * Any results are allowed.
       */
      IN_NODE,

      /**
       * Elements can be written.
       */
      IN_ARRAY
   }

   /**
    * The indentation characters. Indent larger than this constant will be
    * truncated.
    */
   private static final char[] INDENT;

   static {
      INDENT = new char[20];
      Arrays.fill(INDENT, ' ');
   }

   private static MathContext ROUNDING_CONTEXT = new MathContext(6);

   /**
    * Set the number of significant digits for double output.
    * 
    * @param digits
    *           the number of digits. Set to null to disable rounding.
    */
   public static void setRoundingDigits(Integer digits) {
      if (digits == null) {
         ROUNDING_CONTEXT = null;
      } else {
         ROUNDING_CONTEXT = new MathContext(digits);
      }
   }

   /**
    * Set the number of significant digits for double output.
    * 
    * @param digits
    *           the number of digits. Set to null to disable rounding.
    * @param mode
    *           see {@link RoundingMode} for possible modes
    */
   public static void setRoundingDigits(Integer digits, RoundingMode mode) {
      if (digits == null) {
         ROUNDING_CONTEXT = null;
      } else {
         ROUNDING_CONTEXT = new MathContext(digits, mode);
      }
   }

   /**
    * Current state of the result writer.
    */
   private State state;

   /**
    * A stack of the result states which resulted to the current state. This is
    * used to switch back to a previous state.
    */
   private final Deque<State> stateStack;

   private final OutputStream output;
   private XMLStreamWriter writer;

   private final ResultFileType resultType;
   private final int seedIndex;
   private final int batchNumber;
   private final int totalNumberOfBatches;

   private final String fileName;
   private final PathHelper pathHelper;

   /**
    * Constructs a new XML result writer.
    * 
    * @param output
    *           the output stream for the XML result tree. The stream will not
    *           be closed by this writer
    * @param logNameForErrorMessages
    */
   public XMLResultWriter(OutputStream output, ResultFileType resultType, int seedIndex, int batchNumber, int totalNumberOfBatches, String logNameForErrorMessages) {
      if (output == null) {
         throw new NullPointerException();
      }
      this.output = output;

      state = State.DOCUMENT;
      stateStack = new ArrayDeque<State>();
      this.resultType = resultType;
      this.seedIndex = seedIndex;
      this.batchNumber = batchNumber;
      this.totalNumberOfBatches = totalNumberOfBatches;
      this.fileName = logNameForErrorMessages;
      this.pathHelper = new PathHelper();
   }

   @Override
   public void close() throws PrintException {
      state = State.FINISHED;
      try {
         output.close();
      } catch (IOException ioe) {
         throw new PrintException(fileName, pathHelper.getPathName(), "I/O problem while closing underlying stream", ioe);
      }
   }

   private void assertState(State... expected) throws PrintException {
      boolean isExpected = false;
      for (State e : expected) {
         if (state == e) {
            isExpected = true;
         }
      }
      if (!isExpected) {
         throw new PrintException(fileName, pathHelper.getPathName(), "invalid state - expected states " + Arrays.toString(expected) + ", but state is " + state);
      }
   }

   private void assertNextState(State... expected) throws PrintException {
      boolean isExpected = false;
      for (State e : expected) {
         if (stateStack.peek() == e) {
            isExpected = true;
         }
      }
      if (!isExpected) {
         throw new PrintException(fileName, pathHelper.getPathName(), "invalid next state - expected states " + Arrays.toString(expected) + ", but state is " + stateStack.peek());
      }
   }

   private void writeIndent() throws PrintException {
      writeNewLine();

      try {
         writer.writeCharacters(INDENT, 0, Math.min(INDENT.length, stateStack.size()));
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }
   }

   private void writeNewLine() throws PrintException {
      try {
         writer.writeCharacters("\n");
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }
   }

   private void writeResult(String type, String name, String value) throws PrintException {
      assertState(State.IN_NODE);

      writeIndent();

      try {
         writer.writeEmptyElement(XMLConstants.RESULT);
         writer.writeAttribute(XMLConstants.TYPE_ATTRIBUTE, type);
         writer.writeAttribute(XMLConstants.NAME_ATTRIBUTE, name);
         writer.writeAttribute(XMLConstants.VALUE_ATTRIBUTE, value);
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }
   }

   @Override
   public void writeStartResults() throws PrintException {
      assertState(State.DOCUMENT);

      try {
         writer = XMLOutputFactory.newInstance().createXMLStreamWriter(output, XMLConstants.ENCODING);

         // XML header
         writer.writeStartDocument(XMLConstants.ENCODING, XMLConstants.XMLVERSION);
         writeNewLine();

         // Version
         writer.writeStartElement(XMLConstants.SIMULATION_RESULTS);
         writer.writeAttribute(XMLConstants.VERSION_ATTRIBUTE, XMLConstants.VERSION);

         writer.writeAttribute(XMLConstants.SEED_INDEX_ATTRIBUTE, Integer.toString(seedIndex));

         switch (resultType) {
         case EXPORT:
            writer.writeAttribute(XMLConstants.RESULT_TYPE_ATTRIBUTE, XMLConstants.EXPORT);
            writer.writeAttribute(XMLConstants.BATCH_NUMBER_ATTRIBUTE, Integer.toString(batchNumber));
            writer.writeAttribute(XMLConstants.TOTAL_BATCH_NUMBER_ATTRIBUTE, Integer.toString(totalNumberOfBatches));
            break;
         case BATCH:
            writer.writeAttribute(XMLConstants.RESULT_TYPE_ATTRIBUTE, XMLConstants.BATCH);
            writer.writeAttribute(XMLConstants.BATCH_NUMBER_ATTRIBUTE, Integer.toString(batchNumber));
            writer.writeAttribute(XMLConstants.TOTAL_BATCH_NUMBER_ATTRIBUTE, Integer.toString(totalNumberOfBatches));
            writer.writeAttribute(XMLConstants.ROUNDING_PRECISION, ROUNDING_CONTEXT == null ? "" : Integer.toString(ROUNDING_CONTEXT.getPrecision()));
            break;
         case FINAL:
            writer.writeAttribute(XMLConstants.RESULT_TYPE_ATTRIBUTE, XMLConstants.FINAL);
            writer.writeAttribute(XMLConstants.TOTAL_BATCH_NUMBER_ATTRIBUTE, Integer.toString(totalNumberOfBatches));
            writer.writeAttribute(XMLConstants.ROUNDING_PRECISION, ROUNDING_CONTEXT == null ? "" : Integer.toString(ROUNDING_CONTEXT.getPrecision()));
            break;
         }

      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }

      // After the root node we're finished
      stateStack.push(State.FINISHED);
      state = State.AT_ROOT;
   }

   @Override
   public void writeEndResults() throws PrintException {
      assertState(State.AT_ROOT);

      state = stateStack.pop();

      writeIndent();

      try {
         writer.writeEndElement();
         writeNewLine();

         writer.close();
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      } finally {
         writer = null;
         // output stream will be closed in close()
      }
   }

   @Override
   public void writeDouble(String name, double value) throws PrintException {
      pathHelper.enterSubTree(name);
      String valueAsString = null;
      switch (resultType) {
      case BATCH:
      case FINAL:
         if (ROUNDING_CONTEXT == null || Double.isNaN(value) || Double.isInfinite(value))
            valueAsString = Double.toString(value);
         else
            valueAsString = BigDecimal.valueOf(value).round(ROUNDING_CONTEXT).stripTrailingZeros().toPlainString();
         break;

      case EXPORT:
         valueAsString = Double.toHexString(value);
         break;
      }

      writeResult(XMLConstants.DOUBLE_TYPE, name, valueAsString);
      pathHelper.leaveSubTree();
   }

   @Override
   public void writeEndArray() throws PrintException {
      assertState(State.IN_ARRAY);

      state = stateStack.pop();

      writeIndent();

      try {
         writer.writeEndElement();
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }

      pathHelper.leaveSubTree();
   }

   @Override
   public void writeEndArrayItem() throws PrintException {
      assertState(State.IN_NODE);
      // Nodes used as result items can't be finished with this method
      assertNextState(State.IN_ARRAY);

      state = stateStack.pop();

      writeIndent();

      try {
         writer.writeEndElement();
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }

      pathHelper.leaveSubTree();
   }

   @Override
   public void writeEndNode() throws PrintException {
      assertState(State.IN_NODE);
      // Nodes used as array elements can't be finished with this method
      assertNextState(State.AT_ROOT, State.DOCUMENT, State.IN_NODE, State.FINISHED);

      state = stateStack.pop();

      writeIndent();

      try {
         writer.writeEndElement();
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }

      pathHelper.leaveSubTree();
   }

   @Override
   public void writeLong(String name, long value) throws PrintException {
      pathHelper.enterSubTree(name);
      writeResult(XMLConstants.LONG_TYPE, name, Long.toString(value));
      pathHelper.leaveSubTree();
   }

   @Override
   public void writeStartArray(String name) throws PrintException {
      pathHelper.enterSubTree(name);

      assertState(State.IN_NODE);

      writeIndent();

      try {
         writer.writeStartElement(XMLConstants.ARRAY);
         writer.writeAttribute(XMLConstants.NAME_ATTRIBUTE, name);
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }

      stateStack.push(State.IN_NODE);
      state = State.IN_ARRAY;
   }

   @Override
   public void writeStartArrayItem(long index) throws PrintException {
      pathHelper.enterSubTree(Long.toString(index));

      assertState(State.IN_ARRAY);

      writeIndent();

      assert (index >= 0);

      try {
         writer.writeStartElement(XMLConstants.ARRAY_ITEM);
         writer.writeAttribute(XMLConstants.INDEX_ATTRIBUTE, Long.toString(index));
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }

      stateStack.push(State.IN_ARRAY);

      // Array elements can't be named
      state = State.IN_NODE;
   }

   @Override
   public void writeStartNode(String type, String name) throws PrintException {
      pathHelper.enterSubTree(name);

      assertState(State.IN_NODE, State.AT_ROOT);

      writeIndent();

      try {
         writer.writeStartElement(XMLConstants.NODE);
         if (type != null)
            writer.writeAttribute(XMLConstants.TYPE_ATTRIBUTE, type);
         writer.writeAttribute(XMLConstants.NAME_ATTRIBUTE, name);
      } catch (XMLStreamException xse) {
         throw new PrintException(fileName, pathHelper.getPathName(), xse);
      }

      switch (state) {
      case AT_ROOT:
         stateStack.push(State.AT_ROOT);
         break;

      case IN_NODE:
         stateStack.push(State.IN_NODE);
         break;

      default:
         throw new AssertionError();
      }

      state = State.IN_NODE;
   }

   @Override
   public void writeString(String name, String value) throws PrintException {
      pathHelper.enterSubTree(name);
      writeResult(XMLConstants.STRING_TYPE, name, value);
      pathHelper.leaveSubTree();
   }
}
