/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.boundary;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * <code>
   BoundaryStatistic</code> is an abstract base class derived from
 * {@link ikr.simlib.statistics.Statistic}. It is the base class of
 * {@link ikr.simlib.statistics.boundary.StdBoundaryStatistic} and
 * {@link ikr.simlib.statistics.boundary.JainBoundaryStatistic}. These two
 * classes are used to find the quantile for a given percentage. The class
 * {@link ikr.simlib.statistics.boundary.StdBoundaryStatistic} provides an
 * implementation to calculate quantiles for a given number of samples and
 * percentage. The class
 * {@link ikr.simlib.statistics.boundary.JainBoundaryStatistic} does the same
 * with an alternative method, the P2 algorithm described in
 * "R. JAIN, I. CHLAMTAC: "The P**2 algorithm for dynamic calculation of
 * quantiles and histograms without storing observations." Communications of the
 * ACM, Vol. 28, Oct. 1985, pp. 1076-1085.". Thus, for constructor call,
 * {@link ikr.simlib.statistics.boundary.StdBoundaryStatistic} requires
 * additionally to the percentage (e.g., 0.95) also the expected number of
 * samples. <br>
 * <br>
 * The following functions are available for evaluation:
 * <ul>
 * <li>
 * {@link #getQuantile()} returns the value of the quantile.
 * <li>
 * {@link #getPercentage()} returns the value of the input percentage.
 * <li>
 * {@link #getNoOfSamples()} returns the number of all collected samples.
 * </ul>
 * <p>
 * <strong>Boundary statistics can't be used with batch
 * parallelization.</strong>
 * </p>
 * 
 * <p>
 * Following keywords are defined for printing the results:
 * </p>
 * 
 * <dl>
 * <dt>
 * quantile</dt>
 * <dd>
 * value found to be the quantile</dd>
 * <dt>
 * prob</dt>
 * <dd>
 * input percentage</dd>
 * <dt>
 * numberofevents</dt>
 * <dd>
 * number of events in all batches</dd>
 * </dl>
 * <br>
 * <br>
 * A special case of a quantile is the median, which is the 50%-quantile. The
 * dedicated {@link ikr.simlib.statistics.median.MedianStatistic} has been
 * implemented to measure medians. <br>
 * <br>
 * <center><img src="{@docRoot} /ikr/simlib/doc-files/medianP65.svg" width="600"
 * alt="Accuracy of the median algorithm"> <br>
 * Accuracy of the median algorithm </center></td> <br>
 * The graph in Figure 1 shows the accuracy of the median approximation in
 * comparison to the accurate value when define the statistic with negative
 * exponentially distributed values. A total of 10,000 samples have been used in
 * this test. <br>
 * Tests have shown that a <code>
   MaxBufferSize
</code> of 10-20% results a deviation to the accurate
 * value about 0.01%.
 */
public abstract class BoundaryStatistic extends Statistic implements Printable, UnweightedSampleCollector<Double> {

   public BoundaryStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      if (type == ResultType.FINAL) {
         this.printComments(writer);
         writer.writeDouble(ResultKeywords.QUANTILE, getQuantile());
         writer.writeDouble(ResultKeywords.PROBABILITY, getPercentage());
         writer.writeLong(ResultKeywords.NUMBER_OF_EVENTS, getNoOfSamples());
      }
   }

   public abstract void update(double sample);

   @Override
   public void update(Double sample) {
      update(sample.doubleValue());
   }

   public abstract double getQuantile();

   public abstract double getPercentage();

   public abstract int getNoOfSamples();
}
