/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.boundary;

import ikr.simlib.model.SimNode;

/**
 * <code>JainBoundaryStatistic</code> provides an implementation for for
 * <code>BoundaryStatistic</code> with an alternative method (the P2 Algorithm)
 * to <code>StdBoundaryStatistic</code>. Note that the percentage-parameter in
 * the constructor must be between 0 and 1.
 * 
 * <p>
 * A more detailed description of the algorithms can be found in the student
 * thesis of Luis Lopez (No 1535) or in LDB No 1507:<br>
 * <strong>Raj Jain, Imrich Chlamtac:</strong> <em>The P2 algorithm for dynamic
 * calculation of quantiles and histograms without storing observations</em>,
 * Communications of the ACM, Vol. 28, Iss. 10, October 1985, pp. 1076 - 1085
 * </p>
 */
public class JainBoundaryStatistic extends BoundaryStatistic {

   protected double[] height;
   protected double[] desiredPos;
   protected int[] actualPos;
   protected double[] increment;
   private final double percentage;
   private int noOfSamples;

   public JainBoundaryStatistic(double percentage, SimNode ownNode) {
      super(ownNode);

      height = new double[5];
      desiredPos = new double[5];
      actualPos = new int[5];
      increment = new double[5];

      this.percentage = percentage;

      resetStatistic();
   }

   @Override
   public int getNoOfSamples() {
      return noOfSamples;
   }

   @Override
   public double getPercentage() {
      return percentage;
   }

   @Override
   public double getQuantile() {
      double d = desiredPos[2] - actualPos[2];
      desiredPos[2] = actualPos[2];
      double newheight = height[2] + (d / (actualPos[3] - actualPos[1])) * ((actualPos[2] - actualPos[1] + d) * (height[3] - height[2]) / (actualPos[3] - actualPos[2]) + (actualPos[3] - actualPos[2] - d) * (height[2] - height[1]) / (actualPos[2] - actualPos[1]));

      if ((d > 0 && (newheight > height[1])) || (d < 0 && (newheight < height[3]))) {
         height[2] = newheight;
      } else {
         if (d < 0) {
            height[2] = height[2] + d * (height[2] - height[1]) / (actualPos[2] - actualPos[1]);
         } else {
            height[2] = height[2] + d * (height[3] - height[3]) / (actualPos[3] - actualPos[2]);
         }
      }

      if ((percentage > 0.5) && (desiredPos[4] > 1 + 2 / (1 - percentage))) {
         return height[2];
      } else if ((percentage < 0.5) && (desiredPos[4] > 1 + 2 / percentage)) {
         return height[2];
      } else {
         return Double.NaN;
      }
   }

   @Override
   public void update(double sample) {
      int i, k = 0;

      noOfSamples++;

      // store and sort first five samples
      if (noOfSamples <= 5) {
         for (i = noOfSamples - 2; i >= 0; i--) {
            if (sample > height[i]) {
               break;
            } else {
               height[i + 1] = height[i];
            }
         }
         height[i + 1] = sample;
         return;
      }

      // begin updating of markers for new value

      // find markers greater than sample or update minimum/maximum

      if (sample < height[0]) {
         height[0] = sample;
         k = 0;
      } else if (sample >= height[0] && sample <= height[1]) {
         k = 0;
      } else if (sample >= height[1] && sample <= height[2]) {
         k = 1;
      } else if (sample >= height[2] && sample <= height[3]) {
         k = 2;
      } else if (sample >= height[3] && sample <= height[4]) {
         k = 3;
      } else if (sample > height[4]) {
         height[4] = sample;
         k = 3;
      }

      // increment markers positions
      for (i = k + 1; i <= 4; i++) {
         actualPos[i] += 1;
      }

      // update desired positions
      for (i = 0; i <= 4; i++) {
         desiredPos[i] += increment[i];
      }

      // adjust heights of markers 2-4 if necessary
      for (i = 1; i <= 3; i++) {
         double d = desiredPos[i] - actualPos[i];
         if ((d >= 1 && (actualPos[i + 1] - actualPos[i] > 1)) || (d <= -1 && (actualPos[i - 1] - actualPos[i] < -1))) {
            d = (d >= 0) ? 1.0 : -1.0;

            double newheight = height[i] + (d / (actualPos[i + 1] - actualPos[i - 1])) * ((actualPos[i] - actualPos[i - 1] + d) * (height[i + 1] - height[i]) / (actualPos[i + 1] - actualPos[i]) + (actualPos[i + 1] - actualPos[i] - d) * (height[i] - height[i - 1]) / (actualPos[i] - actualPos[i - 1]));

            if ((newheight > height[i - 1]) && (newheight < height[i + 1])) {
               height[i] = newheight;
            } else {
               int di = (int) d;
               height[i] = height[i] + d * (height[i + di] - height[i]) / (actualPos[i + di] - actualPos[i]);
            }
            actualPos[i] += (int) d;
         }
      }
   }

   @Override
   public void computeMeasures(int batchNumber) {
      // Not handled
   }

   @Override
   public void resetBatchStatistic() {
      // Not handled
   }

   @Override
   public void resetStatistic() {
      for (int l = 0; l < 5; l++) {
         height[l] = 0;
         actualPos[l] = l + 1;
      }

      desiredPos[0] = 1.0;
      desiredPos[1] = 1.0 + 2.0 * percentage;
      desiredPos[2] = 1.0 + 4.0 * percentage;
      desiredPos[3] = 3.0 + 2.0 * percentage;
      desiredPos[4] = 5.0;

      increment[0] = 0;
      increment[1] = percentage / 2.0;
      increment[2] = percentage;
      increment[3] = (1 + percentage) / 2.0;
      increment[4] = 1.0;

      noOfSamples = 0;
   }

}
