/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.confidenceinterval;

public class StudentCalc extends Student {

   public StudentCalc(double confidenceLevel) {
      super(confidenceLevel);
   }

   protected double computeStudentUValue(double level, long samples) {
      final double raccu = 1.0E-15;
      double cvalue;
      double ivalue;
      double ssize;
      double caccu;
      ivalue = 0.5 / (1.0 - level);
      cvalue = ivalue;
      ssize = ivalue / 2.0;

      // Let's do a successive Approximation to get the Limits of the Integral
      // Easy and effective

      do {
         if (computeCurrentConfidenceLevel(cvalue, samples) > level) {
            cvalue -= ssize;
         } else {
            cvalue += ssize;
         }
         caccu = Math.abs(ssize / cvalue);
         ssize /= 2.0;
      } while (raccu < caccu);

      return cvalue;
   }

   protected double computeCurrentConfidenceLevel(double value, long samples) {
      int m;
      long maxm;
      double us = value * value;
      double a = Math.sqrt(samples - 1.0);
      double as = samples - 1.0;
      double gamma1 = 1.0;
      double gamma2 = 1.0;
      double aux1;
      double aux2 = 0;
      double sum;
      double p;

      // Perhaps this Routine is not the fastest, but it returns excellent
      // results and with the description of the UsersManual you
      // will understand, what happens.
      if (samples % 2 != 0) {
         maxm = (samples - 1) / 2; // n is odd
         sum = 2 * Math.log(value / a + Math.sqrt(us / as + 1.0));
         aux2 = -2 * value * Math.sqrt(as + us) / as;
         p = 1.0;
         for (m = 1; m <= maxm; m++) {
            aux1 = (2.0 * (m - 1.0)) / (as * (2 * m - 1.0));
            aux2 *= (2 * (m - 1) - 1.0) / ((2 * m - 1.0) * (as + us));
            sum = sum * aux1 + aux2;
            p *= samples - 1.0;
            gamma1 *= ((2 * m - 1) / 2.0);
            if (m > 2) {
               gamma2 *= (m - 1);
            }
            if (gamma1 > 1e100) {
               gamma1 /= 1e100;
               gamma2 /= 1e100;
            }
            if (p > 1e100) {
               p /= 1e100;
               sum *= 1e100;
               aux2 *= 1e100;
            }
         }
         return (p * sum * gamma1 / gamma2);
      } else {
         maxm = (samples / 2 - 1); // n is even
         sum = (1.0 / Math.sqrt(samples - 1.0)) * Math.atan(value / Math.sqrt(samples - 1.0));
         p = Math.sqrt(samples - 1.0);
         for (m = 1; m <= maxm; m++) {
            aux1 = (2.0 * m - 1.0) / (2.0 * m * as);
            if (m == 1) {
               aux2 = value / (2 * as * (as + us));
            } else {
               aux2 *= (2 * (m - 1)) / (2 * m * (as + us));
            }
            sum = sum * aux1 + aux2;
            gamma1 *= m; // G(n)=(n-1)!
            gamma2 *= ((2 * m - 1) / 2.0);
            p *= samples - 1.0;
            if (gamma1 > 1e100) {
               gamma1 /= 1e100;
               gamma2 /= 1e100;
            }
            if (p > 1e100) {
               p /= 1e100;
               sum *= 1e100;
               aux2 *= 1e100;
            }
         }
         return (2 * p * sum * gamma1 / (gamma2 * Math.PI));
      }
   }

   @Override
   public double getConfidenceInterval(long samples, double deviation) {
      double interval = Double.NaN;

      if (samples < 2) { // ConfidenceInterval not defined
         // for less than 2 samples
         return interval;
      }
      if ((getConfidenceLevel() < 0) || (getConfidenceLevel() >= 1)) {
         // Wrong ConfidenceLevel
         return interval;
      }
      interval = computeStudentUValue(getConfidenceLevel(), samples) * deviation / Math.sqrt(samples);
      return interval;
   }
}
