/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Summation;

import java.util.Arrays;

/**
 * <code>StdDistributionStatistic</code> extends
 * {@link ikr.simlib.statistics.distribution.DistributionStatistic} and provides
 * an implementation for a distribution statistic.
 */
public class StdDistributionStatistic extends DistributionStatistic implements BatchExportable, ReflectionConstructable2<SimNode, Parameters> {

   private static final String ARRAY_SIZE = "ArraySize";
   private static final String LOWER_BOUND = "LowerBound";
   private static final String UPPER_BOUND = "UpperBound";

   /** as required by {@link ReflectionConstructable} */
   public static StdDistributionStatistic createInstance(SimNode ownNode, Parameters pars) {
      final int arraySize = pars.get(ownNode, ARRAY_SIZE).asInteger();
      final double lowerBound = pars.get(ownNode, LOWER_BOUND).asDouble();
      final double upperBound = pars.get(ownNode, UPPER_BOUND).asDouble();

      return new StdDistributionStatistic(arraySize, lowerBound, upperBound, ownNode);
   }

   public static final String EXPORT_UNDERFLOWS = ResultKeyword.asExport("underflows");
   public static final String EXPORT_OVERFLOWS = ResultKeyword.asExport("overflows");
   public static final String EXPORT_SUMOFWEIGHTS = ResultKeyword.asExport("sumofweights");
   public static final String EXPORT_WEIGHT = ResultKeyword.asExport("weight");
   public static final String EXPORT_BINS = ResultKeyword.asExport("bins");

   public final int arraySize;
   public final double classWidth;
   public final double lowerLimit;
   public final double upperLimit;

   public final Summation underflowProb;
   public final Summation overflowProb;
   public final Summation[] marginalDist;
   public final Summation[] cumulativeDist;

   private double underflows;
   private double overflows;
   private final double[] weights;
   private double sumOfWeights;

   public StdDistributionStatistic(int arraySize, double lowerLimit, double upperLimit, SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      if (arraySize <= 0) {
         throw new SimulationException("array size must be > 0");
      }

      if (upperLimit <= lowerLimit) {
         throw new SimulationException("upper limit must be greater than lower limit");
      }

      this.arraySize = arraySize;
      this.lowerLimit = lowerLimit;
      this.upperLimit = upperLimit;
      weights = new double[arraySize];
      underflowProb = new Summation();
      overflowProb = new Summation();

      classWidth = (upperLimit - lowerLimit) / arraySize;

      marginalDist = new Summation[arraySize];
      cumulativeDist = new Summation[arraySize];
      for (int i = 0; i < arraySize; i++) {
         marginalDist[i] = new Summation();
         cumulativeDist[i] = new Summation();
      }

      resetStatistic();
   }

   @Override
   public int getArraySize() {
      return arraySize;
   }

   @Override
   public double getBatchDistribution(int bucketIndex) {
      double probabilitySum;

      if (sumOfWeights > 0) {
         double sum = underflows;
         for (int i = 0; i <= bucketIndex; i++) {
            sum += weights[i];
         }
         probabilitySum = sum / sumOfWeights;
      } else {
         probabilitySum = Double.NaN;
      }

      return probabilitySum;
   }

   @Override
   public double getBatchOverflowProb() {
      double probability;

      if (sumOfWeights > 0) {
         probability = overflows / sumOfWeights;
      } else {
         probability = Double.NaN;
      }

      return probability;
   }

   @Override
   public double getBatchProbability(int bucketIndex) {
      double probability;

      if (sumOfWeights > 0) {
         probability = weights[bucketIndex] / sumOfWeights;
      } else {
         probability = Double.NaN;
      }

      return probability;
   }

   @Override
   public double getBatchUnderflowProb() {
      double probability;

      if (sumOfWeights > 0) {
         probability = underflows / sumOfWeights;
      } else {
         probability = Double.NaN;
      }

      return probability;
   }

   @Override
   public double getDistribution(int bucketIndex) {
      return cumulativeDist[bucketIndex].getMean();
   }

   @Override
   public double getDistributionConfInterval(int bucketIndex) {
      return cumulativeDist[bucketIndex].getConfidenceInterval();
   }

   @Override
   public double getLowerBound(int bucketIndex) {
      return lowerLimit + bucketIndex * classWidth;
   }

   @Override
   public double getOverflowConfInterval() {
      return overflowProb.getConfidenceInterval();
   }

   @Override
   public double getOverflowProbability() {
      return overflowProb.getMean();
   }

   @Override
   public double getProbability(int bucketIndex) {
      return marginalDist[bucketIndex].getMean();
   }

   @Override
   public double getProbabilityConfInterval(int bucketIndex) {
      return marginalDist[bucketIndex].getConfidenceInterval();
   }

   @Override
   public double getUnderflowConfInterval() {
      return underflowProb.getConfidenceInterval();
   }

   @Override
   public double getUnderflowProbability() {
      return underflowProb.getMean();
   }

   @Override
   public double getUpperBound(int bucketIndex) {
      return lowerLimit + (bucketIndex + 1) * classWidth;
   }

   @Override
   public void update(double sample, double weight) {
      if (weight < 0) {
         throw new SimulationException("weight must be >= 0!");
      }

      if (Double.isNaN(sample) || Double.isNaN(weight)) {
         sumOfWeights = Double.NaN;
         underflows = Double.NaN;
         overflows = Double.NaN;
         for (int i = 0; i < weights.length; i++) {
            weights[i] = Double.NaN;
         }
      }

      if (sample <= lowerLimit) {
         underflows += weight;
      } else if (sample > upperLimit) {
         overflows += weight;
      } else {
         int index = 0;
         if (sample < lowerLimit + classWidth)
            // takes care of special case in which sample approx.
            // fLowerLimit and which could lead to int index = -1
            index = 0;
         else
            index = arraySize - 1 - (int) ((upperLimit - sample) / classWidth);

         weights[index] += weight;
      }

      sumOfWeights += weight;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      if (sumOfWeights > 0) {
         underflowProb.update(underflows / sumOfWeights);
         overflowProb.update(overflows / sumOfWeights);

         double probabilitySum = underflows / sumOfWeights;
         for (int i = 0; i < arraySize; i++) {
            double probability = weights[i] / sumOfWeights;
            marginalDist[i].update(probability);
            probabilitySum += probability;
            cumulativeDist[i].update(probabilitySum);
         }

      } else {

         underflowProb.update(Double.NaN);
         overflowProb.update(Double.NaN);

         for (int i = 0; i < arraySize; i++) {
            marginalDist[i].update(Double.NaN);
            cumulativeDist[i].update(Double.NaN);
         }
      }
   }

   @Override
   public void resetBatchStatistic() {
      underflows = 0;
      overflows = 0;
      sumOfWeights = 0;

      Arrays.fill(weights, 0);
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();

      underflowProb.reset();
      overflowProb.reset();

      for (int i = 0; i < arraySize; i++) {
         marginalDist[i].reset();
         cumulativeDist[i].reset();
      }
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeDouble(EXPORT_UNDERFLOWS, underflows);
      writer.writeDouble(EXPORT_OVERFLOWS, overflows);
      writer.writeDouble(EXPORT_SUMOFWEIGHTS, sumOfWeights);

      writer.writeStartArray(EXPORT_BINS);
      for (int i = 0; i < arraySize; i++) {
         writer.writeStartArrayItem(i);
         writer.writeDouble(EXPORT_WEIGHT, weights[i]);
         writer.writeEndArrayItem();
      }
      writer.writeEndArray();
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      underflows = reader.readDouble(EXPORT_UNDERFLOWS);
      overflows = reader.readDouble(EXPORT_OVERFLOWS);
      sumOfWeights = reader.readDouble(EXPORT_SUMOFWEIGHTS);

      reader.readStartArray(EXPORT_BINS);
      for (int i = 0; i < arraySize; i++) {
         reader.readStartArrayItem(i);
         weights[i] = reader.readDouble(EXPORT_WEIGHT);
         reader.readEndArrayItem();
      }
      reader.readEndArray();

      computeMeasures(0);
   }
}
