/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution.trace;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.distribution.DistributionStatistic;
import ikr.simlib.statistics.distribution.StdDistributionStatistic;
import ikr.simlib.statistics.distribution.trace.data.DataSource;
import ikr.simlib.statistics.distribution.trace.data.DataStorage;
import ikr.simlib.statistics.distribution.trace.data.Sample;
import ikr.simlib.statistics.distribution.trace.dist.CombinedDistributionData;
import ikr.simlib.statistics.distribution.trace.dist.DropDistributionData;
import ikr.simlib.statistics.distribution.trace.dist.MinMaxDistributionData;
import ikr.simlib.statistics.distribution.trace.quantile.LinearQuantiles;
import ikr.simlib.statistics.distribution.trace.quantile.QuantileDescription;

import java.util.ArrayList;
import java.util.List;

/**
 * Similar to the simlib {@link StdDistributionStatistic}, but buckets are
 * determined automatically. All samples are collected during the simulation. At
 * the end, the values are sorted and the quantiles are caluclated.
 * <p>
 * Output: This statistic is based on quantiles, so the size of "buckets" varies
 * (different to the {@link StdDistributionStatistic}). The main output is the
 * distribution over all drops. In addition, the distribution of each drop is
 * written to the result file separately. To have an approximation of the
 * estimation accuracy, the minimum and maximum value (over all drops) for each
 * quantile is calculated. Be careful: With few drops, this can become narrow,
 * although it might be inaccurate!
 */
public abstract class DistributionStatisticBase extends DistributionStatistic implements BatchExportable {

   private final List<DataSource> batchLists;
   private final QuantileDescription quantiles;
   private final boolean writeDropDistributions;
   private boolean inTransientPhase = false;
   private CombinedDistributionData distributionData = null;
   private List<DropDistributionData> dropDistributions = null;
   private MinMaxDistributionData minDistribution = null;
   private MinMaxDistributionData maxDistribution = null;
   private boolean distributionInvalid = false;
   private DataStorage batchData;
   /** store this for exporting (after end of batch) */
   private DataSource lastBatchData;
   private long samplesInBatch = 0;
   private long totalSamples = 0;

   public DistributionStatisticBase(SimNode ownNode, int numberOfQuantiles) {
      this(ownNode, numberOfQuantiles, false);
   }

   public DistributionStatisticBase(SimNode ownNode, int numberOfQuantiles, boolean writeDropDistributions) {
      this(ownNode, new LinearQuantiles(numberOfQuantiles), writeDropDistributions);
   }

   public DistributionStatisticBase(SimNode ownNode, QuantileDescription quantiles, boolean writeDropDistributions) {
      super(ownNode);
      this.quantiles = quantiles;
      this.writeDropDistributions = writeDropDistributions;
      batchLists = new ArrayList<DataSource>();
      simNode.setExportHandler(this);
      simNode.setPrintHandler(this);
   }

   @Override
   public void update(double sample, double weight) {
      if (!inTransientPhase) {
         assert batchData != null;
         batchData.write(new Sample(sample, weight));
         samplesInBatch++;
      }
   }

   protected abstract DataStorage createNewDataStorage(int batchNumber);

   @Override
   public void handleStartBatch(int batchNumber) {
      super.handleStartBatch(batchNumber);
      assert batchData == null;
      batchData = createNewDataStorage(batchNumber);
      lastBatchData = null;
   }

   @Override
   public void handleStartTransientPhase() {
      super.handleStartTransientPhase();
      inTransientPhase = true;
   }

   @Override
   public void handleStopTransientPhase() {
      super.handleStopTransientPhase();
      inTransientPhase = false;
   }

   @Override
   public void handleStopBatch(int batchNumber) {
      super.handleStopBatch(batchNumber);
      assert batchData != null;

      // sort the file here, because it needs less disk space when it's sorted
      batchData.writingCompleted();
      batchData.sortInPlace();
      batchLists.add(batchData.getDataSource());
      lastBatchData = batchData.getDataSource();
      totalSamples += samplesInBatch;

      batchData = null;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      // noop (called by super.handleStopBatch())
   }

   @Override
   public void resetBatchStatistic() {
      // called by super.handleStartBatch()
      samplesInBatch = 0;
   }

   @Override
   public void resetStatistic() {
      // noop (called by handleInitSimulation(), handleStartTransientPhase(),
      // handleStopTransientPhase())
   }

   /**************************************************************************
    * print server stuff
    ***************************************************************************/

   private void calculateDistributionData() {
      assert distributionData == null;
      assert dropDistributions == null;
      assert minDistribution == null;
      assert maxDistribution == null;
      assert distributionInvalid == false;

      // trace files are already sorted
      distributionData = new CombinedDistributionData(quantiles, batchLists);
      dropDistributions = new ArrayList<DropDistributionData>();
      for (DataSource bData : batchLists)
         dropDistributions.add(new DropDistributionData(quantiles, bData));
      minDistribution = new MinMaxDistributionData(quantiles, dropDistributions, true);
      maxDistribution = new MinMaxDistributionData(quantiles, dropDistributions, false);
      if (!writeDropDistributions)
         dropDistributions.clear();
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      if (distributionData == null && distributionInvalid == false)
         calculateDistributionData();

      switch (type) {
      case FINAL:
         if (distributionInvalid) {
            writer.writeLong(ResultKeywords.NUMBER_OF_SAMPLES, 0);
         } else {
            writer.writeLong(ResultKeywords.NUMBER_OF_SAMPLES, totalSamples);
            writer.writeDouble(ResultKeywords.MIN, distributionData.getSmallestValue());

            // combined distribution
            writer.writeStartArray(ResultKeywords.PDF.toString());

            for (int index = 0; index < quantiles.getNumberOfQuantiles(); index++) {
               writer.writeStartArrayItem(index);
               writer.writeDouble(ResultKeywords.UBOUND, distributionData.getQuantileValue(index));
               writer.writeDouble(ResultKeywords.DISTRIBUTION, distributionData.getQuantileProbability(index));
               writer.writeDouble(ResultKeywords.CDIST, 1.0 - distributionData.getQuantileProbability(index));
               writer.writeDouble(ResultKeywords.MIN, minDistribution.getQuantileValue(index));
               writer.writeDouble(ResultKeywords.MAX, maxDistribution.getQuantileValue(index));
               writer.writeEndArrayItem();
            }
            writer.writeEndArray();

            // drop distributions
            if (writeDropDistributions) {
               writer.writeStartArray("DropDistributions");
               for (int drop = 0; drop < dropDistributions.size(); drop++) {
                  final DropDistributionData dropDist = dropDistributions.get(drop);

                  writer.writeStartArrayItem(drop);

                  writer.writeDouble(ResultKeywords.MIN, dropDist.getSmallestValue());

                  writer.writeStartArray(ResultKeywords.PDF.toString());

                  for (int index = 0; index < quantiles.getNumberOfQuantiles(); index++) {
                     writer.writeStartArrayItem(index);
                     writer.writeDouble(ResultKeywords.UBOUND, dropDist.getQuantileValue(index));
                     writer.writeDouble(ResultKeywords.DISTRIBUTION, dropDist.getQuantileProbability(index));
                     writer.writeEndArrayItem();
                  }
                  writer.writeEndArray();

                  writer.writeEndArrayItem();
               }
               writer.writeEndArray();
            }
         }

         break;

      case BATCH:
         break;
      }

   }

   /**************************************************************************
    * import & export
    ***************************************************************************/
   private final static String EXPORT_NUM_SAMPLES = "NumberOfSamples";

   abstract void exportBatchData(ResultWriter writer, DataSource data) throws PrintException;

   abstract DataSource importBatchData(ResultReader reader) throws ImportException;

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeLong(EXPORT_NUM_SAMPLES, samplesInBatch);
      exportBatchData(writer, lastBatchData);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      samplesInBatch = reader.readLong(EXPORT_NUM_SAMPLES);
      totalSamples += samplesInBatch;
      batchLists.add(importBatchData(reader));
   }

   /**************************************************************************
    * distribution statistic interface
    ***************************************************************************/

   @Override
   public int getArraySize() {
      return quantiles.getNumberOfQuantiles();
   }

   @Override
   public double getLowerBound(int bucketIndex) {
      return bucketIndex == 0 ? distributionData.getSmallestValue() : distributionData.getQuantileValue(bucketIndex - 1);
   }

   @Override
   public double getUpperBound(int bucketIndex) {
      return distributionData.getQuantileValue(bucketIndex);
   }

   @Override
   public double getUnderflowProbability() {
      return 0;
   }

   @Override
   public double getUnderflowConfInterval() {
      return 0;
   }

   @Override
   public double getOverflowProbability() {
      return 0;
   }

   @Override
   public double getOverflowConfInterval() {
      return 0;
   }

   @Override
   public double getProbability(int bucketIndex) {
      // TODO Auto-generated method stub
      return 0;
   }

   @Override
   public double getProbabilityConfInterval(int bucketIndex) {
      // TODO Auto-generated method stub
      return 0;
   }

   @Override
   public double getDistribution(int bucketIndex) {
      return distributionData.getQuantileProbability(bucketIndex);
   }

   @Override
   public double getDistributionConfInterval(int bucketIndex) {
      // TODO Auto-generated method stub
      return 0;
   }

   @Override
   public double getBatchUnderflowProb() {
      return 0;
   }

   @Override
   public double getBatchOverflowProb() {
      return 0;
   }

   @Override
   public double getBatchProbability(int bucketIndex) {
      return 0;
   }

   @Override
   public double getBatchDistribution(int bucketIndex) {
      // return dropDist.getQuantileProbability(bucketIndex);
      return 0;
   }

}
