/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution.trace.data;

import ikr.simlib.model.SimNode;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;

/**
 * writes a series of double values to a file
 */
public class TraceWriter {
   private static final int BUFFER_SIZE = 102400; // in byte
   protected static final String FILE_NAME_PATTERN = "Trace_%s_Batch_%d.dat";
   private final DataOutputStream batchOutputStream;
   private final File file;
   private final boolean mergeAdjacentSamples;
   private long samplesInFile;

   public TraceWriter(SimNode namingNode, int batchNumber, boolean mergeAdjacentSamples) {
      this(new File(String.format(FILE_NAME_PATTERN, namingNode.getFullName(), batchNumber)), mergeAdjacentSamples);
   }

   public TraceWriter(File file, boolean mergeAdjacentSamples) {
      this.file = file;
      this.mergeAdjacentSamples = mergeAdjacentSamples;
      try {
         batchOutputStream = new DataOutputStream(new BufferedOutputStream(new DeflaterOutputStream(new BufferedOutputStream(new FileOutputStream(file), BUFFER_SIZE), new Deflater(Deflater.BEST_SPEED), BUFFER_SIZE), BUFFER_SIZE));
      } catch (FileNotFoundException e) {
         throw new RuntimeException("distribution trace file could not be created", e);
      }
      samplesInFile = 0;
   }

   public void write(Sample sample) {
      write(sample.value, sample.weight);
   }

   private double lastVal;
   private double lastWeight;
   private boolean lastValuesValid = false;

   public void write(double val, double weight) {
      if (lastValuesValid && (val != lastVal || !mergeAdjacentSamples))
         writeReally();
      if (lastValuesValid) {
         assert mergeAdjacentSamples;
         assert val == lastVal;
         lastWeight += weight;
      } else {
         lastVal = val;
         lastWeight = weight;
         lastValuesValid = true;
      }
   }

   private void writeReally() {
      assert lastValuesValid;
      lastValuesValid = false;
      try {
         batchOutputStream.writeDouble(lastVal);
         batchOutputStream.writeDouble(lastWeight);
         samplesInFile++;
      } catch (IOException e) {
         throw new RuntimeException(e);
      }
   }

   public long getSamplesInFile() {
      return samplesInFile + (lastValuesValid ? 1 : 0);
   }

   public File getFile() {
      return file;
   }

   public void close() {
      if (lastValuesValid)
         writeReally();
      try {
         batchOutputStream.close();
      } catch (IOException e) {
         throw new RuntimeException(e);
      }
   }
}
