/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution.trace.quantile;

/**
 * generates quantiles at logarithmic (base 10) positions. The largest value is
 * always 1, generated for the quantileIndex {@link #getNumberOfQuantiles()}-1.
 * The smallest value is the one given in the constructor. Output for (11, 0.1)
 * is
 * 
 * <pre>
 * 0.1 
 * 0.12589254117941673 
 * 0.15848931924611132
 * 0.19952623149688797 
 * 0.251188643150958 
 * 0.31622776601683794 
 * 0.3981071705534972
 * 0.5011872336272722 
 * 0.6309573444801932 
 * 0.7943282347242815 
 * 1.0
 * </pre>
 * */
public class LogarithmicQuantiles implements QuantileDescription {
   private final int numberOfQuantiles;
   private final double smallestValue;

   public LogarithmicQuantiles(int numberOfQuantiles, double smallestValue) {
      this.numberOfQuantiles = numberOfQuantiles;
      this.smallestValue = smallestValue;
   }

   @Override
   public int getNumberOfQuantiles() {
      return numberOfQuantiles;
   }

   @Override
   public double getRelativeQuantilePosition(int quantileIndex) {
      final double largestExponent = 0;
      final double smallestExponent = Math.log10(smallestValue);
      final double exp = smallestExponent + ((double) quantileIndex / (numberOfQuantiles - 1)) * (largestExponent - smallestExponent);
      return Math.pow(10, exp);
   }

   public static void main(String[] args) {
      LogarithmicQuantiles test = new LogarithmicQuantiles(4, 0.125);
      for (int i = 0; i < test.getNumberOfQuantiles(); i++)
         System.out.println(test.getRelativeQuantilePosition(i));
   }

}
