/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.sample;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * The classes <code>SampleStatistic</code> or
 * {@link ikr.simlib.statistics.sample.StdSampleStatistic} serve to register
 * single measured values (e.g. run times) of the type double with the help of
 * the method {@link #update(double)}. <br>
 * <br>
 * The following functions are available for evaluation:
 * <ul>
 * <li>
 * Mean value of all returned measured values ({@link #getMean()}) including a
 * confidence interval ({@link #getMeanConfidenceInterval()})
 * <li>
 * Standard deviation of all returned measured values ({@link #getDeviation()})
 * including a confidence interval ({@link #getDeviationConfidenceInterval()})
 * <li>
 * Coefficient of variation of all returned measured values (
 * {@link #getCoefficient()}) including a confidence interval (
 * {@link #getCoefficientConfidenceInterval()})
 * <li>
 * Value range ({@link #getMinimum()} and {@link #getMaximum()}) in reference to
 * all measured values
 * <li>
 * Mean value of the returned measured values from the current batch (
 * {@link #getBatchMean()})
 * <li>
 * Coefficient of variation of the returned measured values from the current
 * batch ({@link #getBatchCoefficient()})
 * <li>
 * Variation of the returned measured values from the current batch (
 * {@link #getBatchVariance()})
 * <li>
 * Number of measured values from the batch ({@link #getBatchEvents()})
 * </ul>
 * 
 * <p>
 * For printing the results it defines the following keywords:
 * </p>
 * 
 * <dl>
 * <dt>
 * mean</dt>
 * <dd>
 * mean of all samples in all batches</dd>
 * <dt>
 * cintmean</dt>
 * <dd>
 * confidence interval to mean value</dd>
 * <dt>
 * deviation</dt>
 * <dd>
 * standard deviation of all samples in all batches</dd>
 * <dt>
 * cintdeviation</dt>
 * <dd>
 * confidence interval to deviation value</dd>
 * <dt>
 * cov</dt>
 * <dd>
 * coefficient of std. dev. of all Batchs</dd>
 * <dt>
 * cintcov</dt>
 * <dd>
 * confidence interval to coefficient of variation</dd>
 * <dt>
 * max</dt>
 * <dd>
 * maximum sample of all batches</dd>
 * <dt>
 * min</dt>
 * <dd>
 * minimum sample of all batches</dd>
 * <dt>
 * numberofevents</dt>
 * <dd>
 * number of samples in actual Batch</dd>
 * <dt>
 * bmean</dt>
 * <dd>
 * mean of current batch</dd>
 * <dt>
 * bcov</dt>
 * <dd>
 * coefficient of variation in current batch</dd>
 * <dt>
 * bvariance</dt>
 * <dd>
 * variance of current Batch</dd>
 * <dt>
 * bmax</dt>
 * <dd>
 * maximum value in current batches</dd>
 * <dt>
 * bmin</dt>
 * <dd>
 * minimum value in current batches</dd>
 * </dl>
 */
public abstract class SampleStatistic extends Statistic implements Printable, UnweightedSampleCollector<Double> {

   public SampleStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case BATCH:
         writer.writeDouble(ResultKeywords.BATCH_MEAN, getBatchMean());
         writer.writeDouble(ResultKeywords.BATCH_VARIANCE, getBatchVariance());
         writer.writeDouble(ResultKeywords.BATCH_DEVIATION, getBatchStandardDeviation());
         writer.writeDouble(ResultKeywords.BATCH_COV, getBatchCoefficient());
         writer.writeLong(ResultKeywords.BATCH_NUMBER_OF_EVENTS, getBatchEvents());
         break;

      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.MEAN, getMean());
         writer.writeDouble(ResultKeywords.MEAN_CINT, getMeanConfidenceInterval());
         writer.writeDouble(ResultKeywords.DEVIATION, getDeviation());
         writer.writeDouble(ResultKeywords.DEVIATION_CINT, getDeviationConfidenceInterval());
         writer.writeDouble(ResultKeywords.COV, getCoefficient());
         writer.writeDouble(ResultKeywords.COV_CINT, getCoefficientConfidenceInterval());
         writer.writeDouble(ResultKeywords.MAX, getMaximum());
         writer.writeDouble(ResultKeywords.MIN, getMinimum());
         writer.writeLong(ResultKeywords.NUMBER_OF_EVENTS, getEvents());
         break;
      }
   }

   public abstract void update(double sample);

   @Override
   public void update(Double sample) {
      update(sample.doubleValue());
   }

   public abstract double getMean();

   public abstract double getDeviation();

   public abstract double getCoefficient();

   public abstract long getEvents();

   public abstract double getBatchMean();

   public abstract long getBatchEvents();

   public abstract double getBatchCoefficient();

   public abstract double getBatchVariance();

   public abstract double getBatchStandardDeviation();

   public abstract double getMeanConfidenceInterval();

   public abstract double getDeviationConfidenceInterval();

   public abstract double getCoefficientConfidenceInterval();

   public abstract double getMaximum();

   public abstract double getMinimum();
}
