/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.support;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.statistics.conditional.ConditionalSampleCollectorBase.ToIndexMapper;

/**
 * <code>BucketUtility</code> is an abstract base class of different bucket
 * utilities which relate a value provided to the <code>GetBucket</code> method
 * to a bucket index. The range of values of individual buckets is defined by
 * the base class implementations and follows a linear, geometric or logarithmic
 * law.
 */
public abstract class BucketUtility implements Cloneable, ToIndexMapper<Double> {

   private final double min;
   private final double max;
   private final int arraySize;

   public BucketUtility(double min, double max, int noOfBuckets) {
      if (min > max) {
         throw new SimulationException("overlapping min and max");
      }

      this.min = min;
      this.max = max;
      arraySize = noOfBuckets;
   }

   public double getMin() {
      return min;
   }

   public double getMax() {
      return max;
   }

   public int getArraySize() {
      return arraySize;
   }

   public boolean isInRange(double condVal) {
      return (condVal >= min) && (condVal < max);
   }

   @Override
   public int mapToIndex(Double key) {
      return getBucket(key);
   }

   public abstract int getBucket(double condVal);

   public abstract double getLowerBound(int index);

   public abstract double getUpperBound(int index);
}
