/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.weighted;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.Summation;
import ikr.simlib.statistics.WeightedSampleCollector;

/**
 * The class <code>WeightedMeanStatistic</code> is similar to
 * {@link ikr.simlib.statistics.sample.SampleStatistic} with the difference that
 * in the {@link #update(double, double)} method not only a measured value but
 * also a weight is passed. The default value for that weight is 1. Based on
 * that difference, only first order statistic are calculated. <br>
 * <br>
 * The following functions are available for evaluation:
 * <ul>
 * <li>
 * mean value of all measures {@link #getMean()}
 * <li>
 * mean value of all measures in a batch {@link #getBatchMean()}
 * <li>
 * confidence interval of the mean value {@link #getMeanConfidenceInterval()}
 * </ul>
 * Results can be printed using the following keywords within a print format: <br>
 * <dl>
 * <dt>
 * mean</dt>
 * <dd>
 * Mean value of all measured values in reference to all batches</dd>
 * <dt>
 * cintmean</dt>
 * <dd>
 * Confidence interval of the mean values in +/- notation</dd>
 * <dt>
 * bmean</dt>
 * <dd>
 * Mean value in reference to the current batch</dd>
 * </dl>
 * */
public class WeightedMeanStatistic extends Statistic implements BatchExportable, Printable, WeightedSampleCollector<Double, Double> {

   private static final String EXPORT_SUM_OF_WEIGHTS = ResultKeyword.asExport("sumofweights");
   private static final String EXPORT_SAMPLE_SUM = ResultKeyword.asExport("samplesum");

   private final Summation meanSummation;
   private double sampleSum;
   private double sumOfWeights;

   public WeightedMeanStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
      simNode.setExportHandler(this);

      meanSummation = new Summation();

      resetStatistic();
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case BATCH:
         writer.writeDouble(ResultKeywords.BATCH_MEAN, getBatchMean());
         break;

      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.MEAN, getMean());
         writer.writeDouble(ResultKeywords.MEAN_CINT, getMeanConfidenceInterval());
         break;
      }
   }

   public void update(double sample, double weight) {
      if (weight < 0) {
         throw new SimulationException("weight is < 0!");
      }

      sampleSum += sample * weight;
      sumOfWeights += weight;
   }

   @Override
   public void update(Double sample, Double weight) {
      update(sample.doubleValue(), weight.doubleValue());
   }

   public final void update(double sample) {
      update(sample, 1);
   }

   @Override
   public void update(Double sample) {
      update(sample.doubleValue());
   }

   public double getMean() {
      return meanSummation.getMean();
   }

   public double getBatchMean() {
      if (sumOfWeights > 0) {
         return sampleSum / sumOfWeights;
      } else {
         return Double.NaN;
      }
   }

   public double getMeanConfidenceInterval() {
      return meanSummation.getConfidenceInterval();
   }

   @Override
   public void computeMeasures(int batchNumber) {
      if (sumOfWeights > 0) {
         meanSummation.update(sampleSum / sumOfWeights);
      }
   }

   @Override
   public void resetBatchStatistic() {
      sampleSum = 0;
      sumOfWeights = 0;
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      meanSummation.reset();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeDouble(EXPORT_SAMPLE_SUM, sampleSum);
      writer.writeDouble(EXPORT_SUM_OF_WEIGHTS, sumOfWeights);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      sampleSum = reader.readDouble(EXPORT_SAMPLE_SUM);
      sumOfWeights = reader.readDouble(EXPORT_SUM_OF_WEIGHTS);

      computeMeasures(0);
   }
}
