/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.support;

import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.control.simulationcontrol.callbacks.StopSimulationCallbackI;
import ikr.simlib.exceptions.SimulationException;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashSet;
import java.util.Set;
import java.util.zip.GZIPOutputStream;

/**
 * Helper method(s) to work with trace files in simulations.
 */
public final class TraceHelper {

   private static final int BUFFER_SIZE = 1 * 1024 * 2024; // 1Mbyte
   private static final Set<String> TRACE_FILES = new HashSet<String>();

   /**
    * opens a trace file for writing and returns a print writer attached to that
    * file. The file is automatically closed by a
    * {@link StopSimulationCallbackI}. If fileName is null, no file is opened
    * but null is returned. If fileName ends with ".gz", the file is zipped
    * automatically.
    * <p>
    * The size of the buffer is increased to 1MB to avoid many random disc
    * accesses.
    * 
    * @param fileName
    *           name of the file to open
    * @return a print writer attached to that file (without automatic line
    *         flushing)
    */
   public static PrintWriter openFile(File fileName) {
      final PrintWriter retval;
      if (fileName != null) {
         if (TRACE_FILES.contains(fileName.getName())) {
            throw new SimulationException("Trace file with same name already created");
         }
         TRACE_FILES.add(fileName.getName());
         try {
            if (fileName.getName().endsWith(".gz"))
               retval = new PrintWriter(new GZIPOutputStream(new BufferedOutputStream(new FileOutputStream(fileName), BUFFER_SIZE)));
            else
               retval = new PrintWriter(new BufferedOutputStream(new FileOutputStream(fileName), BUFFER_SIZE));
         } catch (IOException e) {
            throw new RuntimeException(e);
         }
         SimulationControlManager.getInstance().registerStopSimulationCallback(new StopSimulationCallbackI() {
            @Override
            public void handleStopSimulation() {
               retval.close();
            }
         });
      } else
         retval = null;
      return retval;
   }

   private TraceHelper() {
      // disable constructor
   }

}
