/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.support.cmdline;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * <code>ArgParser</code> gives support for command line parsing.
 * 
 * Options use a hyphen and a single character and can take a number of
 * arguments.
 * 
 * The options <code>-?</code> and <code>-h</code> are predefined and should be
 * used to print the usage via <code>printUsage</code>.
 */
public class ArgParser {
   private static class Option {
      boolean available;

      public int minArgs;
      public int maxArgs;
      List<String> args;

      public Option(int minArgs, int maxArgs) {
         this.minArgs = minArgs;
         this.maxArgs = maxArgs;
         args = new ArrayList<String>();
      }
   }

   /**
    * Default output stream for usage messages.
    */
   private static final PrintStream OUTPUT_STREAM = System.out;

   /**
    * Default output stream for error messages.
    */
   private static final PrintStream ERROR_STREAM = System.err;

   /**
    * Valid values for a true boolean argument
    */
   private static final List<String> BOOLEAN_TRUE_VALUES = Arrays.asList(new String[] { "1", "yes", "true", "on" });

   /**
    * Valid values for a true boolean argument
    */
   private static final List<String> BOOLEAN_FALSE_VALUES = Arrays.asList(new String[] { "0", "no", "false", "off" });

   /**
    * Program name.
    */
   private final String name;

   /**
    * Command line usage.
    */
   private final String usage;

   private final int minArgs;
   private int maxArgs;
   private final List<String> args;

   private final Map<Character, Option> options;

   // -- Constructors --

   /**
    * Constructs a new argument parser for the program.
    * 
    * The name and usage are used to provide the command line user with
    * additional information.
    * 
    * @param _name
    *           name of the program
    * @param _usage
    *           usage text for the program
    * @param _minArgs
    *           the minimum number of additional arguments
    * @param _maxArgs
    *           the maximum number of additional arguments or -1 for no maximum.
    */
   public ArgParser(String _name, String _usage, int _minArgs, int _maxArgs) {
      name = _name;
      usage = _usage;

      minArgs = _minArgs;
      if (_maxArgs == -1) {
         maxArgs = Integer.MAX_VALUE;
      } else {
         maxArgs = _maxArgs;
      }

      if (minArgs < 0 || minArgs > maxArgs) {
         throw new IllegalArgumentException("min must be >= 0 and < max");
      }

      args = new ArrayList<String>();

      options = new HashMap<Character, Option>();

      addOption('h');
      addOption('?');
   }

   /**
    * Constructs a new argument parser for the program.
    * 
    * The name and usage are used to provide the command line user with
    * additional information.
    * 
    * @param name
    *           name of the program
    * @param usage
    *           usage text for the program
    * @param args
    *           the number of required arguments
    */
   public ArgParser(String name, String usage, int args) {
      this(name, usage, args, args);
   }

   /**
    * Constructs a new argument parser for the program.
    * 
    * The name and usage are used to provide the command line user with
    * additional information.
    * 
    * No additional arguments besides any options are allowed.
    * 
    * @param name
    *           name of the program
    * @param usage
    *           usage text for the program
    */
   public ArgParser(String name, String usage) {
      this(name, usage, 0, 0);
   }

   // -- Parsing --

   /**
    * Parses the arguments.
    * 
    * If either <code>-h</code> or <code>-?</code> is passed as argument the
    * method will return true.
    * 
    * @return whether a help option was provided
    * 
    * @throws ArgumentException
    *            if any option/argument constraint has been broken.
    * 
    * @see #parseSilent(String[]) to prevent an exception to be thrown
    */
   public boolean parse(String[] arguments) {
      for (Option o : options.values()) {
         o.available = false;
         o.args.clear();
      }

      char optionChar = '\0';
      Option option = null;

      for (String arg : arguments) {
         // Skip empty arguments
         if (arg.isEmpty()) {
            continue;
         }

         // New option?
         if (arg.matches("^-.$")) {
            optionChar = arg.charAt(1);

            // Is it a valid option?
            if (!options.containsKey(optionChar)) {
               throw new ArgumentException("Unknown option -" + optionChar);
            }

            option = options.get(optionChar);

            // Mark the option as seen
            option.available = true;

            // Set current option to null if it takes no arguments
            if (option.maxArgs == 0) {
               option = null;
            }
         } else {
            if (option != null) {

               // Check for too many arguments
               if (option.args.size() == option.maxArgs) {
                  throw new ArgumentException("Too many arguments for option -" + optionChar);
               }

               // Add option argument
               option.args.add(arg);

               option = null;

            } else {
               // Program argument

               if (this.args.size() < maxArgs) {
                  this.args.add(arg);
               } else {
                  throw new ArgumentException("Too many arguments");
               }
            }
         }
      }

      // Check options for minimum arguments
      for (Map.Entry<Character, Option> o : options.entrySet()) {
         option = o.getValue();

         if (option.available && (option.args.size() < option.minArgs)) {
            throw new ArgumentException("Not enough arguments for option -" + o.getKey());
         }
      }

      // Check minimum program arguments
      if (this.args.size() < minArgs) {
         throw new ArgumentException("Not enough arguments");
      }

      return options.get('?').available || options.get('h').available;
   }

   /**
    * Parses the arguments and returns whether an error or a help option
    * occurred.
    */
   public boolean parseSilent(String[] arguments) {
      try {
         if (!parse(arguments)) {
            return false;
         }
      } catch (ArgumentException ae) {
         return true;
      }

      return true;
   }

   // -- Usage --
   public void printUsage(PrintStream out) {
      out.printf("Usage %s: %s%n", name, usage);
   }

   public void printUsage() {
      printUsage(OUTPUT_STREAM);
   }

   public void printError(PrintStream out, String msg) {
      out.printf("Error in %s: %s%n", name, msg);
      printUsage(out);
   }

   public void printError(String msg) {
      printError(ERROR_STREAM, msg);
   }

   // -- Adding options --

   /**
    * Adds an option to the parser.
    * 
    * The option might take any number of arguments as limited by
    * <code>minArgs</code> and <code>maxArgs</code>.
    * 
    * @param _minimumNumberOfArguments
    *           the minimum number of arguments for this option
    * @param _maximumNumberOfArguments
    *           the maximum number of arguments for this option or -1 for no
    *           maximum.
    * 
    * @return this
    */
   public ArgParser addOption(char option, int _minimumNumberOfArguments, int _maximumNumberOfArguments) {
      int minimumNumberOfArguments = _minimumNumberOfArguments;
      int maximumNumberOfArguments = _maximumNumberOfArguments;

      if (maximumNumberOfArguments == -1) {
         maximumNumberOfArguments = Integer.MAX_VALUE;
      }

      if (option == '\0') {
         throw new IllegalArgumentException("Illegal option");
      }

      if (minimumNumberOfArguments < 0 || minimumNumberOfArguments > maximumNumberOfArguments) {
         throw new IllegalArgumentException("min must be >= 0 and < max");
      }

      if (options.containsKey(option)) {
         throw new IllegalArgumentException("Option " + option + " is assigned already");
      }

      options.put(option, new Option(minimumNumberOfArguments, maximumNumberOfArguments));

      return this;
   }

   /**
    * Adds an option with exactly <code>args</code> arguments to the parser.
    * 
    * @param option
    *           the option character
    * @param numberOfArguments
    *           the number of arguments for this option
    * 
    * @return this
    */
   public ArgParser addOption(char option, int numberOfArguments) {
      return addOption(option, numberOfArguments, numberOfArguments);
   }

   /**
    * Adds an option without arguments to the parser.
    * 
    * @param option
    *           the option character
    * 
    * @return this
    */
   public ArgParser addOption(char option) {
      return addOption(option, 0, 0);
   }

   // -- Options --

   public boolean hasOption(char option) {
      return options.containsKey(option) && options.get(option).available;
   }

   // -- Arguments --

   public int getArgumentCount() {
      return args.size();
   }

   public int getArgumentCount(char option) {
      if (!options.containsKey(option)) {
         throw new ArgumentException("Unknown option -" + option);
      }

      return options.get(option).args.size();
   }

   /**
    * Returns whether the program has arguments.
    */
   public boolean hasArguments() {
      return !args.isEmpty();
   }

   /**
    * Returns whether the option has arguments or is provided at all.
    */
   public boolean hasArguments(char option) {
      return hasOption(option) && getArgumentCount(option) != 0;
   }

   /**
    * Returns whether a nth argument is available for the program or not.
    */
   public boolean hasArgument(int n) {
      return n >= 0 && n < getArgumentCount();
   }

   /**
    * Returns whether a nth argument is available for an option or not.
    */
   public boolean hasArgument(char option, int n) {
      return n >= 0 && n < getArgumentCount(option);
   }

   public String getArgument() {
      return getNthArgument(0);
   }

   public List<String> getArguments() {
      return args;
   }

   public String getArgument(String defaultValue) {
      if (!hasArguments()) {
         return defaultValue;
      }

      return getArgument();
   }

   public String getNthArgument(int n) {
      if (!hasArgument(n)) {
         throw new ArgumentException(n + "th argument is missing");
      }

      return args.get(n);
   }

   public List<String> getArguments(char option) {
      // Will fail if option is missing.
      hasArguments(option);

      return options.get(option).args;
   }

   public String getArgument(char option) {
      return getNthArgument(option, 0);
   }

   public String getArgument(char option, String defaultValue) {
      if (!hasArguments(option)) {
         return defaultValue;
      }

      return getArgument(option);
   }

   public String getNthArgument(char option, int n) {
      if (!hasArgument(option, n)) {
         throw new ArgumentException("Option -" + option + "has no " + n + "th argument");
      }

      return options.get(option).args.get(n);
   }

   // -- Parsed arguments --

   // Integers

   public int getIntArgument() {
      return getNthIntArgument(0);
   }

   public int getIntArgument(int defaultValue) {
      if (!hasArguments()) {
         return defaultValue;
      }

      return getIntArgument();
   }

   public int getNthIntArgument(int n) {
      String s = getNthArgument(n);

      try {
         return Integer.parseInt(s);
      } catch (NumberFormatException nfe) {
         throw new ArgumentException("Couldn't parse argument");
      }
   }

   public int getIntArgument(char option) {
      return getNthIntArgument(option, 0);
   }

   public int getIntArgument(char option, int defaultValue) {
      if (!hasArguments(option)) {
         return defaultValue;
      }

      return getIntArgument(option);
   }

   public int getNthIntArgument(char option, int n) {
      String s = getNthArgument(option, n);

      try {
         return Integer.parseInt(s);
      } catch (NumberFormatException nfe) {
         throw new ArgumentException("Couldn't parse argument for option -" + option);
      }
   }

   // Booleans

   public boolean getBoolArgument() {
      return getNthBoolArgument(0);
   }

   public boolean getBoolArgument(boolean defaultValue) {
      if (!hasArguments()) {
         return defaultValue;
      }

      return getBoolArgument();
   }

   public boolean getNthBoolArgument(int n) {
      String s = getNthArgument(n);

      if (BOOLEAN_TRUE_VALUES.contains(s)) {
         return true;
      } else if (BOOLEAN_FALSE_VALUES.contains(s)) {
         return false;
      }

      throw new ArgumentException("Couldn't parse argument");
   }

   public boolean getBoolArgument(char option) {
      return getNthBoolArgument(option, 0);
   }

   public boolean getBoolArgument(char option, boolean defaultValue) {
      if (!hasArguments(option)) {
         return defaultValue;
      }

      return getBoolArgument(option);
   }

   public boolean getNthBoolArgument(char option, int n) {
      String s = getNthArgument(option, n);

      if (BOOLEAN_TRUE_VALUES.contains(s)) {
         return true;
      } else if (BOOLEAN_FALSE_VALUES.contains(s)) {
         return false;
      }

      throw new ArgumentException("Couldn't parse argument for option -" + option);
   }

   // Doubles

   public double getDoubleArgument() {
      return getNthDoubleArgument(0);
   }

   public double getDoubleArgument(double defaultValue) {
      if (!hasArguments()) {
         return defaultValue;
      }

      return getDoubleArgument();
   }

   public double getNthDoubleArgument(int n) {
      String s = getNthArgument(n);

      try {
         return Double.parseDouble(s);
      } catch (NumberFormatException nfe) {
         throw new ArgumentException("Couldn't parse argument");
      }
   }

   public double getDoubleArgument(char option) {
      return getNthDoubleArgument(option, 0);
   }

   public double getDoubleArgument(char option, double defaultValue) {
      if (!hasArguments(option)) {
         return defaultValue;
      }

      return getDoubleArgument(option);
   }

   public double getNthDoubleArgument(char option, int n) {
      String s = getNthArgument(option, n);

      try {
         return Double.parseDouble(s);
      } catch (NumberFormatException nfe) {
         throw new ArgumentException("Couldn't parse argument for option -" + option);
      }
   }
}
