/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.support.compression;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.ProcessBuilder.Redirect;
import java.util.Arrays;

public class CommandFilterStreams {

   private static final int STREAM_BUFFER_SIZE = 16 * 4096;

   public static OutputStream createFilterCommandOutputStream(String[] cmd, OutputStream stdout) {
      final Process p = startProcess(cmd);
      new Thread(new StreamCopyTool(p.getInputStream(), stdout), "CommandFilterOutputStream").start();
      return new BufferedOutputStream(p.getOutputStream(), STREAM_BUFFER_SIZE);
   }

   public static InputStream createFilterCommandInputStream(String[] cmd, InputStream stdin) {
      final Process p = startProcess(cmd);
      new Thread(new StreamCopyTool(stdin, p.getOutputStream()), "CommandFilterOutputStream").start();
      return new BufferedInputStream(p.getInputStream(), STREAM_BUFFER_SIZE);
   }

   private static Process startProcess(String[] cmd) {
      final ProcessBuilder pb = new ProcessBuilder(cmd);
      pb.redirectError(Redirect.INHERIT);
      try {
         final Process p = pb.start();
         new Thread(new Waiter(p, cmd), "WaitForProcess(" + cmd[0] + ")").start();
         return p;
      } catch (IOException e) {
         throw new RuntimeException(e);
      }
   }

   private static class Waiter implements Runnable {
      private final Process p;
      private String[] cmd;

      private Waiter(Process p, String[] cmd) {
         this.p = p;
         this.cmd = cmd;
      }

      @Override
      public void run() {
         try {
            final int exitCode = p.waitFor();
            if (exitCode != 0)
               System.err.println("process " + Arrays.toString(cmd) + " returned " + exitCode);
         } catch (InterruptedException e) {
            e.printStackTrace();
         }
      }
   }

   private static class StreamCopyTool implements Runnable {
      final InputStream in;
      final OutputStream out;

      protected StreamCopyTool(InputStream in, OutputStream out) {
         this.in = new BufferedInputStream(in, STREAM_BUFFER_SIZE);
         this.out = new BufferedOutputStream(out, STREAM_BUFFER_SIZE);
      }

      @Override
      public void run() {
         byte[] buffer = new byte[4096];
         boolean done = false;
         boolean closed = false;
         while (!done) {
            int val;
            try {
               val = in.read(buffer);
            } catch (IOException e) {
               throw new RuntimeException(e);
            }
            if (val == -1)
               done = true;
            else
               try {
                  out.write(buffer, 0, val);
               } catch (IOException e) {
                  // ignore closed stream
                  done = true;
                  closed = true;
               }
         }
         if (!closed)
            try {
               out.close();
            } catch (IOException e) {
               throw new RuntimeException(e);
            }
      }
   }
}
