/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.support.compression;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ServiceLoader;

/**
 * This class provides static methods to
 * <ul>
 * <li>open compressed files for reading</li>
 * <li>create and open compressed files for writing</li>
 * <li>determining the filename suffix for a given compression type</li>
 * </ul>
 * <p>
 * To avoid direct dependencies to compression libraries, the
 * {@link ServiceLoader} concept is used. This allows additional compression
 * types to be added without changing any SimLib class. The SimLib supports
 * uncompressed files as well as the compressed files of the types gzip, xz,
 * lzma and bzip2.
 * <p>
 * To implement additional compression types, you have to
 * <ul>
 * <li>create a class which implements {@link CompressionProvider}</li>
 * <li>add the provider description to the jar file which contains your class
 * (see {@link ServiceLoader} and
 * {@code http://ant.apache.org/manual/Tasks/jar.html}</li>
 * <li>place the jar with your class and the provider description in the
 * classpath of the simulation</li>
 * </ul>
 * 
 */
public class CompressionManager {

   /**
    * Opens the given file for reading.
    * 
    * @param file
    *           the file to open, with extension as it is on the disk
    * @return an InputStream to read the file
    * @throws FileNotFoundException
    *            when the file is not readable
    */
   public static InputStream getInputStream(File file) throws FileNotFoundException {
      return selectProviderByFile(file).getInputStream(file);
   }

   /**
    * Opens the given file for writing. The extension is not appended
    * automatically. See
    * {@link #getFilenameWithExtension(String, CompressionType)}.
    * 
    * @param file
    *           the file to open, with extension as it shall be created
    * @param compressionType
    *           the type of file compression which shall be used
    * @return an OutputStream to write to this file
    * @throws FileNotFoundException
    *            when the file cannot be created
    */
   public static OutputStream getOutputStream(File file, CompressionType compressionType) throws FileNotFoundException {
      return selectProvider(compressionType).getOutputStream(file);
   }

   /**
    * Opens the given file for writing. The file type will be determined by the
    * extension which has to be included in <tt>fileWithExt<tt>.
    * If the file extension is specified within {@link CompressionType}
    * then a file with corresponding compression shall be created.
    * If the extension could not be recognized then a regular file of
    * given name and extension shall be created.
    * 
    * @param file
    *           the file to open, with extension as it shall be created
    * @return an OutputStream to write to this file
    * @throws FileNotFoundException
    *            when the file cannot be created
    */
   public static OutputStream getOutPutStream(File file) throws FileNotFoundException {
      return selectProviderByFile(file).getOutputStream(file);
   }

   private static CompressionProvider selectProviderByFile(File file) {
      CompressionType matchingType = null;
      for (CompressionType type : CompressionType.values()) {
         if (file.getName().endsWith(type.fileExtension)) {
            matchingType = type;
            break; // break, otherwise none will match, too
         }
      }
      /* none is declared last, so it serves as catch-all */
      assert matchingType != null;
      return selectProvider(matchingType);
   }

   private static CompressionProvider selectProvider(CompressionType compressionType) {
      switch (compressionType) {
      case lzma:
         return new LZMACommandCompressionProvider();

      case xz:
         return new XzCommandCompressionProvider();

      case bzip2:
         return new BZip2CommandCompressionProvider();

      case gzip:
         return new GZipCompressionProvider();

      case none:
         return new NoCompressionProvider();
      }
      throw new Error("compression type " + compressionType + " not supported");
   }

   public static String getFilenameWithExtension(String fileName, CompressionType compressionType) {
      return fileName + compressionType.fileExtension;
   }

}
