/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.meters.rate;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.IPPacket;
import ikr.simlib.messages.Message;
import ikr.simlib.meters.OnePointMeter;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable3;
import ikr.simlib.ports.Port;
import ikr.simlib.untested.statistics.fairness.RateFairnessStatistic;

/**
 * With <code>FairnessRateMeter</code> the rates of different connections on one
 * or more ports can be measured. A connection is identified by the destination
 * id of their packets (of type <code>IPPacket</code>). The fairness is
 * evaluated using a fairness statistic.
 */
public class FairnessRateMeter extends OnePointMeter implements ReflectionConstructable3<SimNode, Parameters, Integer> {

   private static final String LD_MIN_BLOCK_SIZE = "LdMinBlockSize";
   private static final String LD_MAX_BLOCK_SIZE = "LdMaxBlockSize";
   private static final String TRACE_FILE = "TraceFile";
   private static final String TIME_BASE = "TimeBase";
   private static final String START_ID = "StartID";
   private static final String MODULO = "Modulo";

   /** as required by {@link ReflectionConstructable} */
   public static FairnessRateMeter createInstance(SimNode ownNode, Parameters pars, Integer noOfIds) {
      final int ldMinBlockSize = pars.get(ownNode, LD_MIN_BLOCK_SIZE).asInteger();
      final int ldMaxBlockSize = pars.get(ownNode, LD_MAX_BLOCK_SIZE).asInteger();
      final String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      final Duration timeBase = pars.get(ownNode, TIME_BASE).asInstance(Duration.class, ownNode.createChildNode(TIME_BASE), pars);
      final int startId = pars.get(ownNode, START_ID).asInteger();
      final int modulo = pars.get(ownNode, MODULO).asInteger();
      return new FairnessRateMeter(noOfIds, timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode, startId, modulo);
   }

   protected final int startId;
   protected final int modulo;

   protected final RateFairnessStatistic rateStatistic;

   public FairnessRateMeter(int noOfIds, Duration timeBase, int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode, int startId, int modulo) {
      this.startId = startId;
      this.modulo = modulo;
      rateStatistic = new RateFairnessStatistic(noOfIds, timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode.createChildNode("RateStatistic"));
   }

   public FairnessRateMeter(int noOfIds, Duration timeBase, int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode, int startId) {
      this(noOfIds, timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode, startId, 1);
   }

   public FairnessRateMeter(int noOfIds, Duration timeBase, int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode) {
      this(noOfIds, timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode, 0, 1);
   }

   protected boolean useMessage(Message msg) {
      return true;
   }

   @Override
   protected void evaluateMessage(Message msg, Port port) {
      if (useMessage(msg)) {
         PointInTime currentTime = Calendar.getInstance().getSystemTime();
         IPPacket packet = (IPPacket) msg;

         int _id = (packet.getDestinationId() - startId) / modulo;
         rateStatistic.update(currentTime, _id, msg.getLength());
      }
   }
}
