/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.meters.rate;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.meters.OnePointMeter;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.ports.Port;
import ikr.simlib.untested.statistics.variancetime.VTRateStatistic;

/**
 * This meter can measure variance-time statistics of rates.
 */
public class VTRateMeter extends OnePointMeter implements ReflectionConstructable2<SimNode, Parameters> {

   private static final String LD_MIN_BLOCK_SIZE = "LdMinBlockSize";
   private static final String LD_MAX_BLOCK_SIZE = "LdMaxBlockSize";
   private static final String TRACE_FILE = "TraceFile";
   private static final String TIME_BASE = "TimeBase";
   private static final String USE_LENGTH_AS_WEIGHT = "UseLengthAsWeight";

   /** as required by {@link ReflectionConstructable} */
   public static VTRateMeter createInstance(SimNode ownNode, Parameters pars) {
      final int ldMinBlockSize = pars.get(ownNode, LD_MIN_BLOCK_SIZE).asInteger();
      final int ldMaxBlockSize = pars.get(ownNode, LD_MAX_BLOCK_SIZE).asInteger();
      final String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      final Duration timeBase = pars.get(ownNode, TIME_BASE).asInstance(Duration.class, ownNode.createChildNode(TIME_BASE), pars);
      final boolean useLengthAsWeight = pars.get(ownNode, USE_LENGTH_AS_WEIGHT).asBoolean();
      return new VTRateMeter(timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, useLengthAsWeight, ownNode);
   }

   protected final boolean useLengthAsWeight;
   protected final VTRateStatistic vtStatistic;

   public VTRateMeter(Duration timeBase, int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, boolean useLengthAsWeight, SimNode ownNode) {
      this.useLengthAsWeight = useLengthAsWeight;
      vtStatistic = new VTRateStatistic(timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode.createChildNode("VTStat"));
   }

   protected boolean useMessage(Message msg) {
      return true;
   }

   @Override
   protected void evaluateMessage(Message msg, Port port) {
      if (useMessage(msg)) {
         PointInTime currentTime = Calendar.getInstance().getSystemTime();
         if (useLengthAsWeight) {
            vtStatistic.update(currentTime, msg.getLength());
         } else {
            vtStatistic.update(currentTime);
         }
      }
   }
}
