/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.boundary;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.boundary.BoundaryStatistic;

/**
 * <code>UnlimitedUnbatchedBoundaryStatistic</code> provides an implementation
 * of <code>BoundaryStatistic</code> to calculate quantiles for a given
 * percentage. Note that the percentage parameter in the constructor must be
 * between 0 and 1.
 */
public class UnlimitedUnbatchedBoundaryStatistic extends BoundaryStatistic implements BatchExportable {

   private final double percentage;

   /**
    * Sorted list of samples
    */
   private final SortedSamples samples;
   private final SortedSamples samplesSinceBatchBegin;

   public UnlimitedUnbatchedBoundaryStatistic(double percentage, SimNode ownNode) {
      super(ownNode);

      if (percentage < 0 || percentage > 1) {
         throw new SimulationException("Percentage out of range");
      }

      this.percentage = percentage;
      samples = new SortedSamples();
      samplesSinceBatchBegin = new SortedSamples();

      resetStatistic();
   }

   @Override
   public int getNoOfSamples() {
      return samples.size();
   }

   @Override
   public double getPercentage() {
      return percentage;
   }

   @Override
   public double getQuantile() {
      final double realPos;

      if (samples.size() == 0) {
         // no data => quantile undefined
         return Double.NaN;
      }

      realPos = samples.size() * percentage;

      final int lowerIndex = (int) Math.floor(realPos);
      final int upperIndex = (int) Math.ceil(realPos);
      assert (lowerIndex < samples.size());
      assert (upperIndex < samples.size());
      final double lowerValue = samples.get(lowerIndex);
      final double upperValue = samples.get(upperIndex);

      final double quantile;
      if (lowerIndex == upperIndex) {
         // no interpolation in case of identical indices
         quantile = lowerValue;
      } else {
         // interpolation
         quantile = lowerValue * (realPos - lowerIndex) + upperValue * (upperIndex - realPos);
      }

      return quantile;
   }

   @Override
   public void update(double sample) {
      samples.add(sample);
      samplesSinceBatchBegin.add(sample);
   }

   @Override
   public void computeMeasures(int batchNumber) {
      // not required
   }

   @Override
   public void resetBatchStatistic() {
      samplesSinceBatchBegin.clear();
   }

   @Override
   public void resetStatistic() {
      samples.clear();
      samplesSinceBatchBegin.clear();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {

      // export samples since batch begin
      samplesSinceBatchBegin.exportSamples(writer);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {

      SortedSamples importedSamples = new SortedSamples();
      importedSamples.importSamples(reader);

      // add samples to existing samples
      for (int index = 0; index < importedSamples.size(); index++) {
         samples.add(importedSamples.get(index));
      }
   }
}
