/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.fairness;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable3;
import ikr.simlib.statistics.Summation;

public class ProbFairnessStatistic extends FairnessStatistic implements ReflectionConstructable3<SimNode, Parameters, Integer> {

   private static final String LD_MIN_BLOCK_SIZE = "LdMinBlockSize";
   private static final String LD_MAX_BLOCK_SIZE = "LdMaxBlockSize";
   private static final String TRACE_FILE = "TraceFile";
   private static final String TIME_BASE = "TimeBase";

   /** as required by {@link ReflectionConstructable} */
   public static ProbFairnessStatistic createInstance(SimNode ownNode, Parameters pars, Integer noOfIds) {
      final int ldMinBlockSize = pars.get(ownNode, LD_MIN_BLOCK_SIZE).asInteger();
      final int ldMaxBlockSize = pars.get(ownNode, LD_MAX_BLOCK_SIZE).asInteger();
      final String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      final double timeBase = pars.get(ownNode, TIME_BASE).asDouble();
      return new ProbFairnessStatistic(noOfIds, timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);
   }

   protected final int noOfIds;
   protected final double timeBase;

   protected double nextEndOfInterval;
   protected int noOfIntervals;
   protected final double[][] totalBlockSums;
   protected final double[][] specialBlockSums;

   public ProbFairnessStatistic(int noOfIds, double timeBase, int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode) {
      super(ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);

      if (noOfIds <= 1) {
         throw new SimulationException("no. of ids mus be > 1");
      }

      this.noOfIds = noOfIds;
      this.timeBase = timeBase;
      nextEndOfInterval = timeBase * minBlockSize;
      noOfIntervals = 0;
      totalBlockSums = new double[noOfIds][ldMaxBlockSize - ldMinBlockSize + 1];
      specialBlockSums = new double[noOfIds][ldMaxBlockSize - ldMinBlockSize + 1];
   }

   public void update(double currentTime, int id, boolean isSpecial, double weight) {
      int i, blockSize;

      totalBlockSums[id][0] += weight;
      if (isSpecial) {
         specialBlockSums[id][0] += weight;
      }
      while (currentTime > nextEndOfInterval) {
         noOfIntervals += minBlockSize;
         for (i = 0, blockSize = minBlockSize; (i <= ldMaxBlockSize - ldMinBlockSize) && (noOfIntervals % blockSize == 0); i++, blockSize *= 2) {
            // TODO: Might need to reuse this object
            Summation summation = new Summation();
            for (int j = 0; j < noOfIds; j++) {
               if (totalBlockSums[j][i] != 0) {
                  double blockMean = specialBlockSums[j][i] / totalBlockSums[j][i];
                  summation.update(blockMean);
                  if (i < ldMaxBlockSize - ldMinBlockSize) {
                     totalBlockSums[j][i + 1] += totalBlockSums[j][i];
                     specialBlockSums[j][i + 1] += specialBlockSums[j][i];
                  }
                  totalBlockSums[j][i] = 0;
                  specialBlockSums[j][i] = 0;
               }
            }

            double cov = summation.getCoefficientOfVariation();

            if (Double.isNaN(cov)) {
               blockCoVSums[i] = Double.NaN;
               blockFISums[i] = Double.NaN;

            } else if (Double.isNaN(blockCoVSums[i])) {
               // do not update blockFISums and blockCoVSums
            } else {
               blockCoVSums[i] += cov;
               blockFISums[i] += 1.0 / (1.0 + cov * cov);
               noOfBlocks[i]++;
            }

            if (traceFile != null) {
               traceFile.printf("Block%d %g %g%n", i + ldMinBlockSize, currentTime, cov);
            }
         }
         nextEndOfInterval += timeBase * minBlockSize;
      }
   }

   public final void update(double currentTime, int id, boolean isSpecial) {
      update(currentTime, id, isSpecial, 1);
   }
}
