/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.variancetime;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * <code>VTBurstLevelRateStatistic</code> provides an implementation to estimate
 * the VT plot of rates on burst level.
 * 
 * @see "Dissertation Juergen Enssle (LDB 28226)"
 */
public class VTBurstLevelRateStatistic extends VTStatistic {

   private static final String LD_MIN_BLOCK_SIZE = "LdMinBlockSize";
   private static final String LD_MAX_BLOCK_SIZE = "LdMaxBlockSize";
   private static final String TRACE_FILE = "TraceFile";
   private static final String TIME_BASE = "TimeBase";

   /** as required by {@link ReflectionConstructable} */
   public static VTBurstLevelRateStatistic createInstance(SimNode ownNode, Parameters pars) {
      final int ldMinBlockSize = pars.get(ownNode, LD_MIN_BLOCK_SIZE).asInteger();
      final int ldMaxBlockSize = pars.get(ownNode, LD_MAX_BLOCK_SIZE).asInteger();
      final String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      final double timeBase = pars.get(ownNode, TIME_BASE).asDouble();
      return new VTBurstLevelRateStatistic(timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);
   }

   protected final double timeBase;

   protected double nextEndOfInterval;
   protected int noOfIntervals;
   protected double currentRate;
   protected double lastUpdateTime;
   protected final double[] blockSums;

   public VTBurstLevelRateStatistic(double timeBase, int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode) {
      super(ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);

      this.timeBase = timeBase;
      nextEndOfInterval = timeBase * minBlockSize;
      noOfIntervals = 0;
      currentRate = 0;
      lastUpdateTime = 0;
      blockSums = new double[ldMaxBlockSize - ldMinBlockSize + 1];
   }

   public void update(double currentTime, double newRate) {
      int i, blockSize;

      while (currentTime > nextEndOfInterval) {
         noOfIntervals += minBlockSize;
         blockSums[0] += currentRate * (currentTime - lastUpdateTime);
         for (i = 0, blockSize = minBlockSize; (i <= ldMaxBlockSize - ldMinBlockSize) && (noOfIntervals % blockSize == 0); i++, blockSize *= 2) {
            double blockMean = blockSums[i] / (blockSize * timeBase);
            blockMeansSummations[i].update(blockMean);
            if (traceFile != null) {
               traceFile.printf("VTBlock%d %g %g%n", i + ldMinBlockSize, currentTime, blockMean);
            }
            if (i < ldMaxBlockSize - ldMinBlockSize) {
               blockSums[i + 1] += blockSums[i];
            }
            blockSums[i] = 0;
         }
         lastUpdateTime = nextEndOfInterval;
         nextEndOfInterval += timeBase * minBlockSize;
      }
      blockSums[0] += currentRate * (currentTime - lastUpdateTime);
      lastUpdateTime = currentTime;
      currentRate = newRate;
   }
}
