/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.variancetime;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * <code>VTSampleStatistic</code> provides an implementation to estimate the VT
 * plot of samples.
 */
public class VTSampleStatistic extends VTStatistic {

   private static final String LD_MIN_BLOCK_SIZE = "LdMinBlockSize";
   private static final String LD_MAX_BLOCK_SIZE = "LdMaxBlockSize";
   private static final String TRACE_FILE = "TraceFile";

   /** as required by {@link ReflectionConstructable} */
   public static VTSampleStatistic createInstance(SimNode ownNode, Parameters pars) {
      final int ldMinBlockSize = pars.get(ownNode, LD_MIN_BLOCK_SIZE).asInteger();
      final int ldMaxBlockSize = pars.get(ownNode, LD_MAX_BLOCK_SIZE).asInteger();
      final String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      return new VTSampleStatistic(ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);
   }

   protected final double[] blockSums;
   protected int noOfSamples;

   public VTSampleStatistic(int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode) {
      super(ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);

      blockSums = new double[ldMaxBlockSize - ldMinBlockSize + 1];
      noOfSamples = 0;
   }

   public void update(double sample) {
      int i, blockSize;

      // faster implementation
      noOfSamples++;
      blockSums[0] += sample;
      for (i = 0, blockSize = minBlockSize; (i <= ldMaxBlockSize - ldMinBlockSize) && (noOfSamples % blockSize == 0); i++, blockSize *= 2) {
         double blockMean = blockSums[i] / blockSize;
         blockMeansSummations[i].update(blockMean);
         if (traceFile != null) {
            traceFile.printf("VTBlock%d %d %g%n", i + ldMinBlockSize, noOfSamples, blockMean);
         }
         if (i < ldMaxBlockSize - ldMinBlockSize) {
            blockSums[i + 1] += blockSums[i];
         }
         blockSums[i] = 0;
      }
   }

   public double getNoOfSamples() {
      return noOfSamples;
   }
}
